@@ -7,277 +7,242 @@
 #ifndef nsDOMMutationObserver_h
 #define nsDOMMutationObserver_h
 
+#include <utility>
+
 #include "mozilla/Attributes.h"
-#include "mozilla/Move.h"
-#include "nsCycleCollectionParticipant.h"
-#include "nsPIDOMWindow.h"
-#include "nsIScriptContext.h"
-#include "nsStubAnimationObserver.h"
-#include "nsCOMArray.h"
-#include "nsTArray.h"
-#include "nsIVariant.h"
-#include "nsContentList.h"
-#include "mozilla/dom/Element.h"
-#include "nsClassHashtable.h"
-#include "nsNodeUtils.h"
-#include "nsIDOMMutationEvent.h"
-#include "nsWrapperCache.h"
-#include "mozilla/dom/MutationObserverBinding.h"
-#include "nsIDocument.h"
 #include "mozilla/dom/Animation.h"
+#include "mozilla/dom/Document.h"
+#include "mozilla/dom/MutationObserverBinding.h"
+#include "mozilla/dom/Nullable.h"
+#include "nsCOMArray.h"
+#include "nsClassHashtable.h"
+#include "nsContentList.h"
+#include "nsCycleCollectionParticipant.h"
+#include "nsGlobalWindowInner.h"
 #include "nsIAnimationObserver.h"
-#include "nsGlobalWindow.h"
+#include "nsIMutationObserver.h"
+#include "nsPIDOMWindow.h"
+#include "nsStubAnimationObserver.h"
+#include "nsTArray.h"
+#include "nsWrapperCache.h"
+
+class nsIPrincipal;
 
 class nsDOMMutationObserver;
 using mozilla::dom::MutationObservingInfo;
 
-class nsDOMMutationRecord final : public nsISupports,
-                                  public nsWrapperCache
-{
-  virtual ~nsDOMMutationRecord() {}
+namespace mozilla::dom {
+class Element;
+}
 
-public:
-  typedef nsTArray<RefPtr<mozilla::dom::Animation>> AnimationArray;
+class nsDOMMutationRecord final : public nsISupports, public nsWrapperCache {
+  virtual ~nsDOMMutationRecord() = default;
+
+ public:
+  using AnimationArray = nsTArray<RefPtr<mozilla::dom::Animation>>;
 
   nsDOMMutationRecord(nsAtom* aType, nsISupports* aOwner)
-  : mType(aType), mAttrNamespace(VoidString()), mPrevValue(VoidString()), mOwner(aOwner)
-  {
-  }
+      : mType(aType),
+        mAttrNamespace(VoidString()),
+        mPrevValue(VoidString()),
+        mOwner(aOwner) {}
 
-  nsISupports* GetParentObject() const
-  {
-    return mOwner;
-  }
+  nsISupports* GetParentObject() const { return mOwner; }
 
-  virtual JSObject* WrapObject(JSContext* aCx, JS::Handle<JSObject*> aGivenProto) override
-  {
-    return mozilla::dom::MutationRecordBinding::Wrap(aCx, this, aGivenProto);
+  virtual JSObject* WrapObject(JSContext* aCx,
+                               JS::Handle<JSObject*> aGivenProto) override {
+    return mozilla::dom::MutationRecord_Binding::Wrap(aCx, this, aGivenProto);
   }
 
   NS_DECL_CYCLE_COLLECTING_ISUPPORTS
-  NS_DECL_CYCLE_COLLECTION_SCRIPT_HOLDER_CLASS(nsDOMMutationRecord)
+  NS_DECL_CYCLE_COLLECTION_WRAPPERCACHE_CLASS(nsDOMMutationRecord)
 
-  void GetType(mozilla::dom::DOMString& aRetVal) const
-  {
+  void GetType(mozilla::dom::DOMString& aRetVal) const {
     aRetVal.SetKnownLiveAtom(mType, mozilla::dom::DOMString::eNullNotExpected);
   }
 
-  nsINode* GetTarget() const
-  {
-    return mTarget;
-  }
+  nsINode* GetTarget() const { return mTarget; }
 
   nsINodeList* AddedNodes();
 
   nsINodeList* RemovedNodes();
 
-  nsINode* GetPreviousSibling() const
-  {
-    return mPreviousSibling;
-  }
+  nsINode* GetPreviousSibling() const { return mPreviousSibling; }
 
-  nsINode* GetNextSibling() const
-  {
-    return mNextSibling;
-  }
+  nsINode* GetNextSibling() const { return mNextSibling; }
 
-  void GetAttributeName(mozilla::dom::DOMString& aRetVal) const
-  {
+  void GetAttributeName(mozilla::dom::DOMString& aRetVal) const {
     aRetVal.SetKnownLiveAtom(mAttrName,
                              mozilla::dom::DOMString::eTreatNullAsNull);
   }
 
-  void GetAttributeNamespace(mozilla::dom::DOMString& aRetVal) const
-  {
+  void GetAttributeNamespace(mozilla::dom::DOMString& aRetVal) const {
     aRetVal.SetKnownLiveString(mAttrNamespace);
   }
 
-  void GetOldValue(mozilla::dom::DOMString& aRetVal) const
-  {
+  void GetOldValue(mozilla::dom::DOMString& aRetVal) const {
     aRetVal.SetKnownLiveString(mPrevValue);
   }
 
-  void GetAddedAnimations(AnimationArray& aRetVal) const
-  {
-    aRetVal = mAddedAnimations;
+  void GetAddedAnimations(AnimationArray& aRetVal) const {
+    aRetVal = mAddedAnimations.Clone();
   }
 
-  void GetRemovedAnimations(AnimationArray& aRetVal) const
-  {
-    aRetVal = mRemovedAnimations;
+  void GetRemovedAnimations(AnimationArray& aRetVal) const {
+    aRetVal = mRemovedAnimations.Clone();
   }
 
-  void GetChangedAnimations(AnimationArray& aRetVal) const
-  {
-    aRetVal = mChangedAnimations;
+  void GetChangedAnimations(AnimationArray& aRetVal) const {
+    aRetVal = mChangedAnimations.Clone();
   }
 
-  nsCOMPtr<nsINode>             mTarget;
-  RefPtr<nsAtom>             mType;
-  RefPtr<nsAtom>             mAttrName;
-  nsString                      mAttrNamespace;
-  nsString                      mPrevValue;
+  nsCOMPtr<nsINode> mTarget;
+  RefPtr<nsAtom> mType;
+  RefPtr<nsAtom> mAttrName;
+  nsString mAttrNamespace;
+  nsString mPrevValue;
   RefPtr<nsSimpleContentList> mAddedNodes;
   RefPtr<nsSimpleContentList> mRemovedNodes;
-  nsCOMPtr<nsINode>             mPreviousSibling;
-  nsCOMPtr<nsINode>             mNextSibling;
-  AnimationArray                mAddedAnimations;
-  AnimationArray                mRemovedAnimations;
-  AnimationArray                mChangedAnimations;
+  nsCOMPtr<nsINode> mPreviousSibling;
+  nsCOMPtr<nsINode> mNextSibling;
+  AnimationArray mAddedAnimations;
+  AnimationArray mRemovedAnimations;
+  AnimationArray mChangedAnimations;
 
   RefPtr<nsDOMMutationRecord> mNext;
-  nsCOMPtr<nsISupports>         mOwner;
+  nsCOMPtr<nsISupports> mOwner;
 };
 
 // Base class just prevents direct access to
 // members to make sure we go through getters/setters.
-class nsMutationReceiverBase : public nsStubAnimationObserver
-{
-public:
-  virtual ~nsMutationReceiverBase() { }
+class nsMutationReceiverBase : public nsStubAnimationObserver {
+ public:
+  virtual ~nsMutationReceiverBase() = default;
 
   nsDOMMutationObserver* Observer();
   nsINode* Target() { return mParent ? mParent->Target() : mTarget; }
   nsINode* RegisterTarget() { return mRegisterTarget; }
 
   bool Subtree() { return mParent ? mParent->Subtree() : mSubtree; }
-  void SetSubtree(bool aSubtree)
-  {
+  void SetSubtree(bool aSubtree) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mSubtree = aSubtree;
   }
 
   bool ChildList() { return mParent ? mParent->ChildList() : mChildList; }
-  void SetChildList(bool aChildList)
-  {
+  void SetChildList(bool aChildList) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mChildList = aChildList;
   }
 
-  bool CharacterData()
-  {
+  bool CharacterData() {
     return mParent ? mParent->CharacterData() : mCharacterData;
   }
-  void SetCharacterData(bool aCharacterData)
-  {
+  void SetCharacterData(bool aCharacterData) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mCharacterData = aCharacterData;
   }
 
-  bool CharacterDataOldValue()
-  {
+  bool CharacterDataOldValue() const {
     return mParent ? mParent->CharacterDataOldValue() : mCharacterDataOldValue;
   }
-  void SetCharacterDataOldValue(bool aOldValue)
-  {
+  void SetCharacterDataOldValue(bool aOldValue) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mCharacterDataOldValue = aOldValue;
   }
 
-  bool NativeAnonymousChildList()
-  {
-    return mParent ? mParent->NativeAnonymousChildList() : mNativeAnonymousChildList;
+  bool Attributes() const {
+    return mParent ? mParent->Attributes() : mAttributes;
   }
-  void SetNativeAnonymousChildList(bool aOldValue)
-  {
-    NS_ASSERTION(!mParent, "Shouldn't have parent");
-    mNativeAnonymousChildList = aOldValue;
-  }
-
-  bool Attributes() { return mParent ? mParent->Attributes() : mAttributes; }
-  void SetAttributes(bool aAttributes)
-  {
+  void SetAttributes(bool aAttributes) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mAttributes = aAttributes;
   }
 
-  bool AllAttributes()
-  {
-    return mParent ? mParent->AllAttributes()
-                   : mAllAttributes;
+  bool AllAttributes() const {
+    return mParent ? mParent->AllAttributes() : mAllAttributes;
   }
-  void SetAllAttributes(bool aAll)
-  {
+  void SetAllAttributes(bool aAll) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mAllAttributes = aAll;
   }
 
-  bool Animations() { return mParent ? mParent->Animations() : mAnimations; }
-  void SetAnimations(bool aAnimations)
-  {
+  bool Animations() const {
+    return mParent ? mParent->Animations() : mAnimations;
+  }
+  void SetAnimations(bool aAnimations) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mAnimations = aAnimations;
   }
 
-  bool AttributeOldValue() {
-    return mParent ? mParent->AttributeOldValue()
-                   : mAttributeOldValue;
+  bool AttributeOldValue() const {
+    return mParent ? mParent->AttributeOldValue() : mAttributeOldValue;
   }
-  void SetAttributeOldValue(bool aOldValue)
-  {
+  void SetAttributeOldValue(bool aOldValue) {
     NS_ASSERTION(!mParent, "Shouldn't have parent");
     mAttributeOldValue = aOldValue;
   }
 
-  nsTArray<RefPtr<nsAtom>>& AttributeFilter() { return mAttributeFilter; }
-  void SetAttributeFilter(nsTArray<RefPtr<nsAtom>>&& aFilter)
-  {
-    NS_ASSERTION(!mParent, "Shouldn't have parent");
-    mAttributeFilter.Clear();
-    mAttributeFilter = mozilla::Move(aFilter);
+  bool ChromeOnlyNodes() const {
+    return mParent ? mParent->ChromeOnlyNodes() : mChromeOnlyNodes;
   }
 
-  void AddClone(nsMutationReceiverBase* aClone)
-  {
+  void SetChromeOnlyNodes(bool aChromeOnlyNodes) {
+    NS_ASSERTION(!mParent, "Shouldn't have parent");
+    mChromeOnlyNodes = aChromeOnlyNodes;
+  }
+
+  nsTArray<RefPtr<nsAtom>>& AttributeFilter() { return mAttributeFilter; }
+  void SetAttributeFilter(nsTArray<RefPtr<nsAtom>>&& aFilter) {
+    NS_ASSERTION(!mParent, "Shouldn't have parent");
+    mAttributeFilter.Clear();
+    mAttributeFilter = std::move(aFilter);
+  }
+
+  void AddClone(nsMutationReceiverBase* aClone) {
     mTransientReceivers.AppendObject(aClone);
   }
 
-  void RemoveClone(nsMutationReceiverBase* aClone)
-  {
+  void RemoveClone(nsMutationReceiverBase* aClone) {
     mTransientReceivers.RemoveObject(aClone);
   }
 
-protected:
+ protected:
   nsMutationReceiverBase(nsINode* aTarget, nsDOMMutationObserver* aObserver)
-    : mTarget(aTarget)
-    , mObserver(aObserver)
-    , mRegisterTarget(aTarget)
-    , mSubtree(false)
-    , mChildList(false)
-    , mCharacterData(false)
-    , mCharacterDataOldValue(false)
-    , mNativeAnonymousChildList(false)
-    , mAttributes(false)
-    , mAllAttributes(false)
-    , mAttributeOldValue(false)
-    , mAnimations(false)
-  {
-  }
+      : mTarget(aTarget),
+        mObserver(aObserver),
+        mRegisterTarget(aTarget),
+        mSubtree(false),
+        mChildList(false),
+        mCharacterData(false),
+        mCharacterDataOldValue(false),
+        mAttributes(false),
+        mAllAttributes(false),
+        mAttributeOldValue(false),
+        mAnimations(false) {}
 
   nsMutationReceiverBase(nsINode* aRegisterTarget,
                          nsMutationReceiverBase* aParent)
-    : mTarget(nullptr)
-    , mObserver(nullptr)
-    , mParent(aParent)
-    , mRegisterTarget(aRegisterTarget)
-    , mKungFuDeathGrip(aParent->Target())
-    , mSubtree(false)
-    , mChildList(false)
-    , mCharacterData(false)
-    , mCharacterDataOldValue(false)
-    , mNativeAnonymousChildList(false)
-    , mAttributes(false)
-    , mAllAttributes(false)
-    , mAttributeOldValue(false)
-    , mAnimations(false)
-  {
+      : mTarget(nullptr),
+        mObserver(nullptr),
+        mParent(aParent),
+        mRegisterTarget(aRegisterTarget),
+        mKungFuDeathGrip(aParent->Target()),
+        mSubtree(false),
+        mChildList(false),
+        mCharacterData(false),
+        mCharacterDataOldValue(false),
+        mAttributes(false),
+        mAllAttributes(false),
+        mAttributeOldValue(false),
+        mAnimations(false),
+        mChromeOnlyNodes(false) {
     NS_ASSERTION(mParent->Subtree(), "Should clone a non-subtree observer!");
   }
 
   virtual void AddMutationObserver() = 0;
 
-  void AddObserver()
-  {
+  void AddObserver() {
     AddMutationObserver();
     mRegisterTarget->SetMayHaveDOMMutationObserver();
     mRegisterTarget->OwnerDoc()->SetMayHaveDOMMutationObservers();
@@ -285,103 +250,68 @@ protected:
 
   bool IsObservable(nsIContent* aContent);
 
-  bool ObservesAttr(nsINode* aRegisterTarget,
-                    mozilla::dom::Element* aElement,
-                    int32_t aNameSpaceID,
-                    nsAtom* aAttr)
-  {
-    if (mParent) {
-      return mParent->ObservesAttr(aRegisterTarget, aElement, aNameSpaceID, aAttr);
-    }
-    if (!Attributes() ||
-        (!Subtree() && aElement != Target()) ||
-        (Subtree() && aRegisterTarget->SubtreeRoot() != aElement->SubtreeRoot()) ||
-        !IsObservable(aElement)) {
-      return false;
-    }
-    if (AllAttributes()) {
-      return true;
-    }
-
-    if (aNameSpaceID != kNameSpaceID_None) {
-      return false;
-    }
-
-    nsTArray<RefPtr<nsAtom>>& filters = AttributeFilter();
-    for (size_t i = 0; i < filters.Length(); ++i) {
-      if (filters[i] == aAttr) {
-        return true;
-      }
-    }
-    return false;
-  }
+  bool ObservesAttr(nsINode* aRegisterTarget, mozilla::dom::Element* aElement,
+                    int32_t aNameSpaceID, nsAtom* aAttr);
 
   // The target for the MutationObserver.observe() method.
-  nsINode*                           mTarget;
-  nsDOMMutationObserver*             mObserver;
-  RefPtr<nsMutationReceiverBase>   mParent; // Cleared after microtask.
+  nsINode* mTarget;
+  nsDOMMutationObserver* mObserver;
+  RefPtr<nsMutationReceiverBase> mParent;  // Cleared after microtask.
   // The node to which Gecko-internal nsIMutationObserver was registered to.
   // This is different than mTarget when dealing with transient observers.
-  nsINode*                           mRegisterTarget;
+  nsINode* mRegisterTarget;
   nsCOMArray<nsMutationReceiverBase> mTransientReceivers;
   // While we have transient receivers, keep the original mutation receiver
   // alive so it doesn't go away and disconnect all its transient receivers.
-  nsCOMPtr<nsINode>                  mKungFuDeathGrip;
+  nsCOMPtr<nsINode> mKungFuDeathGrip;
 
-private:
-  bool                               mSubtree;
-  bool                               mChildList;
-  bool                               mCharacterData;
-  bool                               mCharacterDataOldValue;
-  bool                               mNativeAnonymousChildList;
-  bool                               mAttributes;
-  bool                               mAllAttributes;
-  bool                               mAttributeOldValue;
-  bool                               mAnimations;
-  nsTArray<RefPtr<nsAtom>>          mAttributeFilter;
+ private:
+  nsTArray<RefPtr<nsAtom>> mAttributeFilter;
+  bool mSubtree : 1;
+  bool mChildList : 1;
+  bool mCharacterData : 1;
+  bool mCharacterDataOldValue : 1;
+  bool mAttributes : 1;
+  bool mAllAttributes : 1;
+  bool mAttributeOldValue : 1;
+  bool mAnimations : 1;
+  bool mChromeOnlyNodes : 1;
 };
 
-
-class nsMutationReceiver : public nsMutationReceiverBase
-{
-protected:
+class nsMutationReceiver : public nsMutationReceiverBase {
+ protected:
   virtual ~nsMutationReceiver() { Disconnect(false); }
 
-public:
+ public:
   static nsMutationReceiver* Create(nsINode* aTarget,
-                                    nsDOMMutationObserver* aObserver)
-  {
+                                    nsDOMMutationObserver* aObserver) {
     nsMutationReceiver* r = new nsMutationReceiver(aTarget, aObserver);
     r->AddObserver();
     return r;
   }
 
   static nsMutationReceiver* Create(nsINode* aRegisterTarget,
-                                    nsMutationReceiverBase* aParent)
-  {
+                                    nsMutationReceiverBase* aParent) {
     nsMutationReceiver* r = new nsMutationReceiver(aRegisterTarget, aParent);
     aParent->AddClone(r);
     r->AddObserver();
     return r;
   }
 
-  nsMutationReceiver* GetParent()
-  {
+  nsMutationReceiver* GetParent() {
     return static_cast<nsMutationReceiver*>(mParent.get());
   }
 
-  void RemoveClones()
-  {
+  void RemoveClones() {
     for (int32_t i = 0; i < mTransientReceivers.Count(); ++i) {
       nsMutationReceiver* r =
-        static_cast<nsMutationReceiver*>(mTransientReceivers[i]);
+          static_cast<nsMutationReceiver*>(mTransientReceivers[i]);
       r->DisconnectTransientReceiver();
     }
     mTransientReceivers.Clear();
   }
 
-  void DisconnectTransientReceiver()
-  {
+  void DisconnectTransientReceiver() {
     if (mRegisterTarget) {
       mRegisterTarget->RemoveMutationObserver(this);
       mRegisterTarget = nullptr;
@@ -397,53 +327,45 @@ public:
   NS_DECL_ISUPPORTS
 
   NS_DECL_NSIMUTATIONOBSERVER_ATTRIBUTEWILLCHANGE
-  NS_DECL_NSIMUTATIONOBSERVER_NATIVEANONYMOUSCHILDLISTCHANGE
   NS_DECL_NSIMUTATIONOBSERVER_CHARACTERDATAWILLCHANGE
   NS_DECL_NSIMUTATIONOBSERVER_CONTENTAPPENDED
   NS_DECL_NSIMUTATIONOBSERVER_CONTENTINSERTED
   NS_DECL_NSIMUTATIONOBSERVER_CONTENTREMOVED
   NS_DECL_NSIMUTATIONOBSERVER_NODEWILLBEDESTROYED
 
-  virtual void AttributeSetToCurrentValue(nsIDocument* aDocument,
-                                          mozilla::dom::Element* aElement,
+  virtual void AttributeSetToCurrentValue(mozilla::dom::Element* aElement,
                                           int32_t aNameSpaceID,
-                                          nsAtom* aAttribute) override
-  {
+                                          nsAtom* aAttribute) override {
     // We can reuse AttributeWillChange implementation.
-    AttributeWillChange(aDocument, aElement, aNameSpaceID, aAttribute,
-                        nsIDOMMutationEvent::MODIFICATION, nullptr);
+    AttributeWillChange(aElement, aNameSpaceID, aAttribute,
+                        AttrModType::Modification);
   }
 
-protected:
+ protected:
   nsMutationReceiver(nsINode* aTarget, nsDOMMutationObserver* aObserver);
 
   nsMutationReceiver(nsINode* aRegisterTarget, nsMutationReceiverBase* aParent)
-  : nsMutationReceiverBase(aRegisterTarget, aParent)
-  {
+      : nsMutationReceiverBase(aRegisterTarget, aParent) {
     NS_ASSERTION(!static_cast<nsMutationReceiver*>(aParent)->GetParent(),
                  "Shouldn't create deep observer hierarchies!");
   }
 
-  virtual void AddMutationObserver() override
-  {
+  virtual void AddMutationObserver() override {
     mRegisterTarget->AddMutationObserver(this);
   }
 };
 
-class nsAnimationReceiver : public nsMutationReceiver
-{
-public:
+class nsAnimationReceiver : public nsMutationReceiver {
+ public:
   static nsAnimationReceiver* Create(nsINode* aTarget,
-                                     nsDOMMutationObserver* aObserver)
-  {
+                                     nsDOMMutationObserver* aObserver) {
     nsAnimationReceiver* r = new nsAnimationReceiver(aTarget, aObserver);
     r->AddObserver();
     return r;
   }
 
   static nsAnimationReceiver* Create(nsINode* aRegisterTarget,
-                                     nsMutationReceiverBase* aParent)
-  {
+                                     nsMutationReceiverBase* aParent) {
     nsAnimationReceiver* r = new nsAnimationReceiver(aRegisterTarget, aParent);
     aParent->AddClone(r);
     r->AddObserver();
@@ -456,21 +378,20 @@ public:
   NS_DECL_NSIANIMATIONOBSERVER_ANIMATIONCHANGED
   NS_DECL_NSIANIMATIONOBSERVER_ANIMATIONREMOVED
 
-protected:
-  virtual ~nsAnimationReceiver() {}
+ protected:
+  virtual ~nsAnimationReceiver() = default;
 
   nsAnimationReceiver(nsINode* aTarget, nsDOMMutationObserver* aObserver)
-    : nsMutationReceiver(aTarget, aObserver) {}
+      : nsMutationReceiver(aTarget, aObserver) {}
 
   nsAnimationReceiver(nsINode* aRegisterTarget, nsMutationReceiverBase* aParent)
-    : nsMutationReceiver(aRegisterTarget, aParent) {}
+      : nsMutationReceiver(aRegisterTarget, aParent) {}
 
-  virtual void AddMutationObserver() override
-  {
+  virtual void AddMutationObserver() override {
     mRegisterTarget->AddAnimationObserver(this);
   }
 
-private:
+ private:
   enum AnimationMutation {
     eAnimationMutation_Added,
     eAnimationMutation_Changed,
@@ -482,69 +403,53 @@ private:
 };
 
 #define NS_DOM_MUTATION_OBSERVER_IID \
-{ 0x0c3b91f8, 0xcc3b, 0x4b08, \
-  { 0x9e, 0xab, 0x07, 0x47, 0xa9, 0xe4, 0x65, 0xb4 } }
+  {0x0c3b91f8, 0xcc3b, 0x4b08, {0x9e, 0xab, 0x07, 0x47, 0xa9, 0xe4, 0x65, 0xb4}}
 
-class nsDOMMutationObserver final : public nsISupports,
-                                    public nsWrapperCache
-{
-public:
-  nsDOMMutationObserver(already_AddRefed<nsPIDOMWindowInner>&& aOwner,
-                        mozilla::dom::MutationCallback& aCb,
-                        bool aChrome)
-  : mOwner(aOwner), mLastPendingMutation(nullptr), mPendingMutationCount(0),
-    mCallback(&aCb), mWaitingForRun(false), mIsChrome(aChrome),
-    mMergeAttributeRecords(false), mId(++sCount)
-  {
-  }
+class nsDOMMutationObserver final : public nsISupports, public nsWrapperCache {
+ public:
+  nsDOMMutationObserver(nsCOMPtr<nsPIDOMWindowInner>&& aOwner,
+                        mozilla::dom::MutationCallback& aCb)
+      : mOwner(std::move(aOwner)),
+        mLastPendingMutation(nullptr),
+        mPendingMutationCount(0),
+        mCallback(&aCb),
+        mWaitingForRun(false),
+        mMergeAttributeRecords(false),
+        mId(++sCount) {}
   NS_DECL_CYCLE_COLLECTING_ISUPPORTS
   NS_DECL_CYCLE_COLLECTION_SCRIPT_HOLDER_CLASS(nsDOMMutationObserver)
-  NS_DECLARE_STATIC_IID_ACCESSOR(NS_DOM_MUTATION_OBSERVER_IID)
+  NS_INLINE_DECL_STATIC_IID(NS_DOM_MUTATION_OBSERVER_IID)
 
-  static already_AddRefed<nsDOMMutationObserver>
-  Constructor(const mozilla::dom::GlobalObject& aGlobal,
-              mozilla::dom::MutationCallback& aCb,
-              mozilla::ErrorResult& aRv);
+  static already_AddRefed<nsDOMMutationObserver> Constructor(
+      const mozilla::dom::GlobalObject&, mozilla::dom::MutationCallback&,
+      mozilla::ErrorResult&);
 
-  virtual JSObject* WrapObject(JSContext* aCx, JS::Handle<JSObject*> aGivenProto) override
-  {
-    return mozilla::dom::MutationObserverBinding::Wrap(aCx, this, aGivenProto);
+  JSObject* WrapObject(JSContext* aCx,
+                       JS::Handle<JSObject*> aGivenProto) override {
+    return mozilla::dom::MutationObserver_Binding::Wrap(aCx, this, aGivenProto);
   }
 
-  nsISupports* GetParentObject() const
-  {
-    return mOwner;
-  }
-
-  bool IsChrome()
-  {
-    return mIsChrome;
-  }
+  nsISupports* GetParentObject() const { return mOwner; }
 
   void Observe(nsINode& aTarget,
                const mozilla::dom::MutationObserverInit& aOptions,
-               mozilla::ErrorResult& aRv);
+               nsIPrincipal& aSubjectPrincipal, mozilla::ErrorResult& aRv);
 
   void Disconnect();
 
-  void TakeRecords(nsTArray<RefPtr<nsDOMMutationRecord> >& aRetVal);
+  void TakeRecords(nsTArray<RefPtr<nsDOMMutationRecord>>& aRetVal);
 
-  void HandleMutation();
+  MOZ_CAN_RUN_SCRIPT void HandleMutation();
 
-  void GetObservingInfo(nsTArray<Nullable<MutationObservingInfo>>& aResult,
-                        mozilla::ErrorResult& aRv);
+  void GetObservingInfo(
+      nsTArray<mozilla::dom::Nullable<MutationObservingInfo>>& aResult,
+      mozilla::ErrorResult& aRv);
 
   mozilla::dom::MutationCallback* MutationCallback() { return mCallback; }
 
-  bool MergeAttributeRecords()
-  {
-    return mMergeAttributeRecords;
-  }
+  bool MergeAttributeRecords() { return mMergeAttributeRecords; }
 
-  void SetMergeAttributeRecords(bool aVal)
-  {
-    mMergeAttributeRecords = aVal;
-  }
+  void SetMergeAttributeRecords(bool aVal) { mMergeAttributeRecords = aVal; }
 
   // If both records are for 'attributes' type and for the same target and
   // attribute name and namespace are the same, we can skip the newer record.
@@ -552,36 +457,39 @@ public:
   bool MergeableAttributeRecord(nsDOMMutationRecord* aOldRecord,
                                 nsDOMMutationRecord* aRecord);
 
-  void AppendMutationRecord(already_AddRefed<nsDOMMutationRecord> aRecord)
-  {
+  void AppendMutationRecord(already_AddRefed<nsDOMMutationRecord> aRecord) {
     RefPtr<nsDOMMutationRecord> record = aRecord;
     MOZ_ASSERT(record);
     if (!mLastPendingMutation) {
       MOZ_ASSERT(!mFirstPendingMutation);
-      mFirstPendingMutation = record.forget();
+      mFirstPendingMutation = std::move(record);
       mLastPendingMutation = mFirstPendingMutation;
     } else {
       MOZ_ASSERT(mFirstPendingMutation);
-      mLastPendingMutation->mNext = record.forget();
+      mLastPendingMutation->mNext = std::move(record);
       mLastPendingMutation = mLastPendingMutation->mNext;
     }
     ++mPendingMutationCount;
   }
 
-  void ClearPendingRecords()
-  {
-    mFirstPendingMutation = nullptr;
+  void ClearPendingRecords() {
+    // Break down the pending mutation record list so that cycle collector
+    // can delete the objects sooner.
+    RefPtr<nsDOMMutationRecord> current = std::move(mFirstPendingMutation);
     mLastPendingMutation = nullptr;
     mPendingMutationCount = 0;
+    while (current) {
+      current = std::move(current->mNext);
+    }
   }
 
   // static methods
   static void QueueMutationObserverMicroTask();
 
+  MOZ_CAN_RUN_SCRIPT
   static void HandleMutations(mozilla::AutoSlowOperation& aAso);
 
-  static bool AllScheduledMutationObserversAreSuppressed()
-  {
+  static bool AllScheduledMutationObserversAreSuppressed() {
     if (sScheduledMutationObservers) {
       uint32_t len = sScheduledMutationObservers->Length();
       if (len > 0) {
@@ -600,20 +508,18 @@ public:
   static void LeaveMutationHandling();
 
   static void Shutdown();
-protected:
+
+ protected:
   virtual ~nsDOMMutationObserver();
 
   friend class nsMutationReceiver;
   friend class nsAnimationReceiver;
   friend class nsAutoMutationBatch;
   friend class nsAutoAnimationMutationBatch;
-  nsMutationReceiver* GetReceiverFor(nsINode* aNode,
-                                     bool aMayCreate,
+  nsMutationReceiver* GetReceiverFor(nsINode* aNode, bool aMayCreate,
                                      bool aWantsAnimations);
   void RemoveReceiver(nsMutationReceiver* aReceiver);
 
-  already_AddRefed<nsIVariant> TakeRecords();
-
   void GetAllSubtreeObserversFor(nsINode* aNode,
                                  nsTArray<nsMutationReceiver*>& aObservers);
   void ScheduleForRun();
@@ -622,66 +528,65 @@ protected:
   nsDOMMutationRecord* CurrentRecord(nsAtom* aType);
   bool HasCurrentRecord(const nsAString& aType);
 
-  bool Suppressed()
-  {
+  bool Suppressed() {
     return mOwner && nsGlobalWindowInner::Cast(mOwner)->IsInSyncOperation();
   }
 
+  MOZ_CAN_RUN_SCRIPT
   static void HandleMutationsInternal(mozilla::AutoSlowOperation& aAso);
 
   static void AddCurrentlyHandlingObserver(nsDOMMutationObserver* aObserver,
                                            uint32_t aMutationLevel);
 
-  nsCOMPtr<nsPIDOMWindowInner>                       mOwner;
+  nsCOMPtr<nsPIDOMWindowInner> mOwner;
 
-  nsCOMArray<nsMutationReceiver>                     mReceivers;
-  nsClassHashtable<nsISupportsHashKey,
-                   nsCOMArray<nsMutationReceiver> >  mTransientReceivers;
+  nsCOMArray<nsMutationReceiver> mReceivers;
+  nsClassHashtable<nsISupportsHashKey, nsCOMArray<nsMutationReceiver>>
+      mTransientReceivers;
   // MutationRecords which are being constructed.
-  AutoTArray<nsDOMMutationRecord*, 4>              mCurrentMutations;
+  AutoTArray<nsDOMMutationRecord*, 4> mCurrentMutations;
   // MutationRecords which will be handed to the callback at the end of
   // the microtask.
-  RefPtr<nsDOMMutationRecord>                      mFirstPendingMutation;
-  nsDOMMutationRecord*                               mLastPendingMutation;
-  uint32_t                                           mPendingMutationCount;
+  RefPtr<nsDOMMutationRecord> mFirstPendingMutation;
+  nsDOMMutationRecord* mLastPendingMutation;
+  uint32_t mPendingMutationCount;
 
-  RefPtr<mozilla::dom::MutationCallback>           mCallback;
+  RefPtr<mozilla::dom::MutationCallback> mCallback;
 
-  bool                                               mWaitingForRun;
-  bool                                               mIsChrome;
-  bool                                               mMergeAttributeRecords;
+  bool mWaitingForRun;
+  bool mMergeAttributeRecords;
 
-  uint64_t                                           mId;
+  uint64_t mId;
 
-  static uint64_t                                    sCount;
-  static AutoTArray<RefPtr<nsDOMMutationObserver>, 4>* sScheduledMutationObservers;
+  static uint64_t sCount;
+  static AutoTArray<RefPtr<nsDOMMutationObserver>, 4>*
+      sScheduledMutationObservers;
 
-  static uint32_t                                    sMutationLevel;
+  static uint32_t sMutationLevel;
   static AutoTArray<AutoTArray<RefPtr<nsDOMMutationObserver>, 4>, 4>*
-                                                     sCurrentlyHandlingObservers;
+      sCurrentlyHandlingObservers;
 };
 
-NS_DEFINE_STATIC_IID_ACCESSOR(nsDOMMutationObserver, NS_DOM_MUTATION_OBSERVER_IID)
-
-class nsAutoMutationBatch
-{
-public:
+class nsAutoMutationBatch {
+ public:
   nsAutoMutationBatch()
-  : mPreviousBatch(nullptr), mBatchTarget(nullptr), mRemovalDone(false),
-    mFromFirstToLast(false), mAllowNestedBatches(false)
-  {
-  }
+      : mPreviousBatch(nullptr),
+        mBatchTarget(nullptr),
+        mRemovalDone(false),
+        mFromFirstToLast(false),
+        mAllowNestedBatches(false) {}
 
   nsAutoMutationBatch(nsINode* aTarget, bool aFromFirstToLast,
                       bool aAllowNestedBatches)
-  : mPreviousBatch(nullptr), mBatchTarget(nullptr), mRemovalDone(false),
-    mFromFirstToLast(false), mAllowNestedBatches(false)
-  {
+      : mPreviousBatch(nullptr),
+        mBatchTarget(nullptr),
+        mRemovalDone(false),
+        mFromFirstToLast(false),
+        mAllowNestedBatches(false) {
     Init(aTarget, aFromFirstToLast, aAllowNestedBatches);
   }
 
-  void Init(nsINode* aTarget, bool aFromFirstToLast, bool aAllowNestedBatches)
-  {
+  void Init(nsINode* aTarget, bool aFromFirstToLast, bool aAllowNestedBatches) {
     if (aTarget && aTarget->OwnerDoc()->MayHaveDOMMutationObservers()) {
       if (sCurrentBatch && !sCurrentBatch->mAllowNestedBatches) {
         return;
@@ -705,19 +610,12 @@ public:
 
   ~nsAutoMutationBatch() { NodesAdded(); }
 
-  static bool IsBatching()
-  {
-    return !!sCurrentBatch;
-  }
+  static bool IsBatching() { return !!sCurrentBatch; }
 
-  static nsAutoMutationBatch* GetCurrentBatch()
-  {
-    return sCurrentBatch;
-  }
+  static nsAutoMutationBatch* GetCurrentBatch() { return sCurrentBatch; }
 
   static void UpdateObserver(nsDOMMutationObserver* aObserver,
-                             bool aWantsChildList)
-  {
+                             bool aWantsChildList) {
     uint32_t l = sCurrentBatch->mObservers.Length();
     for (uint32_t i = 0; i < l; ++i) {
       if (sCurrentBatch->mObservers[i].mObserver == aObserver) {
@@ -732,49 +630,43 @@ public:
     bo->mWantsChildList = aWantsChildList;
   }
 
-
   static nsINode* GetBatchTarget() { return sCurrentBatch->mBatchTarget; }
 
   // Mutation receivers notify the batch about removed child nodes.
-  static void NodeRemoved(nsIContent* aChild)
-  {
+  static void NodeRemoved(nsIContent* aChild) {
     if (IsBatching() && !sCurrentBatch->mRemovalDone) {
       uint32_t len = sCurrentBatch->mRemovedNodes.Length();
-      if (!len ||
-          sCurrentBatch->mRemovedNodes[len - 1] != aChild) {
+      if (!len || sCurrentBatch->mRemovedNodes[len - 1] != aChild) {
         sCurrentBatch->mRemovedNodes.AppendElement(aChild);
       }
     }
   }
 
   // Called after new child nodes have been added to the batch target.
-  void NodesAdded()
-  {
+  void NodesAdded() {
     if (sCurrentBatch != this) {
       return;
     }
 
-    nsIContent* c =
-      mPrevSibling ? mPrevSibling->GetNextSibling() :
-                     mBatchTarget->GetFirstChild();
+    nsIContent* c = mPrevSibling ? mPrevSibling->GetNextSibling()
+                                 : mBatchTarget->GetFirstChild();
     for (; c != mNextSibling; c = c->GetNextSibling()) {
       mAddedNodes.AppendElement(c);
     }
     Done();
   }
 
-private:
-  struct BatchObserver
-  {
+ private:
+  struct BatchObserver {
     nsDOMMutationObserver* mObserver;
-    bool                   mWantsChildList;
+    bool mWantsChildList;
   };
 
   static nsAutoMutationBatch* sCurrentBatch;
   nsAutoMutationBatch* mPreviousBatch;
   AutoTArray<BatchObserver, 2> mObservers;
-  nsTArray<nsCOMPtr<nsIContent> > mRemovedNodes;
-  nsTArray<nsCOMPtr<nsIContent> > mAddedNodes;
+  nsTArray<nsCOMPtr<nsIContent>> mRemovedNodes;
+  nsTArray<nsCOMPtr<nsIContent>> mAddedNodes;
   nsINode* mBatchTarget;
   bool mRemovalDone;
   bool mFromFirstToLast;
@@ -783,20 +675,16 @@ private:
   nsCOMPtr<nsINode> mNextSibling;
 };
 
-class nsAutoAnimationMutationBatch
-{
+class nsAutoAnimationMutationBatch {
   struct Entry;
 
-public:
-  explicit nsAutoAnimationMutationBatch(nsIDocument* aDocument)
-  {
+ public:
+  explicit nsAutoAnimationMutationBatch(mozilla::dom::Document* aDocument) {
     Init(aDocument);
   }
 
-  void Init(nsIDocument* aDocument)
-  {
-    if (!aDocument ||
-        !aDocument->MayHaveDOMMutationObservers() ||
+  void Init(mozilla::dom::Document* aDocument) {
+    if (!aDocument || !aDocument->MayHaveDOMMutationObservers() ||
         sCurrentBatch) {
       return;
     }
@@ -805,25 +693,17 @@ public:
     nsDOMMutationObserver::EnterMutationHandling();
   }
 
-  ~nsAutoAnimationMutationBatch()
-  {
-    Done();
-  }
+  ~nsAutoAnimationMutationBatch() { Done(); }
 
   void Done();
 
-  static bool IsBatching()
-  {
-    return !!sCurrentBatch;
-  }
+  static bool IsBatching() { return !!sCurrentBatch; }
 
-  static nsAutoAnimationMutationBatch* GetCurrentBatch()
-  {
+  static nsAutoAnimationMutationBatch* GetCurrentBatch() {
     return sCurrentBatch;
   }
 
-  static void AddObserver(nsDOMMutationObserver* aObserver)
-  {
+  static void AddObserver(nsDOMMutationObserver* aObserver) {
     if (sCurrentBatch->mObservers.Contains(aObserver)) {
       return;
     }
@@ -831,8 +711,7 @@ public:
   }
 
   static void AnimationAdded(mozilla::dom::Animation* aAnimation,
-                             nsINode* aTarget)
-  {
+                             nsINode* aTarget) {
     if (!IsBatching()) {
       return;
     }
@@ -846,9 +725,15 @@ public:
         case eState_Removed:
           entry->mState = eState_RemainedPresent;
           break;
-        default:
-          NS_NOTREACHED("shouldn't have observed an animation being added "
-                        "twice");
+        case eState_Added:
+          // FIXME bug 1189015
+          NS_ERROR("shouldn't have observed an animation being added twice");
+          break;
+        case eState_RemainedPresent:
+          MOZ_ASSERT_UNREACHABLE(
+              "shouldn't have observed an animation "
+              "remaining present");
+          break;
       }
     } else {
       entry = sCurrentBatch->AddEntry(aAnimation, aTarget);
@@ -858,12 +743,11 @@ public:
   }
 
   static void AnimationChanged(mozilla::dom::Animation* aAnimation,
-                               nsINode* aTarget)
-  {
+                               nsINode* aTarget) {
     Entry* entry = sCurrentBatch->FindEntry(aAnimation, aTarget);
     if (entry) {
       NS_ASSERTION(entry->mState == eState_RemainedPresent ||
-                   entry->mState == eState_Added,
+                       entry->mState == eState_Added,
                    "shouldn't have observed an animation being changed after "
                    "being removed");
       entry->mChanged = true;
@@ -875,8 +759,7 @@ public:
   }
 
   static void AnimationRemoved(mozilla::dom::Animation* aAnimation,
-                               nsINode* aTarget)
-  {
+                               nsINode* aTarget) {
     Entry* entry = sCurrentBatch->FindEntry(aAnimation, aTarget);
     if (entry) {
       switch (entry->mState) {
@@ -886,9 +769,15 @@ public:
         case eState_Added:
           entry->mState = eState_RemainedAbsent;
           break;
-        default:
-          NS_NOTREACHED("shouldn't have observed an animation being removed "
-                        "twice");
+        case eState_RemainedAbsent:
+          MOZ_ASSERT_UNREACHABLE(
+              "shouldn't have observed an animation "
+              "remaining absent");
+          break;
+        case eState_Removed:
+          // FIXME bug 1189015
+          NS_ERROR("shouldn't have observed an animation being removed twice");
+          break;
       }
     } else {
       entry = sCurrentBatch->AddEntry(aAnimation, aTarget);
@@ -897,9 +786,8 @@ public:
     }
   }
 
-private:
-  Entry* FindEntry(mozilla::dom::Animation* aAnimation, nsINode* aTarget)
-  {
+ private:
+  Entry* FindEntry(mozilla::dom::Animation* aAnimation, nsINode* aTarget) {
     EntryArray* entries = mEntryTable.Get(aTarget);
     if (!entries) {
       return nullptr;
@@ -913,9 +801,8 @@ private:
     return nullptr;
   }
 
-  Entry* AddEntry(mozilla::dom::Animation* aAnimation, nsINode* aTarget)
-  {
-    EntryArray* entries = sCurrentBatch->mEntryTable.LookupOrAdd(aTarget);
+  Entry* AddEntry(mozilla::dom::Animation* aAnimation, nsINode* aTarget) {
+    EntryArray* entries = sCurrentBatch->mEntryTable.GetOrInsertNew(aTarget);
     if (entries->IsEmpty()) {
       sCurrentBatch->mBatchTargets.AppendElement(aTarget);
     }
@@ -931,8 +818,7 @@ private:
     eState_Removed
   };
 
-  struct Entry
-  {
+  struct Entry {
     RefPtr<mozilla::dom::Animation> mAnimation;
     State mState;
     bool mChanged;
@@ -940,7 +826,7 @@ private:
 
   static nsAutoAnimationMutationBatch* sCurrentBatch;
   AutoTArray<nsDOMMutationObserver*, 2> mObservers;
-  typedef nsTArray<Entry> EntryArray;
+  using EntryArray = nsTArray<Entry>;
   nsClassHashtable<nsPtrHashKey<nsINode>, EntryArray> mEntryTable;
   // List of nodes referred to by mEntryTable so we can sort them
   // For a specific pseudo element, we use its parent element as the
@@ -948,12 +834,27 @@ private:
   nsTArray<nsINode*> mBatchTargets;
 };
 
-inline
-nsDOMMutationObserver*
-nsMutationReceiverBase::Observer()
-{
-  return mParent ?
-    mParent->Observer() : static_cast<nsDOMMutationObserver*>(mObserver);
+inline nsDOMMutationObserver* nsMutationReceiverBase::Observer() {
+  return mParent ? mParent->Observer()
+                 : static_cast<nsDOMMutationObserver*>(mObserver);
 }
 
+class MOZ_RAII nsDOMMutationEnterLeave {
+ public:
+  explicit nsDOMMutationEnterLeave(mozilla::dom::Document* aDoc)
+      : mNeeded(aDoc->MayHaveDOMMutationObservers()) {
+    if (mNeeded) {
+      nsDOMMutationObserver::EnterMutationHandling();
+    }
+  }
+  ~nsDOMMutationEnterLeave() {
+    if (mNeeded) {
+      nsDOMMutationObserver::LeaveMutationHandling();
+    }
+  }
+
+ private:
+  const bool mNeeded;
+};
+
 #endif
