@@ -9,13 +9,13 @@
 
 #ifdef USE_CAIRO
 #  include "DrawTargetCairo.h"
+#  include "PathCairo.h"
 #  include "SourceSurfaceCairo.h"
 #endif
 
-#ifdef USE_SKIA
-#  include "DrawTargetSkia.h"
-#  include "ScaledFontBase.h"
-#endif
+#include "DrawTargetSkia.h"
+#include "PathSkia.h"
+#include "ScaledFontBase.h"
 
 #if defined(WIN32)
 #  include "ScaledFontWin.h"
@@ -42,41 +42,31 @@
 #endif
 
 #ifdef WIN32
-#  include "DrawTargetD2D1.h"
 #  include "ScaledFontDWrite.h"
 #  include "NativeFontResourceDWrite.h"
 #  include "UnscaledFontDWrite.h"
 #  include <d3d10_1.h>
 #  include <stdlib.h>
-#  include "HelpersD2D.h"
-#  include "DXVA2Manager.h"
+#  include "HelpersWin.h"
+#  include "ImageContainer.h"
+#  include "mozilla/layers/LayersSurfaces.h"
 #  include "mozilla/layers/TextureD3D11.h"
+#  include "mozilla/layers/VideoProcessorD3D11.h"
 #  include "nsWindowsHelpers.h"
 #endif
 
-#include "DrawTargetCapture.h"
-#include "DrawTargetDual.h"
-#include "DrawTargetTiled.h"
 #include "DrawTargetOffset.h"
-#include "DrawTargetWrapAndRecord.h"
 #include "DrawTargetRecording.h"
+#include "PathRecording.h"
 
 #include "SourceSurfaceRawData.h"
 
-#include "DrawEventRecorder.h"
-
-#include "Logging.h"
-
 #include "mozilla/CheckedInt.h"
 
-#include "mozilla/layers/TextureClient.h"
-
 #ifdef MOZ_ENABLE_FREETYPE
 #  include "ft2build.h"
 #  include FT_FREETYPE_H
 #endif
-#include "MainThreadUtils.h"
-#include "mozilla/Preferences.h"
 #include "mozilla/StaticPrefs_gfx.h"
 
 #if defined(MOZ_LOGGING)
@@ -184,11 +174,12 @@ void mozilla_ForgetSharedFTFaceLockOwner(void* aContext, void* aOwner) {
   static_cast<mozilla::gfx::SharedFTFace*>(aContext)->ForgetLockOwner(aOwner);
 }
 
-int mozilla_LockSharedFTFace(void* aContext, void* aOwner) {
+int mozilla_LockSharedFTFace(void* aContext,
+                             void* aOwner) MOZ_NO_THREAD_SAFETY_ANALYSIS {
   return int(static_cast<mozilla::gfx::SharedFTFace*>(aContext)->Lock(aOwner));
 }
 
-void mozilla_UnlockSharedFTFace(void* aContext) {
+void mozilla_UnlockSharedFTFace(void* aContext) MOZ_NO_THREAD_SAFETY_ANALYSIS {
   static_cast<mozilla::gfx::SharedFTFace*>(aContext)->Unlock();
 }
 
@@ -212,25 +203,36 @@ namespace mozilla::gfx {
 #ifdef MOZ_ENABLE_FREETYPE
 FT_Library Factory::mFTLibrary = nullptr;
 StaticMutex Factory::mFTLock;
+
+already_AddRefed<SharedFTFace> FTUserFontData::CloneFace(int aFaceIndex) {
+  if (mFontData) {
+    RefPtr<SharedFTFace> face = Factory::NewSharedFTFaceFromData(
+        nullptr, mFontData, mLength, aFaceIndex, this);
+    if (!face ||
+        (FT_Select_Charmap(face->GetFace(), FT_ENCODING_UNICODE) != FT_Err_Ok &&
+         FT_Select_Charmap(face->GetFace(), FT_ENCODING_MS_SYMBOL) !=
+             FT_Err_Ok)) {
+      return nullptr;
+    }
+    return face.forget();
+  }
+  FT_Face face = Factory::NewFTFace(nullptr, mFilename.c_str(), aFaceIndex);
+  if (face) {
+    return MakeAndAddRef<SharedFTFace>(face, this);
+  }
+  return nullptr;
+}
 #endif
 
 #ifdef WIN32
 // Note: mDeviceLock must be held when mutating these values.
-static uint32_t mDeviceSeq = 0;
 StaticRefPtr<ID3D11Device> Factory::mD3D11Device;
-StaticRefPtr<ID2D1Device> Factory::mD2D1Device;
 StaticRefPtr<IDWriteFactory> Factory::mDWriteFactory;
-StaticRefPtr<ID2D1DeviceContext> Factory::mMTDC;
-StaticRefPtr<ID2D1DeviceContext> Factory::mOffMTDC;
-bool Factory::mDWriteFactoryInitialized = false;
 StaticRefPtr<IDWriteFontCollection> Factory::mDWriteSystemFonts;
 StaticMutex Factory::mDeviceLock;
-StaticMutex Factory::mDTDependencyLock;
 #endif
 
-bool Factory::mBGRSubpixelOrder = false;
-
-DrawEventRecorder* Factory::mRecorder;
+SubpixelOrder Factory::mSubpixelOrder = SubpixelOrder::UNKNOWN;
 
 mozilla::gfx::Config* Factory::sConfig = nullptr;
 
@@ -243,6 +245,8 @@ void Factory::Init(const Config& aConfig) {
 #else
   NativeFontResource::RegisterMemoryReporter();
 #endif
+
+  SourceSurfaceAlignedRawData::RegisterMemoryReporter();
 }
 
 void Factory::ShutDown() {
@@ -319,10 +323,6 @@ bool Factory::AllowedSurfaceSize(const IntSize& aSize) {
   return CheckSurfaceSize(aSize);
 }
 
-bool Factory::CheckBufferSize(int32_t bufSize) {
-  return !sConfig || bufSize < sConfig->mMaxAllocSize;
-}
-
 bool Factory::CheckSurfaceSize(const IntSize& sz, int32_t extentLimit,
                                int32_t allocLimit) {
   if (sz.width <= 0 || sz.height <= 0) {
@@ -369,17 +369,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTarget(BackendType aBackend,
 
   RefPtr<DrawTarget> retVal;
   switch (aBackend) {
-#ifdef WIN32
-    case BackendType::DIRECT2D1_1: {
-      RefPtr<DrawTargetD2D1> newTarget;
-      newTarget = new DrawTargetD2D1();
-      if (newTarget->Init(aSize, aFormat)) {
-        retVal = newTarget;
-      }
-      break;
-    }
-#endif
-#ifdef USE_SKIA
     case BackendType::SKIA: {
       RefPtr<DrawTargetSkia> newTarget;
       newTarget = new DrawTargetSkia();
@@ -388,7 +377,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTarget(BackendType aBackend,
       }
       break;
     }
-#endif
 #ifdef USE_CAIRO
     case BackendType::CAIRO: {
       RefPtr<DrawTargetCairo> newTarget;
@@ -403,10 +391,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTarget(BackendType aBackend,
       return nullptr;
   }
 
-  if (mRecorder && retVal) {
-    return MakeAndAddRef<DrawTargetWrapAndRecord>(mRecorder, retVal);
-  }
-
   if (!retVal) {
     // Failed
     gfxCriticalError(LoggerOptionsBasedOnSize(aSize))
@@ -417,15 +401,27 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTarget(BackendType aBackend,
   return retVal.forget();
 }
 
-already_AddRefed<PathBuilder> Factory::CreateSimplePathBuilder() {
-  RefPtr<PathBuilder> pathBuilder;
-#ifdef USE_SKIA
-  pathBuilder = MakeAndAddRef<PathBuilderSkia>(FillRule::FILL_WINDING);
+already_AddRefed<PathBuilder> Factory::CreatePathBuilder(BackendType aBackend,
+                                                         FillRule aFillRule) {
+  switch (aBackend) {
+    case BackendType::SKIA:
+    case BackendType::WEBGL:
+      return PathBuilderSkia::Create(aFillRule);
+#ifdef USE_CAIRO
+    case BackendType::CAIRO:
+      return PathBuilderCairo::Create(aFillRule);
 #endif
-  if (!pathBuilder) {
-    NS_WARNING("Failed to create a path builder because we don't use Skia");
+    case BackendType::RECORDING:
+      return do_AddRef(new PathBuilderRecording(BackendType::SKIA, aFillRule));
+    default:
+      gfxCriticalNote << "Invalid PathBuilder type specified: "
+                      << (int)aBackend;
+      return nullptr;
   }
-  return pathBuilder.forget();
+}
+
+already_AddRefed<PathBuilder> Factory::CreateSimplePathBuilder() {
+  return CreatePathBuilder(BackendType::SKIA);
 }
 
 already_AddRefed<DrawTarget> Factory::CreateRecordingDrawTarget(
@@ -433,35 +429,10 @@ already_AddRefed<DrawTarget> Factory::CreateRecordingDrawTarget(
   return MakeAndAddRef<DrawTargetRecording>(aRecorder, aDT, aRect);
 }
 
-already_AddRefed<DrawTargetCapture> Factory::CreateCaptureDrawTargetForTarget(
-    gfx::DrawTarget* aTarget, size_t aFlushBytes) {
-  return MakeAndAddRef<DrawTargetCaptureImpl>(aTarget, aFlushBytes);
-}
-
-already_AddRefed<DrawTargetCapture> Factory::CreateCaptureDrawTarget(
-    BackendType aBackend, const IntSize& aSize, SurfaceFormat aFormat) {
-  return MakeAndAddRef<DrawTargetCaptureImpl>(aBackend, aSize, aFormat);
-}
-
-already_AddRefed<DrawTargetCapture> Factory::CreateCaptureDrawTargetForData(
-    BackendType aBackend, const IntSize& aSize, SurfaceFormat aFormat,
-    int32_t aStride, size_t aSurfaceAllocationSize) {
-  MOZ_ASSERT(aSurfaceAllocationSize && aStride);
-
-  BackendType type = aBackend;
-  if (!Factory::DoesBackendSupportDataDrawtarget(aBackend)) {
-    type = BackendType::SKIA;
-  }
-
-  RefPtr<DrawTargetCaptureImpl> dt =
-      new DrawTargetCaptureImpl(type, aSize, aFormat);
-  dt->InitForData(aStride, aSurfaceAllocationSize);
-  return dt.forget();
-}
-
 already_AddRefed<DrawTarget> Factory::CreateDrawTargetForData(
     BackendType aBackend, unsigned char* aData, const IntSize& aSize,
-    int32_t aStride, SurfaceFormat aFormat, bool aUninitialized) {
+    int32_t aStride, SurfaceFormat aFormat, bool aUninitialized,
+    bool aIsClear) {
   MOZ_ASSERT(aData);
   if (!AllowedSurfaceSize(aSize)) {
     gfxCriticalError(LoggerOptionsBasedOnSize(aSize))
@@ -472,16 +443,15 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTargetForData(
   RefPtr<DrawTarget> retVal;
 
   switch (aBackend) {
-#ifdef USE_SKIA
     case BackendType::SKIA: {
       RefPtr<DrawTargetSkia> newTarget;
       newTarget = new DrawTargetSkia();
-      if (newTarget->Init(aData, aSize, aStride, aFormat, aUninitialized)) {
+      if (newTarget->Init(aData, aSize, aStride, aFormat, aUninitialized,
+                          aIsClear)) {
         retVal = newTarget;
       }
       break;
     }
-#endif
 #ifdef USE_CAIRO
     case BackendType::CAIRO: {
       RefPtr<DrawTargetCairo> newTarget;
@@ -498,10 +468,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTargetForData(
       return nullptr;
   }
 
-  if (mRecorder && retVal) {
-    return MakeAndAddRef<DrawTargetWrapAndRecord>(mRecorder, retVal, true);
-  }
-
   if (!retVal) {
     gfxCriticalNote << "Failed to create DrawTarget, Type: " << int(aBackend)
                     << " Size: " << aSize << ", Data: " << hexa((void*)aData)
@@ -511,17 +477,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTargetForData(
   return retVal.forget();
 }
 
-already_AddRefed<DrawTarget> Factory::CreateTiledDrawTarget(
-    const TileSet& aTileSet) {
-  RefPtr<DrawTargetTiled> dt = new DrawTargetTiled();
-
-  if (!dt->Init(aTileSet)) {
-    return nullptr;
-  }
-
-  return dt.forget();
-}
-
 already_AddRefed<DrawTarget> Factory::CreateOffsetDrawTarget(
     DrawTarget* aDrawTarget, IntPoint aTileOrigin) {
   RefPtr<DrawTargetOffset> dt = new DrawTargetOffset();
@@ -535,13 +490,11 @@ already_AddRefed<DrawTarget> Factory::CreateOffsetDrawTarget(
 
 bool Factory::DoesBackendSupportDataDrawtarget(BackendType aType) {
   switch (aType) {
-    case BackendType::DIRECT2D:
-    case BackendType::DIRECT2D1_1:
     case BackendType::RECORDING:
-    case BackendType::CAPTURE:
     case BackendType::NONE:
     case BackendType::BACKEND_LAST:
     case BackendType::WEBRENDER_TEXT:
+    case BackendType::WEBGL:
       return false;
     case BackendType::CAIRO:
     case BackendType::SKIA:
@@ -555,21 +508,16 @@ uint32_t Factory::GetMaxSurfaceSize(BackendType aType) {
   switch (aType) {
     case BackendType::CAIRO:
       return DrawTargetCairo::GetMaxSurfaceSize();
-#ifdef USE_SKIA
     case BackendType::SKIA:
       return DrawTargetSkia::GetMaxSurfaceSize();
-#endif
-#ifdef WIN32
-    case BackendType::DIRECT2D1_1:
-      return DrawTargetD2D1::GetMaxSurfaceSize();
-#endif
     default:
       return 0;
   }
 }
 
 already_AddRefed<NativeFontResource> Factory::CreateNativeFontResource(
-    uint8_t* aData, uint32_t aSize, FontType aFontType, void* aFontContext) {
+    const uint8_t* aData, uint32_t aSize, FontType aFontType,
+    void* aFontContext) {
   switch (aFontType) {
 #ifdef WIN32
     case FontType::DWRITE:
@@ -628,11 +576,10 @@ already_AddRefed<UnscaledFont> Factory::CreateUnscaledFontFromFontDescriptor(
 #ifdef XP_DARWIN
 already_AddRefed<ScaledFont> Factory::CreateScaledFontForMacFont(
     CGFontRef aCGFont, const RefPtr<UnscaledFont>& aUnscaledFont, Float aSize,
-    const DeviceColor& aFontSmoothingBackgroundColor, bool aUseFontSmoothing,
-    bool aApplySyntheticBold) {
+    bool aUseFontSmoothing, bool aApplySyntheticBold, bool aHasColorGlyphs) {
   return MakeAndAddRef<ScaledFontMac>(aCGFont, aUnscaledFont, aSize, false,
-                                      aFontSmoothingBackgroundColor,
-                                      aUseFontSmoothing, aApplySyntheticBold);
+                                      aUseFontSmoothing, aApplySyntheticBold,
+                                      aHasColorGlyphs);
 }
 #endif
 
@@ -654,33 +601,11 @@ already_AddRefed<ScaledFont> Factory::CreateScaledFontForFreeTypeFont(
 }
 #endif
 
-already_AddRefed<DrawTarget> Factory::CreateDualDrawTarget(
-    DrawTarget* targetA, DrawTarget* targetB) {
-  MOZ_ASSERT(targetA && targetB);
-
-  RefPtr<DrawTarget> newTarget = new DrawTargetDual(targetA, targetB);
-
-  RefPtr<DrawTarget> retVal = newTarget;
-
-  if (mRecorder) {
-    retVal = new DrawTargetWrapAndRecord(mRecorder, retVal);
-  }
-
-  return retVal.forget();
+void Factory::SetSubpixelOrder(SubpixelOrder aOrder) {
+  mSubpixelOrder = aOrder;
 }
 
-already_AddRefed<SourceSurface> Factory::CreateDualSourceSurface(
-    SourceSurface* sourceA, SourceSurface* sourceB) {
-  MOZ_ASSERT(sourceA && sourceB);
-
-  RefPtr<SourceSurface> newSource = new SourceSurfaceDual(sourceA, sourceB);
-
-  return newSource.forget();
-}
-
-void Factory::SetBGRSubpixelOrder(bool aBGR) { mBGRSubpixelOrder = aBGR; }
-
-bool Factory::GetBGRSubpixelOrder() { return mBGRSubpixelOrder; }
+SubpixelOrder Factory::GetSubpixelOrder() { return mSubpixelOrder; }
 
 #ifdef MOZ_ENABLE_FREETYPE
 SharedFTFace::SharedFTFace(FT_Face aFace, SharedFTFaceData* aData)
@@ -719,9 +644,15 @@ void Factory::ReleaseFTLibrary(FT_Library aFTLibrary) {
   FT_Done_FreeType(aFTLibrary);
 }
 
-void Factory::LockFTLibrary(FT_Library aFTLibrary) { mFTLock.Lock(); }
+void Factory::LockFTLibrary(FT_Library aFTLibrary)
+    MOZ_CAPABILITY_ACQUIRE(mFTLock) MOZ_NO_THREAD_SAFETY_ANALYSIS {
+  mFTLock.Lock();
+}
 
-void Factory::UnlockFTLibrary(FT_Library aFTLibrary) { mFTLock.Unlock(); }
+void Factory::UnlockFTLibrary(FT_Library aFTLibrary)
+    MOZ_CAPABILITY_RELEASE(mFTLock) MOZ_NO_THREAD_SAFETY_ANALYSIS {
+  mFTLock.Unlock();
+}
 
 FT_Face Factory::NewFTFace(FT_Library aFTLibrary, const char* aFileName,
                            int aFaceIndex) {
@@ -739,11 +670,21 @@ FT_Face Factory::NewFTFace(FT_Library aFTLibrary, const char* aFileName,
 already_AddRefed<SharedFTFace> Factory::NewSharedFTFace(FT_Library aFTLibrary,
                                                         const char* aFilename,
                                                         int aFaceIndex) {
-  if (FT_Face face = NewFTFace(aFTLibrary, aFilename, aFaceIndex)) {
-    return MakeAndAddRef<SharedFTFace>(face);
-  } else {
+  FT_Face face = NewFTFace(aFTLibrary, aFilename, aFaceIndex);
+  if (!face) {
     return nullptr;
   }
+
+  RefPtr<FTUserFontData> data;
+#  ifdef ANDROID
+  // If the font has variations, we may later need to "clone" it in
+  // UnscaledFontFreeType::CreateScaledFont. To support this, we attach an
+  // FTUserFontData that records the filename used to instantiate the face.
+  if (face->face_flags & FT_FACE_FLAG_MULTIPLE_MASTERS) {
+    data = new FTUserFontData(aFilename);
+  }
+#  endif
+  return MakeAndAddRef<SharedFTFace>(face, data);
 }
 
 FT_Face Factory::NewFTFaceFromData(FT_Library aFTLibrary, const uint8_t* aData,
@@ -783,91 +724,13 @@ FT_Error Factory::LoadFTGlyph(FT_Face aFace, uint32_t aGlyphIndex,
 }
 #endif
 
-AutoSerializeWithMoz2D::AutoSerializeWithMoz2D(BackendType aBackendType) {
 #ifdef WIN32
-  // We use a multi-threaded ID2D1Factory1, so that makes the calls through the
-  // Direct2D API thread-safe. However, if the Moz2D objects are using Direct3D
-  // resources we need to make sure that calls through the Direct3D or DXGI API
-  // use the Direct2D synchronization. It's possible that this should be pushed
-  // down into the TextureD3D11 objects, so that we always use this.
-  if (aBackendType == BackendType::DIRECT2D1_1 ||
-      aBackendType == BackendType::DIRECT2D) {
-    D2DFactory()->QueryInterface(
-        static_cast<ID2D1Multithread**>(getter_AddRefs(mMT)));
-    mMT->Enter();
-  }
-#endif
-}
-
-AutoSerializeWithMoz2D::~AutoSerializeWithMoz2D() {
-#ifdef WIN32
-  if (mMT) {
-    mMT->Leave();
-  }
-#endif
-};
-
-#ifdef WIN32
-already_AddRefed<DrawTarget> Factory::CreateDrawTargetForD3D11Texture(
-    ID3D11Texture2D* aTexture, SurfaceFormat aFormat) {
-  MOZ_ASSERT(aTexture);
-
-  RefPtr<DrawTargetD2D1> newTarget;
-
-  newTarget = new DrawTargetD2D1();
-  if (newTarget->Init(aTexture, aFormat)) {
-    RefPtr<DrawTarget> retVal = newTarget;
-
-    if (mRecorder) {
-      retVal = new DrawTargetWrapAndRecord(mRecorder, retVal, true);
-    }
-
-    return retVal.forget();
-  }
-
-  gfxWarning() << "Failed to create draw target for D3D11 texture.";
-
-  // Failed
-  return nullptr;
-}
-
 bool Factory::SetDirect3D11Device(ID3D11Device* aDevice) {
   MOZ_RELEASE_ASSERT(NS_IsMainThread());
 
-  // D2DFactory already takes the device lock, so we get the factory before
-  // entering the lock scope.
-  RefPtr<ID2D1Factory1> factory = D2DFactory();
-
   StaticMutexAutoLock lock(mDeviceLock);
 
   mD3D11Device = aDevice;
-
-  if (mD2D1Device) {
-    mD2D1Device = nullptr;
-    mMTDC = nullptr;
-    mOffMTDC = nullptr;
-  }
-
-  if (!aDevice) {
-    return true;
-  }
-
-  RefPtr<IDXGIDevice> device;
-  aDevice->QueryInterface((IDXGIDevice**)getter_AddRefs(device));
-
-  RefPtr<ID2D1Device> d2dDevice;
-  HRESULT hr = factory->CreateDevice(device, getter_AddRefs(d2dDevice));
-  if (FAILED(hr)) {
-    gfxCriticalError()
-        << "[D2D1] Failed to create gfx factory's D2D1 device, code: "
-        << hexa(hr);
-
-    mD3D11Device = nullptr;
-    return false;
-  }
-
-  mDeviceSeq++;
-  mD2D1Device = d2dDevice;
   return true;
 }
 
@@ -876,16 +739,6 @@ RefPtr<ID3D11Device> Factory::GetDirect3D11Device() {
   return mD3D11Device;
 }
 
-RefPtr<ID2D1Device> Factory::GetD2D1Device(uint32_t* aOutSeqNo) {
-  StaticMutexAutoLock lock(mDeviceLock);
-  if (aOutSeqNo) {
-    *aOutSeqNo = mDeviceSeq;
-  }
-  return mD2D1Device.get();
-}
-
-bool Factory::HasD2D1Device() { return !!GetD2D1Device(); }
-
 RefPtr<IDWriteFactory> Factory::GetDWriteFactory() {
   StaticMutexAutoLock lock(mDeviceLock);
   return mDWriteFactory;
@@ -894,12 +747,10 @@ RefPtr<IDWriteFactory> Factory::GetDWriteFactory() {
 RefPtr<IDWriteFactory> Factory::EnsureDWriteFactory() {
   StaticMutexAutoLock lock(mDeviceLock);
 
-  if (mDWriteFactoryInitialized) {
+  if (mDWriteFactory) {
     return mDWriteFactory;
   }
 
-  mDWriteFactoryInitialized = true;
-
   HMODULE dwriteModule = LoadLibrarySystem32(L"dwrite.dll");
   decltype(DWriteCreateFactory)* createDWriteFactory =
       (decltype(DWriteCreateFactory)*)GetProcAddress(dwriteModule,
@@ -958,73 +809,18 @@ RefPtr<IDWriteFontCollection> Factory::GetDWriteSystemFonts(bool aUpdate) {
   return mDWriteSystemFonts;
 }
 
-RefPtr<ID2D1DeviceContext> Factory::GetD2DDeviceContext() {
-  StaticRefPtr<ID2D1DeviceContext>* ptr;
-
-  if (NS_IsMainThread()) {
-    ptr = &mMTDC;
-  } else {
-    ptr = &mOffMTDC;
-  }
-
-  if (*ptr) {
-    return *ptr;
-  }
-
-  RefPtr<ID2D1Device> device = GetD2D1Device();
-
-  if (!device) {
-    return nullptr;
-  }
-
-  RefPtr<ID2D1DeviceContext> dc;
-  HRESULT hr = device->CreateDeviceContext(
-      D2D1_DEVICE_CONTEXT_OPTIONS_ENABLE_MULTITHREADED_OPTIMIZATIONS,
-      getter_AddRefs(dc));
-
-  if (FAILED(hr)) {
-    gfxCriticalError() << "Failed to create global device context";
-    return nullptr;
-  }
-
-  *ptr = dc;
-
-  return *ptr;
-}
-
-bool Factory::SupportsD2D1() { return !!D2DFactory(); }
-
 BYTE sSystemTextQuality = CLEARTYPE_QUALITY;
 void Factory::SetSystemTextQuality(uint8_t aQuality) {
   sSystemTextQuality = aQuality;
 }
 
-uint64_t Factory::GetD2DVRAMUsageDrawTarget() {
-  return DrawTargetD2D1::mVRAMUsageDT;
-}
-
-uint64_t Factory::GetD2DVRAMUsageSourceSurface() {
-  return DrawTargetD2D1::mVRAMUsageSS;
-}
-
-void Factory::D2DCleanup() {
-  StaticMutexAutoLock lock(mDeviceLock);
-  if (mD2D1Device) {
-    mD2D1Device = nullptr;
-  }
-  DrawTargetD2D1::CleanupD2D();
-}
-
 already_AddRefed<ScaledFont> Factory::CreateScaledFontForDWriteFont(
     IDWriteFontFace* aFontFace, const gfxFontStyle* aStyle,
     const RefPtr<UnscaledFont>& aUnscaledFont, float aSize,
-    bool aUseEmbeddedBitmap, int aRenderingMode,
-    IDWriteRenderingParams* aParams, Float aGamma, Float aContrast,
-    Float aClearTypeLevel) {
+    bool aUseEmbeddedBitmap, bool aUseMultistrikeBold, bool aGDIForced) {
   return MakeAndAddRef<ScaledFontDWrite>(
-      aFontFace, aUnscaledFont, aSize, aUseEmbeddedBitmap,
-      (DWRITE_RENDERING_MODE)aRenderingMode, aParams, aGamma, aContrast,
-      aClearTypeLevel, aStyle);
+      aFontFace, aUnscaledFont, aSize, aUseEmbeddedBitmap, aUseMultistrikeBold,
+      aGDIForced, aStyle);
 }
 
 already_AddRefed<ScaledFont> Factory::CreateScaledFontForGDIFont(
@@ -1035,7 +831,6 @@ already_AddRefed<ScaledFont> Factory::CreateScaledFontForGDIFont(
 }
 #endif  // WIN32
 
-#ifdef USE_SKIA
 already_AddRefed<DrawTarget> Factory::CreateDrawTargetWithSkCanvas(
     SkCanvas* aCanvas) {
   RefPtr<DrawTargetSkia> newTarget = new DrawTargetSkia();
@@ -1044,7 +839,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTargetWithSkCanvas(
   }
   return newTarget.forget();
 }
-#endif
 
 void Factory::PurgeAllCaches() {}
 
@@ -1062,10 +856,6 @@ already_AddRefed<DrawTarget> Factory::CreateDrawTargetForCairoSurface(
   if (newTarget->Init(aSurface, aSize, aFormat)) {
     retVal = newTarget;
   }
-
-  if (mRecorder && retVal) {
-    return MakeAndAddRef<DrawTargetWrapAndRecord>(mRecorder, retVal, true);
-  }
 #endif
   return retVal.forget();
 }
@@ -1156,6 +946,40 @@ already_AddRefed<DataSourceSurface> Factory::CreateDataSourceSurfaceWithStride(
   return nullptr;
 }
 
+already_AddRefed<DataSourceSurface> Factory::CopyDataSourceSurface(
+    DataSourceSurface* aSource) {
+  // Don't worry too much about speed.
+  MOZ_ASSERT(aSource->GetFormat() == SurfaceFormat::R8G8B8A8 ||
+             aSource->GetFormat() == SurfaceFormat::R8G8B8X8 ||
+             aSource->GetFormat() == SurfaceFormat::B8G8R8A8 ||
+             aSource->GetFormat() == SurfaceFormat::B8G8R8X8);
+
+  DataSourceSurface::ScopedMap srcMap(aSource, DataSourceSurface::READ);
+  if (NS_WARN_IF(!srcMap.IsMapped())) {
+    MOZ_ASSERT_UNREACHABLE("CopyDataSourceSurface: Failed to map surface.");
+    return nullptr;
+  }
+
+  IntSize size = aSource->GetSize();
+  SurfaceFormat format = aSource->GetFormat();
+
+  RefPtr<DataSourceSurface> dst = CreateDataSourceSurfaceWithStride(
+      size, format, srcMap.GetStride(), /* aZero */ false);
+  if (NS_WARN_IF(!dst)) {
+    return nullptr;
+  }
+
+  DataSourceSurface::ScopedMap dstMap(dst, DataSourceSurface::WRITE);
+  if (NS_WARN_IF(!dstMap.IsMapped())) {
+    MOZ_ASSERT_UNREACHABLE("CopyDataSourceSurface: Failed to map surface.");
+    return nullptr;
+  }
+
+  SwizzleData(srcMap.GetData(), srcMap.GetStride(), format, dstMap.GetData(),
+              dstMap.GetStride(), format, size);
+  return dst.forget();
+}
+
 void Factory::CopyDataSourceSurface(DataSourceSurface* aSource,
                                     DataSourceSurface* aDest) {
   // Don't worry too much about speed.
@@ -1185,21 +1009,13 @@ void Factory::CopyDataSourceSurface(DataSourceSurface* aSource,
   aDest->Unmap();
 }
 
-already_AddRefed<DrawEventRecorder> Factory::CreateEventRecorderForFile(
-    const char_type* aFilename) {
-  return MakeAndAddRef<DrawEventRecorderFile>(aFilename);
-}
-
-void Factory::SetGlobalEventRecorder(DrawEventRecorder* aRecorder) {
-  mRecorder = aRecorder;
-}
-
 #ifdef WIN32
 
 /* static */
 already_AddRefed<DataSourceSurface>
 Factory::CreateBGRA8DataSourceSurfaceForD3D11Texture(
-    ID3D11Texture2D* aSrcTexture) {
+    ID3D11Texture2D* aSrcTexture, uint32_t aArrayIndex,
+    gfx::ColorSpace2 aColorSpace, gfx::ColorRange aColorRange) {
   D3D11_TEXTURE2D_DESC srcDesc = {0};
   aSrcTexture->GetDesc(&srcDesc);
 
@@ -1209,16 +1025,50 @@ Factory::CreateBGRA8DataSourceSurfaceForD3D11Texture(
   if (NS_WARN_IF(!destTexture)) {
     return nullptr;
   }
-  if (!ReadbackTexture(destTexture, aSrcTexture)) {
+  if (!ReadbackTexture(destTexture, aSrcTexture, aArrayIndex, aColorSpace,
+                       aColorRange)) {
     return nullptr;
   }
   return destTexture.forget();
 }
 
+/* static */ nsresult Factory::CreateSdbForD3D11Texture(
+    ID3D11Texture2D* aSrcTexture, const IntSize& aSrcSize,
+    layers::SurfaceDescriptorBuffer& aSdBuffer,
+    const std::function<layers::MemoryOrShmem(uint32_t)>& aAllocate) {
+  D3D11_TEXTURE2D_DESC srcDesc = {0};
+  aSrcTexture->GetDesc(&srcDesc);
+  if (srcDesc.Width != uint32_t(aSrcSize.width) ||
+      srcDesc.Height != uint32_t(aSrcSize.height) ||
+      srcDesc.Format != DXGI_FORMAT_B8G8R8A8_UNORM) {
+    return NS_ERROR_NOT_IMPLEMENTED;
+  }
+
+  const auto format = gfx::SurfaceFormat::B8G8R8A8;
+  uint8_t* buffer = nullptr;
+  int32_t stride = 0;
+  nsresult rv = layers::Image::AllocateSurfaceDescriptorBufferRgb(
+      aSrcSize, format, buffer, aSdBuffer, stride, aAllocate);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return rv;
+  }
+
+  if (!ReadbackTexture(buffer, stride, aSrcTexture)) {
+    return NS_ERROR_FAILURE;
+  }
+
+  return NS_OK;
+}
+
 /* static */
-template <typename DestTextureT>
-bool Factory::ConvertSourceAndRetryReadback(DestTextureT* aDestCpuTexture,
-                                            ID3D11Texture2D* aSrcTexture) {
+bool Factory::ConvertSourceAndRetryReadback(DataSourceSurface* aDestCpuTexture,
+                                            ID3D11Texture2D* aSrcTexture,
+                                            uint32_t aArrayIndex,
+                                            gfx::ColorSpace2 aColorSpace,
+                                            gfx::ColorRange aColorRange) {
+  MOZ_ASSERT(aDestCpuTexture);
+  MOZ_ASSERT(aSrcTexture);
+
   RefPtr<ID3D11Device> device;
   aSrcTexture->GetDevice(getter_AddRefs(device));
   if (!device) {
@@ -1226,56 +1076,57 @@ bool Factory::ConvertSourceAndRetryReadback(DestTextureT* aDestCpuTexture,
     return false;
   }
 
-  nsAutoCString error;
-  std::unique_ptr<DXVA2Manager> manager(
-      DXVA2Manager::CreateD3D11DXVA(nullptr, error, device));
-  if (!manager) {
-    gfxWarning() << "Failed to create DXVA2 manager!";
+  CD3D11_TEXTURE2D_DESC desc;
+  aSrcTexture->GetDesc(&desc);
+
+  if (desc.Format != DXGI_FORMAT_NV12 && desc.Format != DXGI_FORMAT_P010 &&
+      desc.Format != DXGI_FORMAT_P016) {
+    gfxWarning() << "Unexpected DXGI format";
     return false;
   }
 
+  desc = CD3D11_TEXTURE2D_DESC(
+      DXGI_FORMAT_B8G8R8A8_UNORM, desc.Width, desc.Height, 1, 1,
+      D3D11_BIND_RENDER_TARGET | D3D11_BIND_SHADER_RESOURCE);
+
   RefPtr<ID3D11Texture2D> newSrcTexture;
   HRESULT hr =
-      manager->CopyToBGRATexture(aSrcTexture, getter_AddRefs(newSrcTexture));
+      device->CreateTexture2D(&desc, nullptr, getter_AddRefs(newSrcTexture));
   if (FAILED(hr)) {
-    gfxWarning() << "Failed to copy to BGRA texture.";
+    gfxWarning() << "Failed to create newSrcTexture: " << gfx::hexa(hr);
     return false;
   }
 
-  return ReadbackTexture(aDestCpuTexture, newSrcTexture);
-}
-
-/* static */
-bool Factory::ReadbackTexture(layers::TextureData* aDestCpuTexture,
-                              ID3D11Texture2D* aSrcTexture) {
-  layers::MappedTextureData mappedData;
-  if (!aDestCpuTexture->BorrowMappedData(mappedData)) {
-    gfxWarning() << "Could not access in-memory texture";
+  RefPtr<layers::VideoProcessorD3D11> videoProcessor =
+      layers::VideoProcessorD3D11::Create(device);
+  if (!videoProcessor) {
+    gfxWarning() << "Failed to create VideoProcessorD3D11";
     return false;
   }
 
-  D3D11_TEXTURE2D_DESC srcDesc = {0};
-  aSrcTexture->GetDesc(&srcDesc);
-
-  // Special case: If the source and destination have different formats and the
-  // destination is B8G8R8A8 then convert the source to B8G8R8A8 and readback.
-  if ((srcDesc.Format != DXGIFormat(mappedData.format)) &&
-      (mappedData.format == SurfaceFormat::B8G8R8A8)) {
-    return ConvertSourceAndRetryReadback(aDestCpuTexture, aSrcTexture);
-  }
-
-  if ((IntSize(srcDesc.Width, srcDesc.Height) != mappedData.size) ||
-      (srcDesc.Format != DXGIFormat(mappedData.format))) {
-    gfxWarning() << "Attempted readback between incompatible textures";
+  hr = videoProcessor->Init(aDestCpuTexture->GetSize());
+  if (FAILED(hr)) {
+    gfxWarning() << "Failed to init VideoProcessorD3D11" << gfx::hexa(hr);
     return false;
   }
 
-  return ReadbackTexture(mappedData.data, mappedData.stride, aSrcTexture);
+  layers::VideoProcessorD3D11::InputTextureInfo info(aColorSpace, aColorRange,
+                                                     aArrayIndex, aSrcTexture);
+  if (!videoProcessor->CallVideoProcessorBlt(info, newSrcTexture)) {
+    gfxWarning() << "CallVideoProcessorBlt failed";
+    return false;
+  }
+
+  return ReadbackTexture(aDestCpuTexture, newSrcTexture, 0, aColorSpace,
+                         aColorRange);
 }
 
 /* static */
 bool Factory::ReadbackTexture(DataSourceSurface* aDestCpuTexture,
-                              ID3D11Texture2D* aSrcTexture) {
+                              ID3D11Texture2D* aSrcTexture,
+                              uint32_t aArrayIndex,
+                              gfx::ColorSpace2 aColorSpace,
+                              gfx::ColorRange aColorRange) {
   D3D11_TEXTURE2D_DESC srcDesc = {0};
   aSrcTexture->GetDesc(&srcDesc);
 
@@ -1283,7 +1134,8 @@ bool Factory::ReadbackTexture(DataSourceSurface* aDestCpuTexture,
   // destination is B8G8R8A8 then convert the source to B8G8R8A8 and readback.
   if ((srcDesc.Format != DXGIFormat(aDestCpuTexture->GetFormat())) &&
       (aDestCpuTexture->GetFormat() == SurfaceFormat::B8G8R8A8)) {
-    return ConvertSourceAndRetryReadback(aDestCpuTexture, aSrcTexture);
+    return ConvertSourceAndRetryReadback(aDestCpuTexture, aSrcTexture,
+                                         aArrayIndex, aColorSpace, aColorRange);
   }
 
   if ((IntSize(srcDesc.Width, srcDesc.Height) != aDestCpuTexture->GetSize()) ||
@@ -1297,6 +1149,8 @@ bool Factory::ReadbackTexture(DataSourceSurface* aDestCpuTexture,
     return false;
   }
 
+  MOZ_ASSERT(aArrayIndex == 0);
+
   bool ret =
       ReadbackTexture(mappedSurface.mData, mappedSurface.mStride, aSrcTexture);
   aDestCpuTexture->Unmap();
@@ -1368,7 +1222,7 @@ bool Factory::ReadbackTexture(uint8_t* aDestData, int32_t aDestStride,
   uint32_t width = srcDesc.Width;
   uint32_t height = srcDesc.Height;
   int bpp = BytesPerPixel(gfx::ToPixelFormat(srcDesc.Format));
-  for (int y = 0; y < height; y++) {
+  for (uint32_t y = 0; y < height; y++) {
     memcpy(aDestData + aDestStride * y,
            (unsigned char*)(srcMap.pData) + srcMap.RowPitch * y, width * bpp);
   }
