@@ -7,6 +7,9 @@
 #ifndef mozilla_layers_AsyncPanZoomController_h
 #define mozilla_layers_AsyncPanZoomController_h
 
+#include "Units.h"
+#include "apz/public/APZPublicUtils.h"
+#include "mozilla/layers/CompositorScrollUpdate.h"
 #include "mozilla/layers/GeckoContentController.h"
 #include "mozilla/layers/RepaintRequest.h"
 #include "mozilla/layers/SampleTime.h"
@@ -41,11 +44,12 @@ namespace mozilla {
 
 namespace ipc {
 
-class SharedMemoryBasic;
+class SharedMemory;
 
 }  // namespace ipc
 
 namespace wr {
+struct MinimapData;
 struct SampledScrollOffset;
 }  // namespace wr
 
@@ -177,10 +181,11 @@ struct PointerEventsConsumableFlags {
 class AsyncPanZoomController {
   NS_INLINE_DECL_THREADSAFE_REFCOUNTING(AsyncPanZoomController)
 
-  typedef mozilla::MonitorAutoLock MonitorAutoLock;
-  typedef mozilla::gfx::Matrix4x4 Matrix4x4;
-  typedef mozilla::layers::RepaintRequest::ScrollOffsetUpdateType
-      RepaintUpdateType;
+  using MonitorAutoLock = mozilla::MonitorAutoLock;
+  using Matrix4x4 = mozilla::gfx::Matrix4x4;
+  using RepaintUpdateType =
+      mozilla::layers::RepaintRequest::ScrollOffsetUpdateType;
+  using ScrollAnimationKind = apz::ScrollAnimationKind;
 
  public:
   enum GestureBehavior {
@@ -358,7 +363,7 @@ class AsyncPanZoomController {
   /**
    * See documentation on corresponding method in APZPublicUtils.h
    */
-  static gfx::IntSize GetDisplayportAlignmentMultiplier(
+  static gfx::Size GetDisplayportAlignmentMultiplier(
       const ScreenSize& aBaseSize);
 
   enum class ZoomInProgress {
@@ -366,6 +371,14 @@ class AsyncPanZoomController {
     Yes,
   };
 
+  /**
+   * Enlarges the displayport along both axes based on the velocity.
+   */
+  static CSSSize CalculateDisplayPortSize(
+      const CSSSize& aCompositionSize, const CSSPoint& aVelocity,
+      AsyncPanZoomController::ZoomInProgress aZoomInProgress,
+      const CSSToScreenScale2D& aDpPerCSS);
+
   /**
    * Recalculates the displayport. Ideally, this should paint an area bigger
    * than the composite-to dimensions so that when you scroll down, you don't
@@ -430,7 +443,7 @@ class AsyncPanZoomController {
    */
   bool HasTreeManager(const APZCTreeManager* aTreeManager) const;
 
-  void StartAnimation(AsyncPanZoomAnimation* aAnimation);
+  void StartAnimation(already_AddRefed<AsyncPanZoomAnimation> aAnimation);
 
   /**
    * Cancels any currently running animation.
@@ -445,12 +458,6 @@ class AsyncPanZoomController {
   void ClearOverscroll();
   void ClearPhysicalOverscroll();
 
-  /**
-   * Returns whether this APZC is for an element marked with the 'scrollgrab'
-   * attribute.
-   */
-  bool HasScrollgrab() const { return mScrollMetadata.GetHasScrollgrab(); }
-
   /**
    * Returns whether this APZC has scroll snap points.
    */
@@ -544,7 +551,8 @@ class AsyncPanZoomController {
 
   // Return the directions in which this APZC allows handoff (as governed by
   // overscroll-behavior).
-  ScrollDirections GetAllowedHandoffDirections() const;
+  ScrollDirections GetAllowedHandoffDirections(
+      HandoffConsumer aConsumer = HandoffConsumer::Scrolling) const;
 
   // Return the directions in which this APZC allows overscrolling.
   ScrollDirections GetOverscrollableDirections() const;
@@ -574,7 +582,8 @@ class AsyncPanZoomController {
   bool CanScrollDownwards() const;
 
   // Return true if there is room to scroll upwards.
-  bool CanOverscrollUpwards() const;
+  bool CanOverscrollUpwards(
+      HandoffConsumer aConsumer = HandoffConsumer::Scrolling) const;
 
   /**
    * Convert a point on the scrollbar from this APZC's ParentLayer coordinates
@@ -594,14 +603,24 @@ class AsyncPanZoomController {
   const FrameMetrics& Metrics() const;
   FrameMetrics& Metrics();
 
+  class AutoRecordCompositorScrollUpdate;
   /**
-   * Get the GeckoViewMetrics to be sent to Gecko for the current composite.
+   * Get the CompositorScrollUpdates to be sent to consumers for the current
+   * composite.
    */
-  GeckoViewMetrics GetGeckoViewMetrics() const;
+  std::vector<CompositorScrollUpdate> GetCompositorScrollUpdates();
 
-  // Helper function to compare root frame metrics and update them
-  // Returns true when the metrics have changed and were updated.
-  bool UpdateRootFrameMetricsIfChanged(GeckoViewMetrics& aMetrics);
+ private:
+  // Compositor scroll updates since the last time
+  // SampleCompositedAsyncTransform() was called.
+  // Access to this field should be protected by mRecursiveMutex.
+  std::vector<CompositorScrollUpdate> mUpdatesSinceLastSample;
+
+  CompositorScrollUpdate::Metrics GetCurrentMetricsForCompositorScrollUpdate(
+      const RecursiveMutexAutoLock& aProofOfApzcLock) const;
+
+ public:
+  wr::MinimapData GetMinimapData() const;
 
   // Returns the cached current frame time.
   SampleTime GetFrameTime() const;
@@ -796,6 +815,21 @@ class AsyncPanZoomController {
    */
   void ScrollByAndClamp(const CSSPoint& aOffset);
 
+  /**
+   * A variant of ScrollByAndClamp() that can scroll either the visual
+   * or the layout viewport.
+   */
+  void ScrollByAndClamp(ViewportType aViewportToScroll,
+                        const CSSPoint& aOffset);
+
+  /**
+   * Similar to ScrollByAndClamp() but scrolls to a specified destination.
+   * Can also be thought of as a variant of ClampAndSetVisualScrollOffset()
+   * which can set either the layout or viewport viewport offse.
+   */
+  void ScrollToAndClamp(ViewportType aViewportToScroll,
+                        const CSSPoint& aDestination);
+
   /**
    * Scales the viewport by an amount (note that it multiplies this scale in to
    * the current scale, it doesn't set it to |aScale|). Also considers a focus
@@ -846,7 +880,8 @@ class AsyncPanZoomController {
   /**
    * Gets the relevant point in the event, in external screen coordinates.
    */
-  ExternalPoint GetFirstExternalTouchPoint(const MultiTouchInput& aEvent);
+  static ExternalPoint GetFirstExternalTouchPoint(
+      const MultiTouchInput& aEvent);
 
   /**
    * Gets the amount by which this APZC is overscrolled along both axes.
@@ -858,6 +893,9 @@ class AsyncPanZoomController {
   // the test async properties.
   ParentLayerPoint GetOverscrollAmountInternal() const;
 
+  // Return whether this APZC blocks pull-to-refresh.
+  bool BlocksPullToRefreshForOverflowHidden() const;
+
  protected:
   /**
    * Returns SideBits where this APZC is overscrolled.
@@ -972,11 +1010,12 @@ class AsyncPanZoomController {
    */
   Maybe<LayoutDevicePoint> ConvertToGecko(const ScreenIntPoint& aPoint);
 
-  enum AxisLockMode {
+  enum class AxisLockMode {
     FREE,     /* No locking at all */
     STANDARD, /* Default axis locking mode that remains locked until pan ends */
     STICKY,   /* Allow lock to be broken, with hysteresis */
     DOMINANT_AXIS, /* Only allow movement on one axis */
+    BREAKABLE,     /* Allow lock to be broken until the pan ends */
   };
 
   static AxisLockMode GetAxisLockMode();
@@ -1007,9 +1046,11 @@ class AsyncPanZoomController {
   /* Access to the following two fields is protected by the mRefPtrMonitor,
      since they are accessed on the UI thread but can be cleared on the
      updater thread. */
-  RefPtr<GeckoContentController> mGeckoContentController;
-  RefPtr<GestureEventListener> mGestureEventListener;
-  mutable Monitor mRefPtrMonitor MOZ_UNANNOTATED;
+  RefPtr<GeckoContentController> mGeckoContentController
+      MOZ_GUARDED_BY(mRefPtrMonitor);
+  RefPtr<GestureEventListener> mGestureEventListener
+      MOZ_GUARDED_BY(mRefPtrMonitor);
+  mutable Monitor mRefPtrMonitor;
 
   // This is a raw pointer to avoid introducing a reference cycle between
   // AsyncPanZoomController and APZCTreeManager. Since these objects don't
@@ -1059,7 +1100,7 @@ class AsyncPanZoomController {
   // be held before calling the CanScroll function of |mX| and |mY|. These
   // coupled relationships bring us the burden of taking care of when the
   // monitor should be held, so they should be decoupled in the future.
-  mutable RecursiveMutex mRecursiveMutex MOZ_UNANNOTATED;
+  mutable RecursiveMutex mRecursiveMutex;
 
  private:
   // Metadata of the container layer corresponding to this APZC. This is
@@ -1103,6 +1144,12 @@ class AsyncPanZoomController {
   // controller thread.
   RecentEventsBuffer<PinchGestureInput> mPinchEventBuffer;
 
+  // Stores the touch events that occured within a given timeframe. Ussed to
+  // determine the direction of a touch scroll, which determines which axis
+  // should be locked if STICKY axis locking is used. Should only by accessed
+  // on the controller thread.
+  RecentEventsBuffer<MultiTouchInput> mTouchScrollEventBuffer;
+
   // Most up-to-date constraints on zooming. These should always be reasonable
   // values; for example, allowing a min zoom of 0.0 can cause very bad things
   // to happen. Hold mRecursiveMutex when accessing this.
@@ -1123,17 +1170,17 @@ class AsyncPanZoomController {
   // notification.
   CSSToParentLayerScale mLastNotifiedZoom;
 
+  // Accessing mAnimation needs to be protected by mRecursiveMutex
   RefPtr<AsyncPanZoomAnimation> mAnimation;
 
   UniquePtr<OverscrollEffectBase> mOverscrollEffect;
 
-  // Zoom animation id, used for zooming in WebRender. This should only be
-  // set on the APZC instance for the root content document (i.e. the one we
-  // support zooming on), and is only used if WebRender is enabled. The
-  // animation id itself refers to the transform animation id that was set on
-  // the stacking context in the WR display list. By changing the transform
-  // associated with this id, we can adjust the scaling that WebRender applies,
-  // thereby controlling the zoom.
+  // Zoom animation id, used for zooming. This should only be set on the APZC
+  // instance for the root content document (i.e. the one we support zooming
+  // on). The animation id itself refers to the transform animation id that was
+  // set on the stacking context in the WR display list. By changing the
+  // transform associated with this id, we can adjust the scaling that WebRender
+  // applies, thereby controlling the zoom.
   Maybe<uint64_t> mZoomAnimationId;
 
   // Position on screen where user first put their finger down.
@@ -1274,6 +1321,22 @@ class AsyncPanZoomController {
     return mScrollMetadata.GetMetrics().GetScrollableRect();
   }
 
+  CSSToParentLayerScale GetZoom() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return Metrics().GetZoom();
+  }
+
+  CSSRect GetVisualViewport() const {
+    MOZ_ASSERT(IsRootContent());
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return Metrics().GetVisualViewport();
+  }
+
+  CSSPoint GetLayoutScrollOffset() const {
+    RecursiveMutexAutoLock lock(mRecursiveMutex);
+    return Metrics().GetLayoutScrollOffset();
+  }
+
   // Returns the delta for the given InputData.
   ParentLayerPoint GetDeltaForEvent(const InputData& aEvent) const;
 
@@ -1283,6 +1346,8 @@ class AsyncPanZoomController {
    */
   CSSRect GetCurrentScrollRangeInCssPixels() const;
 
+  bool AllowOneTouchPinch() const;
+
  private:
   /**
    * Advances to the next sample, if there is one, the list of sampled states
@@ -1293,6 +1358,13 @@ class AsyncPanZoomController {
    */
   void AdvanceToNextSample();
 
+  /**
+   * Returns whether we have changes to the scroll offsets which need to be
+   * sampled in the next couple of frames (it depends on how many offsets we
+   * have, currently it's two).
+   */
+  bool HavePendingFrameDelayedOffset() const;
+
   /**
    * Samples the composited async transform, storing the result into
    * mSampledState. This will make the result of
@@ -1343,8 +1415,12 @@ class AsyncPanZoomController {
   std::tuple<ParentLayerPoint, ScreenPoint> GetDisplacementsForPanGesture(
       const PanGestureInput& aEvent);
 
+  CSSPoint ToCSSPixels(ParentLayerPoint value) const;
+  CSSCoord ToCSSPixels(ParentLayerCoord value) const;
+
  private:
   friend class AutoApplyAsyncTestAttributes;
+  friend class AutoDynamicToolbarHider;
 
   bool SuppressAsyncScrollOffset() const;
 
@@ -1387,15 +1463,11 @@ class AsyncPanZoomController {
     ANIMATING_ZOOM,       /* animated zoom to a new rect */
     OVERSCROLL_ANIMATION, /* Spring-based animation used to relieve overscroll
                              once the finger is lifted. */
-    SMOOTH_SCROLL,        /* Smooth scrolling to destination, with physics
-                             controlled by prefs specific to the scroll origin. */
-    SMOOTHMSD_SCROLL,     /* SmoothMSD scrolling to destination. Used by
-                             CSSOM-View smooth scroll-behavior */
-    WHEEL_SCROLL,    /* Smooth scrolling to a destination for a wheel event. */
-    KEYBOARD_SCROLL, /* Smooth scrolling to a destination for a keyboard event.
-                      */
-    AUTOSCROLL,      /* Autoscroll animation. */
-    SCROLLBAR_DRAG   /* Async scrollbar drag. */
+    SMOOTH_SCROLL,        /* Smooth scrolling to destination. May be for a wheel
+                             event, keyboard event, scroll-behavior or scroll
+                             snapping. */
+    AUTOSCROLL,           /* Autoscroll animation. */
+    SCROLLBAR_DRAG        /* Async scrollbar drag. */
   };
   // This is in theory protected by |mRecursiveMutex|; that is, it should be
   // held whenever this is updated. In practice though... see bug 897017.
@@ -1421,6 +1493,18 @@ class AsyncPanZoomController {
    */
   void SetDelayedTransformEnd(bool aDelayedTransformEnd);
 
+  /**
+   * Check whether there is an ongoing smooth scroll animation of
+   * the specified kind.
+   */
+  bool InScrollAnimation(ScrollAnimationKind aKind) const;
+
+  /**
+   * Check whether there is an ongoing smooth scroll animation triggered by
+   * script.
+   */
+  bool InScrollAnimationTriggeredByScript() const;
+
   /**
    * Returns whether the specified PanZoomState does not need to be reset when
    * a scroll offset update is processed.
@@ -1441,6 +1525,7 @@ class AsyncPanZoomController {
 
  private:
   friend class StateChangeNotificationBlocker;
+  friend class ThreadSafeStateChangeNotificationBlocker;
   /**
    * A counter of how many StateChangeNotificationBlockers are active.
    * A non-zero count will prevent state change notifications from
@@ -1472,6 +1557,13 @@ class AsyncPanZoomController {
    */
   void DispatchStateChangeNotification(PanZoomState aOldState,
                                        PanZoomState aNewState);
+
+  /**
+   * Send a TransformBegin notification followed by a TransformEnd
+   * notification.
+   */
+  void SendTransformBeginAndEnd();
+
   /**
    * Internal helpers for checking general state of this apzc.
    */
@@ -1499,7 +1591,7 @@ class AsyncPanZoomController {
    * govern dynamic toolbar and pull-to-refresh behaviour).
    */
   PointerEventsConsumableFlags ArePointerEventsConsumable(
-      TouchBlockState* aBlock, const MultiTouchInput& aInput);
+      const TouchBlockState* aBlock, const MultiTouchInput& aInput) const;
 
   /**
    * Clear internal state relating to touch input handling.
@@ -1557,9 +1649,8 @@ class AsyncPanZoomController {
   friend class AndroidFlingPhysics;
   friend class DesktopFlingPhysics;
   friend class OverscrollAnimation;
-  friend class SmoothMsdScrollAnimation;
   friend class GenericScrollAnimation;
-  friend class WheelScrollAnimation;
+  friend class SmoothScrollAnimation;
   friend class ZoomAnimation;
 
   friend class GenericOverscrollEffect;
@@ -1586,23 +1677,21 @@ class AsyncPanZoomController {
       const RefPtr<const OverscrollHandoffChain>& aOverscrollHandoffChain,
       const RefPtr<const AsyncPanZoomController>& aScrolledApzc);
 
-  void HandleSmoothScrollOverscroll(const ParentLayerPoint& aVelocity,
-                                    SideBits aOverscrollSideBits);
-
   // Start an overscroll animation with the given initial velocity.
   void StartOverscrollAnimation(const ParentLayerPoint& aVelocity,
                                 SideBits aOverscrollSideBits);
 
-  // Start a smooth-scrolling animation to the given destination, with physics
-  // based on the prefs for the indicated origin.
-  void SmoothScrollTo(CSSSnapTarget&& aDestination,
+  // Start a smooth-scrolling animation to the given destination.
+  // |aAnimationKind| must be |Smooth| or |SmoothMsd|
+  // |aOrigin| is only used with |Smooth| (for |SmoothMsd|,
+  // |ScrollOrigin::NotSpecified| may be passed).
+  void SmoothScrollTo(CSSSnapDestination&& aDestination,
                       ScrollTriggeredByScript aTriggeredByScript,
-                      const ScrollOrigin& aOrigin);
+                      ScrollAnimationKind aAnimationKind,
+                      ViewportType aViewportToScroll, ScrollOrigin aOrigin,
+                      TimeStamp aStartTime);
 
-  // Start a smooth-scrolling animation to the given destination, with MSD
-  // physics that is suited for scroll-snapping.
-  void SmoothMsdScrollTo(CSSSnapTarget&& aDestination,
-                         ScrollTriggeredByScript aTriggeredByScript);
+  ParentLayerPoint ConvertDestinationToDelta(CSSPoint& aDestination) const;
 
   // Returns whether overscroll is allowed during an event.
   bool AllowScrollHandoffInCurrentBlock() const;
@@ -1882,7 +1971,7 @@ class AsyncPanZoomController {
                                uint32_t aMagnitude);
 
   // Mutex protecting mCheckerboardEvent
-  Mutex mCheckerboardEventLock MOZ_UNANNOTATED;
+  Mutex mCheckerboardEventLock;
   // This is created when this APZC instance is first included as part of a
   // composite. If a checkerboard event takes place, this is destroyed at the
   // end of the event, and a new one is created on the next composite.
@@ -1904,17 +1993,17 @@ class AsyncPanZoomController {
   // |aUnit| affects the snapping behaviour (see ScrollSnapUtils::
   // GetSnapPointForDestination).
   // Returns true iff. a target snap point was found.
-  Maybe<CSSSnapTarget> MaybeAdjustDeltaForScrollSnapping(
+  Maybe<CSSSnapDestination> MaybeAdjustDeltaForScrollSnapping(
       ScrollUnit aUnit, ScrollSnapFlags aSnapFlags, ParentLayerPoint& aDelta,
       CSSPoint& aStartPosition);
 
   // A wrapper function of MaybeAdjustDeltaForScrollSnapping for
   // ScrollWheelInput.
-  Maybe<CSSSnapTarget> MaybeAdjustDeltaForScrollSnappingOnWheelInput(
+  Maybe<CSSSnapDestination> MaybeAdjustDeltaForScrollSnappingOnWheelInput(
       const ScrollWheelInput& aEvent, ParentLayerPoint& aDelta,
       CSSPoint& aStartPosition);
 
-  Maybe<CSSSnapTarget> MaybeAdjustDestinationForScrollSnapping(
+  Maybe<CSSSnapDestination> MaybeAdjustDestinationForScrollSnapping(
       const KeyboardInput& aEvent, CSSPoint& aDestination,
       ScrollSnapFlags aSnapFlags);
 
@@ -1933,9 +2022,9 @@ class AsyncPanZoomController {
   // |aUnit| affects the snapping behaviour (see ScrollSnapUtils::
   // GetSnapPointForDestination). It should generally be determined by the
   // type of event that's triggering the scroll.
-  Maybe<CSSSnapTarget> FindSnapPointNear(const CSSPoint& aDestination,
-                                         ScrollUnit aUnit,
-                                         ScrollSnapFlags aSnapFlags);
+  Maybe<CSSSnapDestination> FindSnapPointNear(const CSSPoint& aDestination,
+                                              ScrollUnit aUnit,
+                                              ScrollSnapFlags aSnapFlags);
 
   // If |aOriginalEvent| crosses the touch-start tolerance threshold, split it
   // into two events: one that just reaches the threshold, and the remainder.
@@ -1954,6 +2043,12 @@ class AsyncPanZoomController {
       const MultiTouchInput& aOriginalEvent, ScreenCoord aPanThreshold,
       float aVectorLength, ExternalPoint& aExtPoint);
 
+  // Fill out the overscroll gutter with the new expanded contents the
+  // overscroll amount is inside the new scroll range. Returns the scroll
+  // position change delta if filling out happened, CSSPoint() otherwise.
+  CSSPoint MaybeFillOutOverscrollGutter(
+      const RecursiveMutexAutoLock& aProofOfLock);
+
   friend std::ostream& operator<<(
       std::ostream& aOut, const AsyncPanZoomController::PanZoomState& aState);
 };
