@@ -14,19 +14,17 @@ class APZScrollHandoffTester : public APZCTreeManagerTester {
   TestAsyncPanZoomController* rootApzc;
 
   void CreateScrollHandoffLayerTree1() {
-    const char* layerTreeSyntax = "c(t)";
-    nsIntRegion layerVisibleRegion[] = {nsIntRegion(IntRect(0, 0, 100, 100)),
-                                        nsIntRegion(IntRect(0, 50, 100, 50))};
-    root = CreateLayerTree(layerTreeSyntax, layerVisibleRegion, nullptr, lm,
-                           layers);
+    const char* treeShape = "x(x)";
+    LayerIntRect layerVisibleRect[] = {LayerIntRect(0, 0, 100, 100),
+                                       LayerIntRect(0, 50, 100, 50)};
+    CreateScrollData(treeShape, layerVisibleRect);
     SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
                               CSSRect(0, 0, 200, 200));
     SetScrollableFrameMetrics(layers[1],
                               ScrollableLayerGuid::START_SCROLL_ID + 1,
                               CSSRect(0, 0, 100, 100));
     SetScrollHandoff(layers[1], root);
-    registration = MakeUnique<ScopedLayerTreeRegistration>(manager, LayersId{0},
-                                                           root, mcc);
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
     UpdateHitTestingTree();
     rootApzc = ApzcOf(root);
     rootApzc->GetFrameMetrics().SetIsRootContent(
@@ -34,12 +32,11 @@ class APZScrollHandoffTester : public APZCTreeManagerTester {
   }
 
   void CreateScrollHandoffLayerTree2() {
-    const char* layerTreeSyntax = "c(c(t))";
-    nsIntRegion layerVisibleRegion[] = {nsIntRegion(IntRect(0, 0, 100, 100)),
-                                        nsIntRegion(IntRect(0, 0, 100, 100)),
-                                        nsIntRegion(IntRect(0, 50, 100, 50))};
-    root = CreateLayerTree(layerTreeSyntax, layerVisibleRegion, nullptr, lm,
-                           layers);
+    const char* treeShape = "x(x(x))";
+    LayerIntRect layerVisibleRect[] = {LayerIntRect(0, 0, 100, 100),
+                                       LayerIntRect(0, 0, 100, 100),
+                                       LayerIntRect(0, 50, 100, 50)};
+    CreateScrollData(treeShape, layerVisibleRect);
     SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
                               CSSRect(0, 0, 200, 200));
     SetScrollableFrameMetrics(layers[1],
@@ -58,16 +55,15 @@ class APZScrollHandoffTester : public APZCTreeManagerTester {
   }
 
   void CreateScrollHandoffLayerTree3() {
-    const char* layerTreeSyntax = "c(c(t)c(t))";
-    nsIntRegion layerVisibleRegion[] = {
-        nsIntRegion(IntRect(0, 0, 100, 100)),  // root
-        nsIntRegion(IntRect(0, 0, 100, 50)),   // scrolling parent 1
-        nsIntRegion(IntRect(0, 0, 100, 50)),   // scrolling child 1
-        nsIntRegion(IntRect(0, 50, 100, 50)),  // scrolling parent 2
-        nsIntRegion(IntRect(0, 50, 100, 50))   // scrolling child 2
+    const char* treeShape = "x(x(x)x(x))";
+    LayerIntRect layerVisibleRect[] = {
+        LayerIntRect(0, 0, 100, 100),  // root
+        LayerIntRect(0, 0, 100, 50),   // scrolling parent 1
+        LayerIntRect(0, 0, 100, 50),   // scrolling child 1
+        LayerIntRect(0, 50, 100, 50),  // scrolling parent 2
+        LayerIntRect(0, 50, 100, 50)   // scrolling child 2
     };
-    root = CreateLayerTree(layerTreeSyntax, layerVisibleRegion, nullptr, lm,
-                           layers);
+    CreateScrollData(treeShape, layerVisibleRect);
     SetScrollableFrameMetrics(layers[0], ScrollableLayerGuid::START_SCROLL_ID,
                               CSSRect(0, 0, 100, 100));
     SetScrollableFrameMetrics(layers[1],
@@ -86,90 +82,64 @@ class APZScrollHandoffTester : public APZCTreeManagerTester {
     SetScrollHandoff(layers[3], layers[0]);
     SetScrollHandoff(layers[2], layers[1]);
     SetScrollHandoff(layers[4], layers[3]);
-    registration = MakeUnique<ScopedLayerTreeRegistration>(manager, LayersId{0},
-                                                           root, mcc);
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
     UpdateHitTestingTree();
   }
 
   // Creates a layer tree with a parent layer that is only scrollable
   // horizontally, and a child layer that is only scrollable vertically.
   void CreateScrollHandoffLayerTree4() {
-    const char* layerTreeSyntax = "c(t)";
-    nsIntRegion layerVisibleRegion[] = {nsIntRegion(IntRect(0, 0, 100, 100)),
-                                        nsIntRegion(IntRect(0, 0, 100, 100))};
-    root = CreateLayerTree(layerTreeSyntax, layerVisibleRegion, nullptr, lm,
-                           layers);
+    const char* treeShape = "x(x)";
+    LayerIntRect layerVisibleRect[] = {LayerIntRect(0, 0, 100, 100),
+                                       LayerIntRect(0, 0, 100, 100)};
+    CreateScrollData(treeShape, layerVisibleRect);
     SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
                               CSSRect(0, 0, 200, 100));
     SetScrollableFrameMetrics(layers[1],
                               ScrollableLayerGuid::START_SCROLL_ID + 1,
                               CSSRect(0, 0, 100, 200));
     SetScrollHandoff(layers[1], root);
-    registration = MakeUnique<ScopedLayerTreeRegistration>(manager, LayersId{0},
-                                                           root, mcc);
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
     UpdateHitTestingTree();
     rootApzc = ApzcOf(root);
   }
 
-  void CreateScrollgrabLayerTree(bool makeParentScrollable = true) {
-    const char* layerTreeSyntax = "c(t)";
-    nsIntRegion layerVisibleRegion[] = {
-        nsIntRegion(IntRect(0, 0, 100, 100)),  // scroll-grabbing parent
-        nsIntRegion(IntRect(0, 20, 100, 80))   // child
+  // Creates a layer tree with a parent layer that is not scrollable, and a
+  // child layer that is only scrollable vertically.
+  void CreateScrollHandoffLayerTree5() {
+    const char* treeShape = "x(x)";
+    LayerIntRect layerVisibleRect[] = {
+        LayerIntRect(0, 0, 100, 100),  // scrolling parent
+        LayerIntRect(0, 50, 100, 50)   // scrolling child
     };
-    root = CreateLayerTree(layerTreeSyntax, layerVisibleRegion, nullptr, lm,
-                           layers);
-    float parentHeight = makeParentScrollable ? 120 : 100;
+    CreateScrollData(treeShape, layerVisibleRect);
     SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
-                              CSSRect(0, 0, 100, parentHeight));
+                              CSSRect(0, 0, 100, 100));
     SetScrollableFrameMetrics(layers[1],
                               ScrollableLayerGuid::START_SCROLL_ID + 1,
-                              CSSRect(0, 0, 100, 800));
+                              CSSRect(0, 0, 100, 200));
     SetScrollHandoff(layers[1], root);
-    registration = MakeUnique<ScopedLayerTreeRegistration>(manager, LayersId{0},
-                                                           root, mcc);
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
     UpdateHitTestingTree();
     rootApzc = ApzcOf(root);
-    rootApzc->GetScrollMetadata().SetHasScrollgrab(true);
   }
 
-  void TestFlingAcceleration() {
-    // Jack up the fling acceleration multiplier so we can easily determine
-    // whether acceleration occured.
-    const float kAcceleration = 100.0f;
-    SCOPED_GFX_PREF_FLOAT("apz.fling_accel_base_mult", kAcceleration);
-    SCOPED_GFX_PREF_FLOAT("apz.fling_accel_min_fling_velocity", 0.0);
-    SCOPED_GFX_PREF_FLOAT("apz.fling_accel_min_pan_velocity", 0.0);
-
-    RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
-
-    // Pan once, enough to fully scroll the scrollgrab parent and then scroll
-    // and fling the child.
-    Pan(manager, 70, 40);
-
-    // Give the fling animation a chance to start.
-    SampleAnimationsOnce();
-
-    float childVelocityAfterFling1 = childApzc->GetVelocityVector().y;
-
-    // Pan again.
-    Pan(manager, 70, 40);
-
-    // Give the fling animation a chance to start.
-    // This time it should be accelerated.
-    SampleAnimationsOnce();
-
-    float childVelocityAfterFling2 = childApzc->GetVelocityVector().y;
-
-    // We should have accelerated once.
-    // The division by 2 is to account for friction.
-    EXPECT_GT(childVelocityAfterFling2,
-              childVelocityAfterFling1 * kAcceleration / 2);
-
-    // We should not have accelerated twice.
-    // The division by 4 is to account for friction.
-    EXPECT_LE(childVelocityAfterFling2,
-              childVelocityAfterFling1 * kAcceleration * kAcceleration / 4);
+  // Creates a layer tree with a parent layer that is only scrollable
+  // vertically, and a child layer that is only scrollable horizontally.
+  void CreateScrollHandoffLayerTree6() {
+    const char* treeShape = "x(x)";
+    LayerIntRect layerVisibleRect[] = {LayerIntRect(0, 0, 100, 100),
+                                       LayerIntRect(0, 0, 100, 1000)};
+    CreateScrollData(treeShape, layerVisibleRect);
+    SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
+                              CSSRect(0, 0, 100, 1000));
+    SetScrollableFrameMetrics(layers[1],
+                              ScrollableLayerGuid::START_SCROLL_ID + 1,
+                              CSSRect(0, 0, 200, 1000));
+    SetScrollHandoff(layers[1], root);
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
+    UpdateHitTestingTree();
+    rootApzc = ApzcOf(root);
   }
 
   void TestCrossApzcAxisLock() {
@@ -182,9 +152,99 @@ class APZScrollHandoffTester : public APZCTreeManagerTester {
         PanOptions::KeepFingerDown | PanOptions::ExactCoordinates);
 
     childApzc->AssertAxisLocked(ScrollDirection::eVertical);
+    childApzc->AssertStateIsPanningLockedY();
   }
 };
 
+class APZScrollHandoffTesterMock : public APZScrollHandoffTester {
+ public:
+  APZScrollHandoffTesterMock() { CreateMockHitTester(); }
+};
+
+class APZCNestedFlingScrollHandoffTester : public APZCTreeManagerTester {
+ protected:
+  void SetUp() {
+    APZCTreeManagerTester::SetUp();
+    const char* treeShape = "x(x)";
+    LayerIntRect layerVisibleRect[] = {
+        LayerIntRect(0, 0, 800, 1000),
+        LayerIntRect(0, 0, 100, 100),
+    };
+
+    CreateScrollData(treeShape, layerVisibleRect);
+
+    SetScrollableFrameMetrics(root, ScrollableLayerGuid::START_SCROLL_ID,
+                              CSSRect(0, 0, 800, 50000));
+    SetScrollableFrameMetrics(layers[1],
+                              ScrollableLayerGuid::START_SCROLL_ID + 1,
+                              CSSRect(0, 0, 800, 100));
+
+    SetScrollHandoff(layers[1], root);
+
+    // Scroll somewhere into the middle of the scroll range, so that we have
+    // lots of space to scroll in both directions.
+    ModifyFrameMetrics(root, [](ScrollMetadata& aSm, FrameMetrics& aMetrics) {
+      aMetrics.SetVisualScrollUpdateType(
+          FrameMetrics::ScrollOffsetUpdateType::eMainThread);
+      aMetrics.SetVisualDestination(CSSPoint(0, 25000));
+    });
+
+    registration = MakeUnique<ScopedLayerTreeRegistration>(LayersId{0}, mcc);
+    UpdateHitTestingTree();
+
+    subframeApzc = ApzcOf(layers[1]);
+  }
+
+  void ExecuteDirectionChangingPanGesture(
+      const ScreenIntPoint& aStartPoint,
+      std::initializer_list<int32_t> aXDeltas,
+      std::initializer_list<int32_t> aYDeltas) {
+    APZEventResult result = TouchDown(subframeApzc, aStartPoint, mcc->Time());
+
+    // Allowed touch behaviours must be set after sending touch-start.
+    if (result.GetStatus() != nsEventStatus_eConsumeNoDefault) {
+      SetDefaultAllowedTouchBehavior(subframeApzc, result.mInputBlockId);
+    }
+
+    const TimeDuration kTouchTimeDelta100Hz =
+        TimeDuration::FromMilliseconds(10);
+
+    ScreenIntPoint currentLocation = aStartPoint;
+    for (int32_t delta : aXDeltas) {
+      mcc->AdvanceBy(kTouchTimeDelta100Hz);
+      if (delta != 0) {
+        currentLocation.x += delta;
+        (void)TouchMove(subframeApzc, currentLocation, mcc->Time());
+      }
+    }
+
+    ExecuteWait(TimeDuration::FromMilliseconds(255));
+
+    for (int32_t delta : aYDeltas) {
+      mcc->AdvanceBy(kTouchTimeDelta100Hz);
+      if (delta != 0) {
+        currentLocation.y += delta;
+        (void)TouchMove(subframeApzc, currentLocation, mcc->Time());
+      }
+    }
+
+    (void)TouchUp(subframeApzc, currentLocation, mcc->Time());
+  }
+
+  void ExecuteWait(const TimeDuration& aDuration) {
+    TimeDuration remaining = aDuration;
+    const TimeDuration TIME_BETWEEN_FRAMES =
+        TimeDuration::FromSeconds(1) / int64_t(60);
+    while (remaining.ToMilliseconds() > 0) {
+      mcc->AdvanceBy(TIME_BETWEEN_FRAMES);
+      subframeApzc->AdvanceAnimations(mcc->GetSampleTime());
+      remaining -= TIME_BETWEEN_FRAMES;
+    }
+  }
+
+  RefPtr<TestAsyncPanZoomController> subframeApzc;
+  UniquePtr<ScopedLayerTreeRegistration> registration;
+};
 #ifndef MOZ_WIDGET_ANDROID  // Currently fails on Android
 // Here we test that if the processing of a touch block is deferred while we
 // wait for content to send a prevent-default message, overscroll is still
@@ -240,7 +300,7 @@ TEST_F(APZScrollHandoffTester, LayerStructureChangesWhileEventsArePending) {
   // Modify the APZC tree to insert a new APZC 'middle' into the handoff chain
   // between the child and the root.
   CreateScrollHandoffLayerTree2();
-  RefPtr<Layer> middle = layers[1];
+  WebRenderLayerScrollData* middle = layers[1];
   childApzc->SetWaitForMainThread();
   TestAsyncPanZoomController* middleApzc = ApzcOf(middle);
 
@@ -273,17 +333,17 @@ TEST_F(APZScrollHandoffTester, LayerStructureChangesWhileEventsArePending) {
 #ifndef MOZ_WIDGET_ANDROID  // Currently fails on Android
 // Test that putting a second finger down on an APZC while a down-chain APZC
 // is overscrolled doesn't result in being stuck in overscroll.
-TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1073250) {
+TEST_F(APZScrollHandoffTesterMock, StuckInOverscroll_Bug1073250) {
   // Enable overscrolling.
   SCOPED_GFX_PREF_BOOL("apz.overscroll.enabled", true);
   SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
 
   CreateScrollHandoffLayerTree1();
 
   TestAsyncPanZoomController* child = ApzcOf(layers[1]);
 
   // Pan, causing the parent APZC to overscroll.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID);
   Pan(manager, 10, 40, PanOptions::KeepFingerDown);
   EXPECT_FALSE(child->IsOverscrolled());
   EXPECT_TRUE(rootApzc->IsOverscrolled());
@@ -318,17 +378,17 @@ TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1073250) {
 // This is almost exactly like StuckInOverscroll_Bug1073250, except the
 // APZC receiving the input events for the first touch block is the child
 // (and thus not the same APZC that overscrolls, which is the parent).
-TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1231228) {
+TEST_F(APZScrollHandoffTesterMock, StuckInOverscroll_Bug1231228) {
   // Enable overscrolling.
   SCOPED_GFX_PREF_BOOL("apz.overscroll.enabled", true);
   SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
 
   CreateScrollHandoffLayerTree1();
 
   TestAsyncPanZoomController* child = ApzcOf(layers[1]);
 
   // Pan, causing the parent APZC to overscroll.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
   Pan(manager, 60, 90, PanOptions::KeepFingerDown);
   EXPECT_FALSE(child->IsOverscrolled());
   EXPECT_TRUE(rootApzc->IsOverscrolled());
@@ -395,16 +455,16 @@ TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1240202a) {
 #endif
 
 #ifndef MOZ_WIDGET_ANDROID  // Currently fails on Android
-TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1240202b) {
+TEST_F(APZScrollHandoffTesterMock, StuckInOverscroll_Bug1240202b) {
   // Enable overscrolling.
   SCOPED_GFX_PREF_BOOL("apz.overscroll.enabled", true);
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
 
   CreateScrollHandoffLayerTree1();
 
   TestAsyncPanZoomController* child = ApzcOf(layers[1]);
 
   // Pan, causing the parent APZC to overscroll.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
   Pan(manager, 60, 90, PanOptions::KeepFingerDown);
   EXPECT_FALSE(child->IsOverscrolled());
   EXPECT_TRUE(rootApzc->IsOverscrolled());
@@ -415,6 +475,7 @@ TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1240202b) {
 
   // Put the finger down again, interrupting the animation
   // and entering the TOUCHING state.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
   TouchDown(manager, ScreenIntPoint(10, 90), mcc->Time());
 
   // Put a second finger down. Since we're in the TOUCHING state,
@@ -422,13 +483,15 @@ TEST_F(APZScrollHandoffTester, StuckInOverscroll_Bug1240202b) {
   // will not ignore the second finger, instead entering the
   // PINCHING state.
   MultiTouchInput secondFingerDown(MultiTouchInput::MULTITOUCH_START, 0,
-                                   TimeStamp(), 0);
+                                   mcc->Time(), 0);
   // Use the same touch identifier for the first touch (0) as TouchDown(). (A
   // bit hacky.)
   secondFingerDown.mTouches.AppendElement(
       SingleTouchData(0, ScreenIntPoint(10, 90), ScreenSize(0, 0), 0, 0));
   secondFingerDown.mTouches.AppendElement(
       SingleTouchData(1, ScreenIntPoint(10, 80), ScreenSize(0, 0), 0, 0));
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
   manager->ReceiveInputEvent(secondFingerDown);
 
   // Release the fingers.
@@ -464,32 +527,32 @@ TEST_F(APZScrollHandoffTester, OpposingConstrainedAxes_Bug1201098) {
 }
 #endif
 
-// Test that flinging in a direction where one component of the fling goes into
-// overscroll but the other doesn't, results in just the one component being
-// handed off to the parent, while the original APZC continues flinging in the
-// other direction.
-TEST_F(APZScrollHandoffTester, PartialFlingHandoff) {
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
+// Test the behaviour when flinging diagonally and there is room to scroll in
+// one direction but not the other. In the direction where there is room to
+// scroll, the fling should continue. In the direction where there is no room
+// to scroll, the fling should stop without being handed off to the parent.
+TEST_F(APZScrollHandoffTesterMock, DiagonalFlingNoHandoff) {
   SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
 
   CreateScrollHandoffLayerTree1();
 
   // Fling up and to the left. The child APZC has room to scroll up, but not
-  // to the left, so the horizontal component of the fling should be handed
-  // off to the parent APZC.
+  // to the left.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
   Pan(manager, ScreenIntPoint(90, 90), ScreenIntPoint(55, 55));
 
-  RefPtr<TestAsyncPanZoomController> parent = ApzcOf(root);
+  RefPtr<TestAsyncPanZoomController> parent = ApzcOf(layers[0]);
   RefPtr<TestAsyncPanZoomController> child = ApzcOf(layers[1]);
 
-  // Advance the child's fling animation once to give the partial handoff
+  // Advance the child's fling animation once to give any potential handoff
   // a chance to occur.
   mcc->AdvanceByMillis(10);
   child->AdvanceAnimations(mcc->GetSampleTime());
 
-  // Assert that partial handoff has occurred.
+  // Assert that the child is still flinging but the parent is not (no handoff
+  // occurred).
   child->AssertStateIsFling();
-  parent->AssertStateIsFling();
+  parent->AssertStateIsReset();
 }
 
 // Here we test that if two flings are happening simultaneously, overscroll
@@ -527,70 +590,12 @@ TEST_F(APZScrollHandoffTester, SimultaneousFlings) {
   parent2->AssertStateIsFling();
 }
 
-#ifndef MOZ_WIDGET_ANDROID  // Currently fails on Android
-TEST_F(APZScrollHandoffTester, Scrollgrab) {
-  SCOPED_GFX_PREF_BOOL("apz.allow_immediate_handoff", true);
-
-  // Set up the layer tree
-  CreateScrollgrabLayerTree();
-
-  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
-
-  // Pan on the child, enough to fully scroll the scrollgrab parent (20 px)
-  // and leave some more (another 15 px) for the child.
-  Pan(childApzc, 80, 45);
-
-  // Check that the parent and child have scrolled as much as we expect.
-  EXPECT_EQ(20, rootApzc->GetFrameMetrics().GetVisualScrollOffset().y);
-  EXPECT_EQ(15, childApzc->GetFrameMetrics().GetVisualScrollOffset().y);
-}
-#endif
-
-TEST_F(APZScrollHandoffTester, ScrollgrabFling) {
-  SCOPED_GFX_PREF_BOOL("apz.allow_immediate_handoff", true);
-  SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
-
-  // Set up the layer tree
-  CreateScrollgrabLayerTree();
-
-  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
-
-  // Pan on the child, not enough to fully scroll the scrollgrab parent.
-  Pan(childApzc, 80, 70);
-
-  // Check that it is the scrollgrab parent that's in a fling, not the child.
-  rootApzc->AssertStateIsFling();
-  childApzc->AssertStateIsReset();
-}
-
-TEST_F(APZScrollHandoffTester, ScrollgrabFlingAcceleration1) {
-  SCOPED_GFX_PREF_BOOL("apz.allow_immediate_handoff", true);
-  SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
-  CreateScrollgrabLayerTree(true /* make parent scrollable */);
-
-  // Note: Usually, fling acceleration does not work across handoff, because our
-  // fling acceleration code does not propagate the "fling cancel velocity"
-  // across handoff. However, this test sets apz.fling_min_velocity_threshold to
-  // zero, so the "fling cancel velocity" is allowed to be zero, and fling
-  // acceleration succeeds, almost by accident.
-  TestFlingAcceleration();
-}
-
-TEST_F(APZScrollHandoffTester, ScrollgrabFlingAcceleration2) {
-  SCOPED_GFX_PREF_BOOL("apz.allow_immediate_handoff", true);
-  SCOPED_GFX_PREF_FLOAT("apz.fling_min_velocity_threshold", 0.0f);
-  SCOPED_GFX_VAR(UseWebRender, bool, false);
-  CreateScrollgrabLayerTree(false /* do not make parent scrollable */);
-  TestFlingAcceleration();
-}
-
 TEST_F(APZScrollHandoffTester, ImmediateHandoffDisallowed_Pan) {
   SCOPED_GFX_PREF_BOOL("apz.allow_immediate_handoff", false);
 
   CreateScrollHandoffLayerTree1();
 
-  RefPtr<TestAsyncPanZoomController> parentApzc = ApzcOf(root);
+  RefPtr<TestAsyncPanZoomController> parentApzc = ApzcOf(layers[0]);
   RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
 
   // Pan on the child, enough to scroll it to its end and have scroll
@@ -616,12 +621,12 @@ TEST_F(APZScrollHandoffTester, ImmediateHandoffDisallowed_Fling) {
 
   CreateScrollHandoffLayerTree1();
 
-  RefPtr<TestAsyncPanZoomController> parentApzc = ApzcOf(root);
+  RefPtr<TestAsyncPanZoomController> parentApzc = ApzcOf(layers[0]);
   RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
 
   // Pan on the child, enough to get very close to the end, so that the
   // subsequent fling reaches the end and has leftover velocity to hand off.
-  Pan(childApzc, 60, 12);
+  Pan(childApzc, 60, 2);
 
   // Allow the fling to run its course.
   childApzc->AdvanceAnimationsUntilEnd();
@@ -637,7 +642,7 @@ TEST_F(APZScrollHandoffTester, ImmediateHandoffDisallowed_Fling) {
 
   // Pan again on the child. This time, since the child was scrolled to
   // its end when the gesture began, we expect the scroll to be handed off.
-  Pan(childApzc, 60, 50);
+  Pan(childApzc, 60, 40);
 
   // Allow the fling to run its course. The fling should also be handed off.
   childApzc->AdvanceAnimationsUntilEnd();
@@ -647,12 +652,253 @@ TEST_F(APZScrollHandoffTester, ImmediateHandoffDisallowed_Fling) {
   EXPECT_GT(parentApzc->GetFrameMetrics().GetVisualScrollOffset().y, 10);
 }
 
-TEST_F(APZScrollHandoffTester, CrossApzcAxisLock_NoTouchAction) {
-  SCOPED_GFX_PREF_BOOL("layout.css.touch_action.enabled", false);
+TEST_F(APZScrollHandoffTester, CrossApzcAxisLock_TouchAction) {
   TestCrossApzcAxisLock();
 }
 
-TEST_F(APZScrollHandoffTester, CrossApzcAxisLock_TouchAction) {
-  SCOPED_GFX_PREF_BOOL("layout.css.touch_action.enabled", true);
-  TestCrossApzcAxisLock();
+TEST_F(APZScrollHandoffTesterMock, WheelHandoffAfterDirectionReversal) {
+  // Explicitly set the wheel transaction timeout pref because the test relies
+  // on its value.
+  SCOPED_GFX_PREF_INT("mousewheel.transaction.timeout", 1500);
+
+  // Set up a basic scroll handoff layer tree.
+  CreateScrollHandoffLayerTree1();
+
+  rootApzc = ApzcOf(layers[0]);
+  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
+  FrameMetrics& rootMetrics = rootApzc->GetFrameMetrics();
+  FrameMetrics& childMetrics = childApzc->GetFrameMetrics();
+  CSSRect childScrollRange = childMetrics.CalculateScrollRange();
+
+  EXPECT_EQ(0, rootMetrics.GetVisualScrollOffset().y);
+  EXPECT_EQ(0, childMetrics.GetVisualScrollOffset().y);
+
+  ScreenIntPoint cursorLocation(10, 60);  // positioned to hit the subframe
+  ScreenPoint upwardDelta(0, -10);
+  ScreenPoint downwardDelta(0, 10);
+
+  // First wheel upwards. This will have no effect because we're already
+  // scrolled to the top.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+  Wheel(manager, cursorLocation, upwardDelta, mcc->Time());
+  EXPECT_EQ(0, rootMetrics.GetVisualScrollOffset().y);
+  EXPECT_EQ(0, childMetrics.GetVisualScrollOffset().y);
+
+  // Now wheel downwards 6 times. This should scroll the child, and get it
+  // to the bottom of its 50px scroll range.
+  for (size_t i = 0; i < 6; ++i) {
+    mcc->AdvanceByMillis(100);
+    QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+    Wheel(manager, cursorLocation, downwardDelta, mcc->Time());
+  }
+  EXPECT_EQ(0, rootMetrics.GetVisualScrollOffset().y);
+  EXPECT_EQ(childScrollRange.YMost(), childMetrics.GetVisualScrollOffset().y);
+
+  // Wheel downwards an additional 16 times, with 100ms increments.
+  // This should be enough to overcome the 1500ms wheel transaction timeout
+  // and start scrolling the root.
+  for (size_t i = 0; i < 16; ++i) {
+    mcc->AdvanceByMillis(100);
+    QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+    Wheel(manager, cursorLocation, downwardDelta, mcc->Time());
+  }
+  EXPECT_EQ(childScrollRange.YMost(), childMetrics.GetVisualScrollOffset().y);
+  EXPECT_GT(rootMetrics.GetVisualScrollOffset().y, 0);
+}
+
+TEST_F(APZScrollHandoffTesterMock, WheelHandoffNonscrollable) {
+  // Set up a basic scroll layer tree.
+  CreateScrollHandoffLayerTree5();
+
+  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
+  FrameMetrics& childMetrics = childApzc->GetFrameMetrics();
+
+  EXPECT_EQ(0, childMetrics.GetVisualScrollOffset().y);
+
+  ScreenPoint downwardDelta(0, 10);
+  // Positioned to hit the nonscrollable parent frame
+  ScreenIntPoint nonscrollableLocation(40, 10);
+  // Positioned to hit the scrollable subframe
+  ScreenIntPoint scrollableLocation(40, 60);
+
+  // Start the wheel transaction on a nonscrollable parent frame.
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID);
+  Wheel(manager, nonscrollableLocation, downwardDelta, mcc->Time());
+  EXPECT_EQ(0, childMetrics.GetVisualScrollOffset().y);
+
+  // Mouse moves to a scrollable subframe. This should end the transaction.
+  mcc->AdvanceByMillis(100);
+  MouseInput mouseInput(MouseInput::MOUSE_MOVE,
+                        MouseInput::ButtonType::PRIMARY_BUTTON, 0, 0,
+                        scrollableLocation, mcc->Time(), 0);
+  WidgetMouseEvent mouseEvent =
+      mouseInput.ToWidgetEvent<WidgetMouseEvent>(nullptr);
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+  ((APZInputBridge*)manager.get())->ReceiveInputEvent(mouseEvent);
+
+  // Wheel downward should scroll the subframe.
+  mcc->AdvanceByMillis(100);
+  QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+  Wheel(manager, scrollableLocation, downwardDelta, mcc->Time());
+  EXPECT_GT(childMetrics.GetVisualScrollOffset().y, 0);
+}
+
+TEST_F(APZScrollHandoffTesterMock, ChildCloseToEndOfScrollRange) {
+  SCOPED_GFX_PREF_BOOL("apz.overscroll.enabled", true);
+
+  CreateScrollHandoffLayerTree1();
+
+  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
+
+  FrameMetrics& rootMetrics = rootApzc->GetFrameMetrics();
+  FrameMetrics& childMetrics = childApzc->GetFrameMetrics();
+
+  // Zoom the page in by 3x. This needs to be reflected in the zoom level
+  // and composition bounds of both APZCs.
+  rootMetrics.SetZoom(CSSToParentLayerScale(3.0));
+  rootMetrics.SetCompositionBounds(ParentLayerRect(0, 0, 300, 300));
+  childMetrics.SetZoom(CSSToParentLayerScale(3.0));
+  childMetrics.SetCompositionBounds(ParentLayerRect(0, 150, 300, 150));
+
+  // Scroll the child APZC very close to the end of the scroll range.
+  // The scroll offset is chosen such that in CSS pixels it has 0.01 pixels
+  // room to scroll (less than COORDINATE_EPSILON = 0.02), but in ParentLayer
+  // pixels it has 0.03 pixels room (greater than COORDINATE_EPSILON).
+  childMetrics.SetVisualScrollOffset(CSSPoint(0, 49.99));
+
+  EXPECT_FALSE(childApzc->IsOverscrolled());
+
+  CSSPoint childBefore = childApzc->GetFrameMetrics().GetVisualScrollOffset();
+  CSSPoint parentBefore = rootApzc->GetFrameMetrics().GetVisualScrollOffset();
+
+  // Synthesize a pan gesture that tries to scroll the child further down.
+  PanGesture(PanGestureInput::PANGESTURE_START, childApzc,
+             ScreenIntPoint(10, 20), ScreenPoint(0, 40), mcc->Time());
+  mcc->AdvanceByMillis(5);
+  childApzc->AdvanceAnimations(mcc->GetSampleTime());
+
+  PanGesture(PanGestureInput::PANGESTURE_END, childApzc, ScreenIntPoint(10, 21),
+             ScreenPoint(0, 0), mcc->Time());
+
+  CSSPoint childAfter = childApzc->GetFrameMetrics().GetVisualScrollOffset();
+  CSSPoint parentAfter = rootApzc->GetFrameMetrics().GetVisualScrollOffset();
+
+  bool childScrolled = (childBefore != childAfter);
+  bool parentScrolled = (parentBefore != parentAfter);
+
+  // Check that either the child or the parent scrolled.
+  // (With the current implementation of comparing quantities to
+  // COORDINATE_EPSILON in CSS units, it will be the parent, but the important
+  // thing is that at least one of the child or parent scroll, i.e. we're not
+  // stuck in a situation where no scroll offset is changing).
+  EXPECT_TRUE(childScrolled || parentScrolled);
+}
+
+TEST_F(APZScrollHandoffTesterMock, ScrollJump_Bug1812227) {
+  // Set the touch start tolerance to 10 pixels.
+  SCOPED_GFX_PREF_FLOAT("apz.touch_start_tolerance", 10 / manager->GetDPI());
+
+  CreateScrollHandoffLayerTree6();
+  RefPtr<TestAsyncPanZoomController> childApzc = ApzcOf(layers[1]);
+
+  // Throughout the test, we record the composited vertical scroll position
+  // of the root scroll frame after every event or animation frame.
+  std::vector<CSSCoord> rootYScrollPositions;
+  auto SampleScrollPosition = [&]() {
+    rootYScrollPositions.push_back(
+        rootApzc->GetFrameMetrics().GetVisualScrollOffset().y);
+  };
+
+  // Helper function to perform a light upward flick (finger moves upward
+  // ==> page will scroll downward).
+  auto DoLightUpwardFlick = [&](bool aSimulatePaint = false) {
+    // Don't use Pan() because it decreases the touch start tolerance
+    // to almost zero, and here we want to test a codepath related to
+    // the touch start tolerance.
+
+    mcc->AdvanceByMillis(16);
+    QueueMockHitResult(ScrollableLayerGuid::START_SCROLL_ID + 1);
+    TouchDown(manager, {30, 30}, mcc->Time());
+    SampleScrollPosition();
+
+    // If aSimulatePaint=true, simulate a main-thread paint arriving in between
+    // the touch-down (when the input block is created and the cached value
+    // InputBlockState::mTransformToApzc is set) and the first touch-move which
+    // overcomes the touch-tolerance threshold and synthesizes an additional
+    // touch-move event at the threshold. The paint has the effect of resetting
+    // transform to the APZC to zero. The bug occurs if the synthesized
+    // touch-move event incorrectly uses the up-to-date transform to the APZC
+    // rather than the value cached in InputBlockState::mTrasnformToApzc.
+    if (aSimulatePaint) {
+      // For simplicity, simulate a paint with the latest metrics stored on the
+      // APZC. In practice, what would be painted would be from a frame or two
+      // ago, but for reproducing this bug it does not matter.
+      ModifyFrameMetrics(root, [&](ScrollMetadata&, FrameMetrics& aMetrics) {
+        aMetrics = rootApzc->GetFrameMetrics();
+      });
+      ModifyFrameMetrics(layers[1],
+                         [&](ScrollMetadata&, FrameMetrics& aMetrics) {
+                           aMetrics = childApzc->GetFrameMetrics();
+                         });
+      UpdateHitTestingTree();
+    }
+
+    mcc->AdvanceByMillis(16);
+    TouchMove(manager, {30, 10}, mcc->Time());
+    SampleScrollPosition();
+
+    mcc->AdvanceByMillis(16);
+    TouchUp(manager, {30, 10}, mcc->Time());
+    SampleScrollPosition();
+
+    // The root APZC should be flinging.
+    rootApzc->AssertStateIsFling();
+  };
+
+  // Peform one flick.
+  DoLightUpwardFlick();
+
+  // Sample the resulting fling partway. Testing shows it goes well past
+  // y=100, so sample it until y=100.
+  while (SampleAnimationsOnce() && rootYScrollPositions.back() < 100) {
+    SampleScrollPosition();
+  }
+
+  // Perform a second flick, this time simulating a paint in between
+  // the touch-start and touch-move.
+  DoLightUpwardFlick(true);
+
+  // Sample the fling until its completion.
+  while (SampleAnimationsOnce()) {
+    SampleScrollPosition();
+  }
+
+  // Check that the vertical root scroll position is non-decreasing
+  // throughout the course of the test, i.e. it never jumps back up.
+  for (size_t i = 0; i < (rootYScrollPositions.size() - 1); ++i) {
+    CSSCoord before = rootYScrollPositions[i];
+    CSSCoord after = rootYScrollPositions[i + 1];
+    EXPECT_LE(before, after);
+  }
+}
+
+TEST_F(APZCNestedFlingScrollHandoffTester, FlingInOppositeDirection) {
+  RefPtr<TestAsyncPanZoomController> rootApzc = ApzcOf(root);
+
+  ParentLayerPoint startRootOffset = rootApzc->GetCurrentAsyncScrollOffset(
+      AsyncTransformConsumer::eForEventHandling);
+  ExecuteDirectionChangingPanGesture(
+      ScreenIntPoint{569, 710},
+      {-11, -2, -107, -18, -148, -57, -133, -159, -21}, {11, 2, 42, 107, 148});
+  // Let any animation start and run for a few frames
+  ExecuteWait(TimeDuration::FromMilliseconds(154));
+  auto vel = subframeApzc->GetVelocityVector();
+
+  ParentLayerPoint endRootOffset = rootApzc->GetCurrentAsyncScrollOffset(
+      AsyncTransformConsumer::eForEventHandling);
+
+  EXPECT_EQ(vel.y, 0.0);
+  rootApzc->AssertStateIsReset();
+  subframeApzc->AssertStateIsReset();
+  EXPECT_EQ(startRootOffset.y, endRootOffset.y);
 }
