@@ -12,19 +12,21 @@
 #include "mozilla/Assertions.h"  // for MOZ_ASSERT, etc
 #include "mozilla/Attributes.h"  // for override
 #include "mozilla/RefPtr.h"      // for RefPtr, already_AddRefed, etc
+#include "mozilla/dom/ipc/IdType.h"
 #include "mozilla/gfx/Logging.h"
 #include "mozilla/gfx/Matrix.h"
 #include "mozilla/gfx/Point.h"  // for IntSize, IntPoint
 #include "mozilla/gfx/Rect.h"
-#include "mozilla/gfx/Types.h"  // for SurfaceFormat, etc
-#include "mozilla/ipc/FileDescriptor.h"
+#include "mozilla/gfx/Types.h"               // for SurfaceFormat, etc
 #include "mozilla/layers/CompositorTypes.h"  // for TextureFlags, etc
 #include "mozilla/layers/LayersTypes.h"      // for LayerRenderState, etc
+#include "mozilla/layers/LayersMessages.h"
 #include "mozilla/layers/LayersSurfaces.h"
 #include "mozilla/layers/TextureSourceProvider.h"
 #include "mozilla/mozalloc.h"  // for operator delete
 #include "mozilla/Range.h"
-#include "mozilla/UniquePtr.h"  // for UniquePtr
+#include "mozilla/UniquePtr.h"            // for UniquePtr
+#include "mozilla/UniquePtrExtensions.h"  // for UniqueFileHandle
 #include "mozilla/webrender/WebRenderTypes.h"
 #include "nsCOMPtr.h"         // for already_AddRefed
 #include "nsDebug.h"          // for NS_WARNING
@@ -60,6 +62,8 @@ class Compositor;
 class CompositableParentManager;
 class ReadLockDescriptor;
 class CompositorBridgeParent;
+class DXGITextureHostD3D11;
+class DXGIYCbCrTextureHostD3D11;
 class SurfaceDescriptor;
 class HostIPCAllocator;
 class ISurfaceAllocator;
@@ -76,6 +80,7 @@ class RemoteTextureHostWrapper;
 class TextureParent;
 class WebRenderTextureHost;
 class WrappingTextureSourceYCbCrBasic;
+class TextureHostWrapperD3D11;
 
 /**
  * A view on a TextureHost where the texture is internally represented as tiles
@@ -87,7 +92,7 @@ class WrappingTextureSourceYCbCrBasic;
 class BigImageIterator {
  public:
   virtual void BeginBigImageIteration() = 0;
-  virtual void EndBigImageIteration(){};
+  virtual void EndBigImageIteration() {};
   virtual gfx::IntRect GetTileRect() = 0;
   virtual size_t GetTileCount() = 0;
   virtual bool NextTile() = 0;
@@ -417,7 +422,7 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
    */
   static already_AddRefed<TextureHost> Create(
       const SurfaceDescriptor& aDesc, ReadLockDescriptor&& aReadLock,
-      ISurfaceAllocator* aDeallocator, LayersBackend aBackend,
+      HostIPCAllocator* aDeallocator, LayersBackend aBackend,
       TextureFlags aFlags, wr::MaybeExternalImageId& aExternalImageId);
 
   /**
@@ -495,10 +500,14 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
   virtual void SetCropRect(nsIntRect aCropRect) {}
 
   /**
-   * Debug facility.
+   * Return TextureHost's data as DataSourceSurface.
+   *
+   * @param aSurface may be used as returned DataSourceSurface.
+   *
    * XXX - cool kids use Moz2D. See bug 882113.
    */
-  virtual already_AddRefed<gfx::DataSourceSurface> GetAsSurface() = 0;
+  virtual already_AddRefed<gfx::DataSourceSurface> GetAsSurface(
+      gfx::DataSourceSurface* aSurface = nullptr) = 0;
 
   /**
    * XXX - Flags should only be set at creation time, this will be removed.
@@ -527,8 +536,8 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
   static PTextureParent* CreateIPDLActor(
       HostIPCAllocator* aAllocator, const SurfaceDescriptor& aSharedData,
       ReadLockDescriptor&& aDescriptor, LayersBackend aLayersBackend,
-      TextureFlags aFlags, uint64_t aSerial,
-      const wr::MaybeExternalImageId& aExternalImageId);
+      TextureFlags aFlags, const dom::ContentParentId& aContentId,
+      uint64_t aSerial, const wr::MaybeExternalImageId& aExternalImageId);
   static bool DestroyIPDLActor(PTextureParent* actor);
 
   /**
@@ -545,6 +554,8 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
 
   static uint64_t GetTextureSerial(PTextureParent* actor);
 
+  static dom::ContentParentId GetTextureContentId(PTextureParent* actor);
+
   /**
    * Return a pointer to the IPDLActor.
    *
@@ -610,6 +621,16 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
     return nullptr;
   }
 
+  virtual TextureHostWrapperD3D11* AsTextureHostWrapperD3D11() {
+    return nullptr;
+  }
+
+  virtual DXGITextureHostD3D11* AsDXGITextureHostD3D11() { return nullptr; }
+
+  virtual DXGIYCbCrTextureHostD3D11* AsDXGIYCbCrTextureHostD3D11() {
+    return nullptr;
+  }
+
   virtual bool IsWrappingSurfaceTextureHost() { return false; }
 
   // Create the corresponding RenderTextureHost type of this texture, and
@@ -654,6 +675,9 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
     // Passed in the RenderCompositor supports BufferTextureHosts
     // being used directly as external compositor surfaces.
     SUPPORTS_EXTERNAL_BUFFER_TEXTURES,
+
+    // Passed if the caller wants to disable external compositing of TextureHost
+    EXTERNAL_COMPOSITING_DISABLED,
   };
   using PushDisplayItemFlagSet = EnumSet<PushDisplayItemFlag>;
 
@@ -676,12 +700,12 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
 
   virtual bool NeedsYFlip() const;
 
-  virtual void SetAcquireFence(mozilla::ipc::FileDescriptor&& aFenceFd) {}
+  virtual void SetAcquireFence(UniqueFileHandle&& aFenceFd) {}
 
-  virtual void SetReleaseFence(mozilla::ipc::FileDescriptor&& aFenceFd) {}
+  virtual void SetReleaseFence(UniqueFileHandle&& aFenceFd) {}
 
-  virtual mozilla::ipc::FileDescriptor GetAndResetReleaseFence() {
-    return mozilla::ipc::FileDescriptor();
+  virtual UniqueFileHandle GetAndResetReleaseFence() {
+    return UniqueFileHandle();
   }
 
   virtual AndroidHardwareBuffer* GetAndroidHardwareBuffer() const {
@@ -755,6 +779,7 @@ class TextureHost : public AtomicRefCountedWithFinalize<TextureHost> {
   friend class TextureSourceProvider;
   friend class GPUVideoTextureHost;
   friend class WebRenderTextureHost;
+  friend class TextureHostWrapperD3D11;
 };
 
 /**
@@ -776,9 +801,9 @@ class BufferTextureHost : public TextureHost {
 
   virtual ~BufferTextureHost();
 
-  virtual uint8_t* GetBuffer() = 0;
+  virtual uint8_t* GetBuffer() const = 0;
 
-  virtual size_t GetBufferSize() = 0;
+  virtual size_t GetBufferSize() const = 0;
 
   void UnbindTextureSource() override;
 
@@ -799,9 +824,12 @@ class BufferTextureHost : public TextureHost {
 
   gfx::ColorRange GetColorRange() const override;
 
+  gfx::ChromaSubsampling GetChromaSubsampling() const;
+
   gfx::IntSize GetSize() const override { return mSize; }
 
-  already_AddRefed<gfx::DataSourceSurface> GetAsSurface() override;
+  already_AddRefed<gfx::DataSourceSurface> GetAsSurface(
+      gfx::DataSourceSurface* aSurface) override;
 
   bool NeedsDeferredDeletion() const override {
     return TextureHost::NeedsDeferredDeletion() || UseExternalTextures();
@@ -827,6 +855,12 @@ class BufferTextureHost : public TextureHost {
                         const Range<wr::ImageKey>& aImageKeys,
                         PushDisplayItemFlagSet aFlags) override;
 
+  uint8_t* GetYChannel();
+  uint8_t* GetCbChannel();
+  uint8_t* GetCrChannel();
+  int32_t GetYStride() const;
+  int32_t GetCbCrStride() const;
+
  protected:
   bool UseExternalTextures() const { return mUseExternalTextures; }
 
@@ -859,9 +893,9 @@ class ShmemTextureHost : public BufferTextureHost {
 
   void ForgetSharedData() override;
 
-  uint8_t* GetBuffer() override;
+  uint8_t* GetBuffer() const override;
 
-  size_t GetBufferSize() override;
+  size_t GetBufferSize() const override;
 
   const char* Name() override { return "ShmemTextureHost"; }
 
@@ -893,9 +927,9 @@ class MemoryTextureHost : public BufferTextureHost {
 
   void ForgetSharedData() override;
 
-  uint8_t* GetBuffer() override;
+  uint8_t* GetBuffer() const override;
 
-  size_t GetBufferSize() override;
+  size_t GetBufferSize() const override;
 
   const char* Name() override { return "MemoryTextureHost"; }
 
