@@ -24,19 +24,14 @@
 #include "mozilla/gfx/StackArray.h"
 #include "mozilla/widget/WinCompositorWidget.h"
 
-#include "mozilla/EnumeratedArray.h"
 #include "mozilla/ProfilerState.h"
 #include "mozilla/StaticPrefs_gfx.h"
 #include "mozilla/StaticPrefs_layers.h"
-#include "mozilla/Telemetry.h"
 
 #include "D3D11ShareHandleImage.h"
 #include "DeviceAttachmentsD3D11.h"
 #include "BlendShaderConstants.h"
 
-#include <versionhelpers.h>  // For IsWindows8OrGreater
-#include <winsdkver.h>
-
 namespace mozilla {
 
 using namespace gfx;
@@ -164,15 +159,13 @@ bool CompositorD3D11::Initialize(nsCString* const out_failureReason) {
     hr = dxgiFactory->QueryInterface(
         (IDXGIFactory2**)getter_AddRefs(dxgiFactory2));
 
-#if (_WIN32_WINDOWS_MAXVER >= 0x0A00)
     if (gfxVars::UseDoubleBufferingWithCompositor() && SUCCEEDED(hr) &&
         dxgiFactory2) {
       // DXGI_SCALING_NONE is not available on Windows 7 with Platform Update.
       // This looks awful for things like the awesome bar and browser window
       // resizing so we don't use a flip buffer chain here. When using
       // EFFECT_SEQUENTIAL it looks like windows doesn't stretch the surface
-      // when resizing. We chose not to run this before Windows 10 because it
-      // appears sometimes this breaks our ability to test ASAP compositing.
+      // when resizing.
       RefPtr<IDXGISwapChain1> swapChain;
 
       DXGI_SWAP_CHAIN_DESC1 swapDesc;
@@ -211,9 +204,7 @@ bool CompositorD3D11::Initialize(nsCString* const out_failureReason) {
 
     // In some configurations double buffering may have failed with an
     // ACCESS_DENIED error.
-    if (!mSwapChain)
-#endif
-    {
+    if (!mSwapChain) {
       if (mWidget->AsWindows()->GetCompositorHwnd()) {
         // Destroy compositor window.
         mWidget->AsWindows()->DestroyCompositorWindow();
@@ -516,20 +507,21 @@ void CompositorD3D11::SetRenderTarget(CompositingRenderTarget* aRenderTarget) {
   }
 }
 
-ID3D11PixelShader* CompositorD3D11::GetPSForEffect(Effect* aEffect) {
+ID3D11PixelShader* CompositorD3D11::GetPSForEffect(Effect* aEffect,
+                                                   const ClipType aClipType) {
   switch (aEffect->mType) {
     case EffectTypes::RGB: {
       SurfaceFormat format =
           static_cast<TexturedEffect*>(aEffect)->mTexture->GetFormat();
       return (format == SurfaceFormat::B8G8R8A8 ||
               format == SurfaceFormat::R8G8B8A8)
-                 ? mAttachments->mRGBAShader
-                 : mAttachments->mRGBShader;
+                 ? mAttachments->mRGBAShader[aClipType]
+                 : mAttachments->mRGBShader[aClipType];
     }
     case EffectTypes::NV12:
-      return mAttachments->mNV12Shader;
+      return mAttachments->mNV12Shader[aClipType];
     case EffectTypes::YCBCR:
-      return mAttachments->mYCbCrShader;
+      return mAttachments->mYCbCrShader[aClipType];
     default:
       NS_WARNING("No shader to load");
       return nullptr;
@@ -559,7 +551,8 @@ void CompositorD3D11::ClearRect(const gfx::Rect& aRect) {
   scissor.bottom = aRect.YMost();
   mContext->RSSetScissorRects(1, &scissor);
   mContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP);
-  mContext->VSSetShader(mAttachments->mVSQuadShader, nullptr, 0);
+  mContext->VSSetShader(mAttachments->mVSQuadShader[ClipType::ClipNone],
+                        nullptr, 0);
 
   mContext->PSSetShader(mAttachments->mSolidColorShader, nullptr, 0);
   mPSConstants.layerColor[0] = 0;
@@ -649,10 +642,33 @@ void CompositorD3D11::DrawQuad(const gfx::Rect& aRect,
 
   mContext->RSSetScissorRects(1, &scissor);
 
-  RefPtr<ID3D11VertexShader> vertexShader = mAttachments->mVSQuadShader;
+  auto clipType = ClipType::ClipNone;
+
+  if (aEffectChain.mRoundedClipEffect) {
+    clipType = ClipType::RoundedRect;
+
+    mVSConstants.roundedClipRect[0] = aEffectChain.mRoundedClipEffect->mRect.x;
+    mVSConstants.roundedClipRect[1] = aEffectChain.mRoundedClipEffect->mRect.y;
+    mVSConstants.roundedClipRect[2] =
+        aEffectChain.mRoundedClipEffect->mRect.width;
+    mVSConstants.roundedClipRect[3] =
+        aEffectChain.mRoundedClipEffect->mRect.height;
+
+    mPSConstants.roundedClipRadii[0] =
+        aEffectChain.mRoundedClipEffect->mRadii[eCornerTopLeft].width;
+    mPSConstants.roundedClipRadii[1] =
+        aEffectChain.mRoundedClipEffect->mRadii[eCornerBottomLeft].width;
+    mPSConstants.roundedClipRadii[2] =
+        aEffectChain.mRoundedClipEffect->mRadii[eCornerTopRight].width;
+    mPSConstants.roundedClipRadii[3] =
+        aEffectChain.mRoundedClipEffect->mRadii[eCornerBottomRight].width;
+  }
+
+  RefPtr<ID3D11VertexShader> vertexShader =
+      mAttachments->mVSQuadShader[clipType];
 
   RefPtr<ID3D11PixelShader> pixelShader =
-      GetPSForEffect(aEffectChain.mPrimaryEffect);
+      GetPSForEffect(aEffectChain.mPrimaryEffect, clipType);
 
   mContext->VSSetShader(vertexShader, nullptr, 0);
   mContext->PSSetShader(pixelShader, nullptr, 0);
@@ -851,9 +867,9 @@ Maybe<IntRect> CompositorD3D11::BeginFrame(const nsIntRegion& aInvalidRegion,
 
   PrepareStaticVertexBuffer();
 
-  mBackBufferInvalid.Or(mBackBufferInvalid, invalidRegionSafe);
+  mBackBufferInvalid.OrWith(invalidRegionSafe);
   if (mIsDoubleBuffered) {
-    mFrontBufferInvalid.Or(mFrontBufferInvalid, invalidRegionSafe);
+    mFrontBufferInvalid.OrWith(invalidRegionSafe);
   }
 
   // We have to call UpdateRenderTarget after we've determined the invalid regi
