@@ -16,7 +16,6 @@
 #include "gfxPlatform.h"            // for gfxPlatform
 #include "gfxRect.h"                // for gfxRect
 #include "gfxUtils.h"               // for gfxUtils, etc
-#include "mozilla/ArrayUtils.h"     // for ArrayLength
 #include "mozilla/Preferences.h"    // for Preferences
 #include "mozilla/ProfilerLabels.h"
 #include "mozilla/StaticPrefs_gfx.h"
@@ -198,11 +197,9 @@ CompositorOGL::CompositorOGL(widget::CompositorWidget* aWidget,
       mFrameInProgress(false),
       mDestroyed(false),
       mViewportSize(0, 0) {
-  if (aWidget->GetNativeLayerRoot()) {
-    // We can only render into native layers, our GLContext won't have a usable
-    // default framebuffer.
-    mCanRenderToDefaultFramebuffer = false;
-  }
+  // If we render into native layers, our GLContext won't have a usable default
+  // framebuffer.
+  mCanRenderToDefaultFramebuffer = !aWidget->GetNativeLayerRoot();
   MOZ_COUNT_CTOR(CompositorOGL);
 }
 
@@ -422,7 +419,7 @@ bool CompositorOGL::Initialize(nsCString* const out_failureReason) {
     mGLContext->fGenFramebuffers(1, &testFBO);
     GLuint testTexture = 0;
 
-    for (uint32_t i = 0; i < ArrayLength(textureTargets); i++) {
+    for (uint32_t i = 0; i < std::size(textureTargets); i++) {
       GLenum target = textureTargets[i];
       if (!target) continue;
 
@@ -745,9 +742,13 @@ Maybe<IntRect> CompositorOGL::BeginFrame(const nsIntRegion& aInvalidRegion,
     MakeCurrent(ForceMakeCurrent);
 
     mWidgetSize = LayoutDeviceIntSize::FromUnknownSize(rect.Size());
-#ifdef MOZ_WIDGET_GTK
+#ifdef MOZ_WAYLAND
     if (mWidget && mWidget->AsGTK()) {
-      mWidget->AsGTK()->SetEGLNativeWindowSize(mWidgetSize);
+      // Wayland only check we have correct window size to avoid
+      // rendering artifacts.
+      if (!mWidget->AsGTK()->SetEGLNativeWindowSize(mWidgetSize)) {
+        return Nothing();
+      }
     }
 #endif
   } else {
@@ -799,8 +800,8 @@ Maybe<IntRect> CompositorOGL::BeginFrame(const nsIntRegion& aInvalidRegion,
   if (regionToClear.IsEmpty() &&
       mGLContext->IsSupported(GLFeature::invalidate_framebuffer)) {
     GLenum attachments[] = {LOCAL_GL_COLOR};
-    mGLContext->fInvalidateFramebuffer(
-        LOCAL_GL_FRAMEBUFFER, MOZ_ARRAY_LENGTH(attachments), attachments);
+    mGLContext->fInvalidateFramebuffer(LOCAL_GL_FRAMEBUFFER,
+                                       std::size(attachments), attachments);
   } else {
     clearBits |= LOCAL_GL_COLOR_BUFFER_BIT;
   }
@@ -835,67 +836,6 @@ void CompositorOGL::CreateFBOWithTexture(const gfx::IntRect& aRect,
   mGLContext->fGenFramebuffers(1, aFBO);
 }
 
-// Should be called after calls to fReadPixels or fCopyTexImage2D, and other
-// GL read calls.
-static void WorkAroundAppleIntelHD3000GraphicsGLDriverBug(GLContext* aGL) {
-#ifdef XP_MACOSX
-  if (aGL->WorkAroundDriverBugs() &&
-      aGL->Renderer() == GLRenderer::IntelHD3000) {
-    // Work around a bug in the Apple Intel HD Graphics 3000 driver (bug
-    // 1586627, filed with Apple as FB7379358). This bug has been found present
-    // on 10.9.3 and on 10.13.6, so it likely affects all shipped versions of
-    // this driver. (macOS 10.14 does not support this GPU.)
-    // The bug manifests as follows: Reading from a framebuffer puts that
-    // framebuffer into a state such that deleting that framebuffer can break
-    // other framebuffers in certain cases. More specifically, if you have two
-    // framebuffers A and B, the following sequence of events breaks subsequent
-    // drawing to B:
-    //  1. A becomes "most recently read-from framebuffer".
-    //  2. B is drawn to.
-    //  3. A is deleted, and other GL state (such as GL_SCISSOR enabled state)
-    //     is touched.
-    //  4. B is drawn to again.
-    // Now all draws to framebuffer B, including the draw from step 4, will
-    // render at the wrong position and upside down.
-    //
-    // When AfterGLReadCall() is called, the currently bound framebuffer is the
-    // framebuffer that has been read from most recently. So in the presence of
-    // this bug, deleting this framebuffer has now become dangerous. We work
-    // around the bug by creating a new short-lived framebuffer, making that new
-    // framebuffer the most recently read-from framebuffer (using
-    // glCopyTexImage2D), and then deleting it under controlled circumstances.
-    // This deletion is not affected by the bug because our deletion call is not
-    // interleaved with draw calls to another framebuffer and a touching of the
-    // GL scissor enabled state.
-
-    ScopedTexture texForReading(aGL);
-    {
-      // Initialize a 1x1 texture.
-      ScopedBindTexture autoBindTexForReading(aGL, texForReading);
-      aGL->fTexImage2D(LOCAL_GL_TEXTURE_2D, 0, LOCAL_GL_RGBA, 1, 1, 0,
-                       LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE, nullptr);
-      aGL->fTexParameteri(LOCAL_GL_TEXTURE_2D, LOCAL_GL_TEXTURE_MIN_FILTER,
-                          LOCAL_GL_LINEAR);
-      aGL->fTexParameteri(LOCAL_GL_TEXTURE_2D, LOCAL_GL_TEXTURE_MAG_FILTER,
-                          LOCAL_GL_LINEAR);
-    }
-    // Make a framebuffer around the texture.
-    ScopedFramebufferForTexture autoFBForReading(aGL, texForReading);
-    if (autoFBForReading.IsComplete()) {
-      // "Read" from the framebuffer, by initializing a new texture using
-      // glCopyTexImage2D. This flips the bad bit on autoFBForReading.FB().
-      ScopedBindFramebuffer autoFB(aGL, autoFBForReading.FB());
-      ScopedTexture texReadingDest(aGL);
-      ScopedBindTexture autoBindTexReadingDest(aGL, texReadingDest);
-      aGL->fCopyTexImage2D(LOCAL_GL_TEXTURE_2D, 0, LOCAL_GL_RGBA, 0, 0, 1, 1,
-                           0);
-    }
-    // When autoFBForReading goes out of scope, the "poisoned" framebuffer is
-    // deleted, and the bad state seems to go away along with it.
-  }
-#endif
-}
-
 GLuint CompositorOGL::CreateTexture(const IntRect& aRect, bool aCopyFromSource,
                                     GLuint aSourceFrameBuffer,
                                     IntSize* aAllocSize) {
@@ -940,7 +880,6 @@ GLuint CompositorOGL::CreateTexture(const IntRect& aRect, bool aCopyFromSource,
       mGLContext->fCopyTexImage2D(mFBOTextureTarget, 0, LOCAL_GL_RGBA,
                                   clampedRect.X(), FlipY(clampedRect.YMost()),
                                   clampedRectWidth, clampedRectHeight, 0);
-      WorkAroundAppleIntelHD3000GraphicsGLDriverBug(mGLContext);
     } else {
       // Curses, incompatible formats.  Take a slow path.
 
@@ -951,7 +890,6 @@ GLuint CompositorOGL::CreateTexture(const IntRect& aRect, bool aCopyFromSource,
       mGLContext->fReadPixels(clampedRect.X(), clampedRect.Y(),
                               clampedRectWidth, clampedRectHeight,
                               LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE, buf.get());
-      WorkAroundAppleIntelHD3000GraphicsGLDriverBug(mGLContext);
       mGLContext->fTexImage2D(mFBOTextureTarget, 0, LOCAL_GL_RGBA,
                               clampedRectWidth, clampedRectHeight, 0,
                               LOCAL_GL_RGBA, LOCAL_GL_UNSIGNED_BYTE, buf.get());
@@ -990,7 +928,8 @@ GLuint CompositorOGL::CreateTexture(const IntRect& aRect, bool aCopyFromSource,
 }
 
 ShaderConfigOGL CompositorOGL::GetShaderConfigFor(Effect* aEffect,
-                                                  bool aDEAAEnabled) const {
+                                                  bool aDEAAEnabled,
+                                                  bool aRoundedClip) const {
   ShaderConfigOGL config;
 
   switch (aEffect->mType) {
@@ -1027,7 +966,7 @@ ShaderConfigOGL CompositorOGL::GetShaderConfigFor(Effect* aEffect,
           source->GetFormat() == gfx::SurfaceFormat::R8G8B8A8 ||
               source->GetFormat() == gfx::SurfaceFormat::R8G8B8X8 ||
               source->GetFormat() == gfx::SurfaceFormat::R5G6B5_UINT16 ||
-              source->GetFormat() == gfx::SurfaceFormat::YUV422);
+              source->GetFormat() == gfx::SurfaceFormat::YUY2);
       config = ShaderConfigFromTargetAndFormat(source->GetTextureTarget(),
                                                source->GetFormat());
       if (!texturedEffect->mPremultiplied) {
@@ -1037,6 +976,7 @@ ShaderConfigOGL CompositorOGL::GetShaderConfigFor(Effect* aEffect,
     }
   }
   config.SetDEAA(aDEAAEnabled);
+  config.SetRoundedClip(aRoundedClip);
   return config;
 }
 
@@ -1168,8 +1108,10 @@ void CompositorOGL::DrawGeometry(const Geometry& aGeometry,
   bool bEnableAA = StaticPrefs::layers_deaa_enabled() &&
                    !aTransform.Is2DIntegerTranslation();
 
+  bool bRoundedClip = aEffectChain.mRoundedClipEffect;
+
   ShaderConfigOGL config =
-      GetShaderConfigFor(aEffectChain.mPrimaryEffect, bEnableAA);
+      GetShaderConfigFor(aEffectChain.mPrimaryEffect, bEnableAA, bRoundedClip);
 
   config.SetOpacity(aOpacity != 1.f);
   ApplyPrimitiveConfig(config, aGeometry);
@@ -1195,6 +1137,11 @@ void CompositorOGL::DrawGeometry(const Geometry& aGeometry,
                                    source->GetSize().height);
   }
 
+  if (aEffectChain.mRoundedClipEffect) {
+    program->SetRoundedClipRect(aEffectChain.mRoundedClipEffect->mRect);
+    program->SetRoundedClipRadii(aEffectChain.mRoundedClipEffect->mRadii);
+  }
+
   // XXX kip - These calculations could be performed once per layer rather than
   //           for every tile.  This might belong in Compositor.cpp once DEAA
   //           is implemented for DirectX.
@@ -1461,16 +1408,13 @@ void CompositorOGL::InitializeVAO(const GLuint aAttrib, const GLint aComponents,
 #ifdef MOZ_DUMP_PAINTING
 template <typename T>
 void WriteSnapshotToDumpFile_internal(T* aObj, DataSourceSurface* aSurf) {
-  nsCString string(aObj->Name());
-  string.Append('-');
-  string.AppendInt((uint64_t)aObj);
   if (gfxUtils::sDumpPaintFile != stderr) {
-    fprintf_stderr(gfxUtils::sDumpPaintFile, R"(array["%s"]=")",
-                   string.BeginReading());
-  }
-  gfxUtils::DumpAsDataURI(aSurf, gfxUtils::sDumpPaintFile);
-  if (gfxUtils::sDumpPaintFile != stderr) {
-    fprintf_stderr(gfxUtils::sDumpPaintFile, R"(";)");
+    gfxUtils::DumpAsDataURI(aSurf, gfxUtils::sDumpPaintFile);
+  } else {
+    nsCString uri = gfxUtils::GetAsDataURI(aSurf);
+    nsPrintfCString string(R"(array["%s-%)" PRIu64 R"("]="%s";\n)",
+                           aObj->Name(), uint64_t(aObj), uri.BeginReading());
+    fprintf_stderr(gfxUtils::sDumpPaintFile, "%s", string.get());
   }
 }
 
