@@ -8,20 +8,19 @@
 
 #include "DisplayItemClipChain.h"
 #include "FrameMetrics.h"
+#include "mozilla/ScrollContainerFrame.h"
 #include "mozilla/dom/Document.h"
 #include "mozilla/layers/StackingContextHelper.h"
 #include "mozilla/layers/WebRenderLayerManager.h"
+#include "mozilla/StaticPrefs_apz.h"
 #include "mozilla/webrender/WebRenderAPI.h"
 #include "nsDisplayList.h"
 #include "nsRefreshDriver.h"
 #include "nsStyleStructInlines.h"
 #include "UnitTransforms.h"
 
-// clang-format off
-#define CLIP_LOG(...)
-//#define CLIP_LOG(s_, ...) printf_stderr("CLIP(%s): " s_, __func__, ## __VA_ARGS__)
-//#define CLIP_LOG(s_, ...) if (XRE_IsContentProcess()) printf_stderr("CLIP(%s): " s_, __func__, ## __VA_ARGS__)
-// clang-format on
+static mozilla::LazyLogModule sClipLog("wr.clip");
+#define CLIP_LOG(...) MOZ_LOG(sClipLog, LogLevel::Debug, (__VA_ARGS__))
 
 namespace mozilla {
 namespace layers {
@@ -73,7 +72,8 @@ void ClipManager::BeginList(const StackingContextHelper& aStackingContext) {
     }
   }
 
-  CLIP_LOG("  push: clip: %p, asr: %p, scroll = %zu, clip = %zu\n",
+  CLIP_LOG("  push: clip: %p, asr: %p, scroll =%" PRIuPTR ", clip =%" PRIu64
+           "\n",
            clips.mChain, clips.mASR, clips.mScrollId.id,
            clips.mClipChainId.valueOr(wr::WrClipChainId{0}).id);
 
@@ -259,7 +259,7 @@ wr::WrSpaceAndClipChain ClipManager::SwitchItem(nsDisplayListBuilder* aBuilder,
     leafmostASR = ActiveScrolledRoot::PickDescendant(leafmostASR, clip->mASR);
   }
   Maybe<wr::WrSpatialId> leafmostId = DefineScrollLayers(leafmostASR, aItem);
-  Unused << leafmostId;
+  (void)leafmostId;
 
   // Define all the clips in the item's clip chain, and obtain a clip chain id
   // for it.
@@ -274,7 +274,8 @@ wr::WrSpaceAndClipChain ClipManager::SwitchItem(nsDisplayListBuilder* aBuilder,
   clips.UpdateSeparateLeaf(*mBuilder, auPerDevPixel);
   auto spaceAndClipChain = clips.GetSpaceAndClipChain();
 
-  CLIP_LOG("  push: clip: %p, asr: %p, scroll = %zu, clip = %zu\n",
+  CLIP_LOG("  push: clip: %p, asr: %p, scroll = %" PRIuPTR ", clip = %" PRIu64
+           "\n",
            clips.mChain, clips.mASR, clips.mScrollId.id,
            clips.mClipChainId.valueOr(wr::WrClipChainId{0}).id);
 
@@ -320,9 +321,9 @@ Maybe<wr::WrSpatialId> ClipManager::DefineScrollLayers(
   Maybe<wr::WrSpatialId> ancestorSpace =
       DefineScrollLayers(aASR->mParent, aItem);
 
-  Maybe<ScrollMetadata> metadata =
-      aASR->mScrollableFrame->ComputeScrollMetadata(mManager, aItem->Frame(),
-                                                    aItem->ToReferenceFrame());
+  ScrollContainerFrame* scrollContainerFrame = aASR->mScrollContainerFrame;
+  Maybe<ScrollMetadata> metadata = scrollContainerFrame->ComputeScrollMetadata(
+      mManager, aItem->Frame(), aItem->ToReferenceFrame());
   if (!metadata) {
     MOZ_ASSERT_UNREACHABLE("Expected scroll metadata to be available!");
     return ancestorSpace;
@@ -334,12 +335,10 @@ Maybe<wr::WrSpatialId> ClipManager::DefineScrollLayers(
     return ancestorSpace;
   }
 
-  nsIScrollableFrame* scrollableFrame = aASR->mScrollableFrame;
-  nsIFrame* scrollFrame = do_QueryFrame(scrollableFrame);
-  nsPoint offset = scrollFrame->GetOffsetToCrossDoc(aItem->Frame()) +
+  nsPoint offset = scrollContainerFrame->GetOffsetToCrossDoc(aItem->Frame()) +
                    aItem->ToReferenceFrame();
   int32_t auPerDevPixel = aItem->Frame()->PresContext()->AppUnitsPerDevPixel();
-  nsRect scrollPort = scrollableFrame->GetScrollPortRect() + offset;
+  nsRect scrollPort = scrollContainerFrame->GetScrollPortRect() + offset;
   LayoutDeviceRect clipBounds =
       LayoutDeviceRect::FromAppUnits(scrollPort, auPerDevPixel);
 
@@ -371,9 +370,9 @@ Maybe<wr::WrSpatialId> ClipManager::DefineScrollLayers(
   LayoutDevicePoint scrollOffset =
       useRoundedOffset
           ? LayoutDevicePoint::FromAppUnitsRounded(
-                scrollableFrame->GetScrollPosition(), auPerDevPixel)
+                scrollContainerFrame->GetScrollPosition(), auPerDevPixel)
           : LayoutDevicePoint::FromAppUnits(
-                scrollableFrame->GetScrollPosition(), auPerDevPixel);
+                scrollContainerFrame->GetScrollPosition(), auPerDevPixel);
 
   // Currently we track scroll-linked effects at the granularity of documents,
   // not scroll frames, so we consider a scroll frame to have a scroll-linked
@@ -386,9 +385,11 @@ Maybe<wr::WrSpatialId> ClipManager::DefineScrollLayers(
   return Some(mBuilder->DefineScrollLayer(
       viewId, parent, wr::ToLayoutRect(contentRect),
       wr::ToLayoutRect(clipBounds), wr::ToLayoutVector2D(scrollOffset),
-      wr::ToWrAPZScrollGeneration(scrollableFrame->ScrollGenerationOnApz()),
+      wr::ToWrAPZScrollGeneration(
+          scrollContainerFrame->ScrollGenerationOnApz()),
       wr::ToWrHasScrollLinkedEffect(hasScrollLinkedEffect),
-      wr::SpatialKey(uint64_t(scrollFrame), 0, wr::SpatialKeyKind::Scroll)));
+      wr::SpatialKey(uint64_t(scrollContainerFrame), 0,
+                     wr::SpatialKeyKind::Scroll)));
 }
 
 Maybe<wr::WrClipChainId> ClipManager::DefineClipChain(
@@ -401,7 +402,7 @@ Maybe<wr::WrClipChainId> ClipManager::DefineClipChain(
   for (const DisplayItemClipChain* chain = aChain; chain;
        chain = chain->mParent) {
     MOZ_DIAGNOSTIC_ASSERT(chain->mOnStack || !chain->mASR ||
-                          chain->mASR->mScrollableFrame);
+                          chain->mASR->mScrollContainerFrame);
 
     if (!chain->mClip.HasClip()) {
       // This item in the chain is a no-op, skip over it
