@@ -8,14 +8,16 @@
 #include "mozilla/MemoryReporting.h"
 #include "mozilla/Sprintf.h"
 #include "mozilla/StaticPrefs_gfx.h"
+#include "mozilla/gfx/ScaledFontMac.h"
 
-#include "gfxCoreTextShaper.h"
 #include <algorithm>
+
+#include "CoreTextFontList.h"
+#include "gfxCoreTextShaper.h"
 #include "gfxPlatformMac.h"
 #include "gfxContext.h"
 #include "gfxFontUtils.h"
 #include "gfxHarfBuzzShaper.h"
-#include "gfxMacPlatformFontList.h"
 #include "gfxFontConstants.h"
 #include "gfxTextRun.h"
 #include "gfxUtils.h"
@@ -33,12 +35,10 @@ struct TagEquals {
 };
 
 gfxMacFont::gfxMacFont(const RefPtr<UnscaledFontMac>& aUnscaledFont,
-                       MacOSFontEntry* aFontEntry,
-                       const gfxFontStyle* aFontStyle)
+                       CTFontEntry* aFontEntry, const gfxFontStyle* aFontStyle)
     : gfxFont(aUnscaledFont, aFontEntry, aFontStyle),
       mCGFont(nullptr),
       mCTFont(nullptr),
-      mFontSmoothingBackgroundColor(aFontStyle->fontSmoothingBackgroundColor),
       mVariationFont(aFontEntry->HasVariations()) {
   mApplySyntheticBold = aFontStyle->NeedsSyntheticBold(aFontEntry);
 
@@ -155,8 +155,8 @@ bool gfxMacFont::ShapeText(DrawTarget* aDrawTarget, const char16_t* aText,
 
   // Currently, we don't support vertical shaping via CoreText,
   // so we ignore RequiresAATLayout if vertical is requested.
-  auto macFontEntry = static_cast<MacOSFontEntry*>(GetFontEntry());
-  if (macFontEntry->RequiresAATLayout() && !aVertical &&
+  auto ctFontEntry = static_cast<CTFontEntry*>(GetFontEntry());
+  if (ctFontEntry->RequiresAATLayout() && !aVertical &&
       StaticPrefs::gfx_font_rendering_coretext_enabled()) {
     if (!mCoreTextShaper) {
       mCoreTextShaper = MakeUnique<gfxCoreTextShaper>(this);
@@ -166,19 +166,19 @@ bool gfxMacFont::ShapeText(DrawTarget* aDrawTarget, const char16_t* aText,
                                    aShapedText)) {
       PostShapingFixup(aDrawTarget, aText, aOffset, aLength, aVertical,
                        aShapedText);
-      if (GetFontEntry()->HasTrackingTable()) {
+      if (ctFontEntry->HasTrackingTable()) {
         // Convert font size from device pixels back to CSS px
         // to use in selecting tracking value
         float trackSize = GetAdjustedSize() *
                           aShapedText->GetAppUnitsPerDevUnit() /
                           AppUnitsPerCSSPixel();
         float tracking =
-            GetFontEntry()->TrackingForCSSPx(trackSize) * mFUnitsConvFactor;
+            ctFontEntry->TrackingForCSSPx(trackSize) * mFUnitsConvFactor;
         // Applying tracking is a lot like the adjustment we do for
         // synthetic bold: we want to apply between clusters, not to
         // non-spacing glyphs within a cluster. So we can reuse that
         // helper here.
-        aShapedText->AdjustAdvancesForSyntheticBold(tracking, aOffset, aLength);
+        aShapedText->ApplyTrackingToClusters(tracking, aOffset, aLength);
       }
       return true;
     }
@@ -221,8 +221,8 @@ void gfxMacFont::InitMetrics() {
   // return the true value for OpenType/CFF fonts (it normalizes to 1000,
   // which then leads to metrics errors when we read the 'hmtx' table to
   // get glyph advances for HarfBuzz, see bug 580863)
-  AutoCFRelease<CFDataRef> headData =
-      ::CGFontCopyTableForTag(mCGFont, TRUETYPE_TAG('h', 'e', 'a', 'd'));
+  AutoCFTypeRef<CFDataRef> headData(
+      ::CGFontCopyTableForTag(mCGFont, TRUETYPE_TAG('h', 'e', 'a', 'd')));
   if (headData) {
     if (size_t(::CFDataGetLength(headData)) >= sizeof(HeadTable)) {
       const HeadTable* head =
@@ -255,7 +255,7 @@ void gfxMacFont::InitMetrics() {
   // use CG's idea of unitsPerEm, which may differ from the "true" value in
   // the head table of the font (see bug 580863)
   gfxFloat cgConvFactor;
-  if (static_cast<MacOSFontEntry*>(mFontEntry.get())->IsCFF()) {
+  if (static_cast<CTFontEntry*>(mFontEntry.get())->IsCFF()) {
     cgConvFactor = mAdjustedSize / ::CGFontGetUnitsPerEm(mCGFont);
   } else {
     cgConvFactor = mFUnitsConvFactor;
@@ -278,8 +278,8 @@ void gfxMacFont::InitMetrics() {
     mMetrics.capHeight = ::CGFontGetCapHeight(mCGFont) * cgConvFactor;
   }
 
-  AutoCFRelease<CFDataRef> cmap =
-      ::CGFontCopyTableForTag(mCGFont, TRUETYPE_TAG('c', 'm', 'a', 'p'));
+  AutoCFTypeRef<CFDataRef> cmap(
+      ::CGFontCopyTableForTag(mCGFont, TRUETYPE_TAG('c', 'm', 'a', 'p')));
 
   uint32_t glyphID;
   mMetrics.zeroWidth = GetCharWidth(cmap, '0', &glyphID, cgConvFactor);
@@ -322,7 +322,7 @@ void gfxMacFont::InitMetrics() {
       delete mHarfBuzzShaper.exchange(nullptr);
       mAdjustedSize = mStyle.GetAdjustedSize(aspect);
       mFUnitsConvFactor = mAdjustedSize / upem;
-      if (static_cast<MacOSFontEntry*>(mFontEntry.get())->IsCFF()) {
+      if (static_cast<CTFontEntry*>(mFontEntry.get())->IsCFF()) {
         cgConvFactor = mAdjustedSize / ::CGFontGetUnitsPerEm(mCGFont);
       } else {
         cgConvFactor = mFUnitsConvFactor;
@@ -434,61 +434,6 @@ gfxFloat gfxMacFont::GetCharWidth(CFDataRef aCmap, char16_t aUniChar,
   return 0;
 }
 
-/* static */
-CTFontRef gfxMacFont::CreateCTFontFromCGFontWithVariations(
-    CGFontRef aCGFont, CGFloat aSize, bool aInstalledFont,
-    CTFontDescriptorRef aFontDesc) {
-  // Avoid calling potentially buggy variation APIs on pre-Sierra macOS
-  // versions (see bug 1331683).
-  //
-  // And on HighSierra, CTFontCreateWithGraphicsFont properly carries over
-  // variation settings from the CGFont to CTFont, so we don't need to do
-  // the extra work here -- and this seems to avoid Core Text crashiness
-  // seen in bug 1454094.
-  //
-  // However, for installed fonts it seems we DO need to copy the variations
-  // explicitly even on 10.13, otherwise fonts fail to render (as in bug
-  // 1455494) when non-default values are used. Fortunately, the crash
-  // mentioned above occurs with data fonts, not (AFAICT) with system-
-  // installed fonts.
-  //
-  // So we only need to do this "the hard way" on Sierra, and on HighSierra
-  // for system-installed fonts; in other cases just let the standard CTFont
-  // function do its thing.
-  //
-  // NOTE in case this ever needs further adjustment: there is similar logic
-  // in four places in the tree (sadly):
-  //    CreateCTFontFromCGFontWithVariations in gfxMacFont.cpp
-  //    CreateCTFontFromCGFontWithVariations in ScaledFontMac.cpp
-  //    CreateCTFontFromCGFontWithVariations in cairo-quartz-font.c
-  //    ctfont_create_exact_copy in SkFontHost_mac.cpp
-
-  CTFontRef ctFont;
-  if (aInstalledFont) {
-    AutoCFRelease<CFDictionaryRef> variations = ::CGFontCopyVariations(aCGFont);
-    if (variations) {
-      AutoCFRelease<CFDictionaryRef> varAttr = ::CFDictionaryCreate(
-          nullptr, (const void**)&kCTFontVariationAttribute,
-          (const void**)&variations, 1, &kCFTypeDictionaryKeyCallBacks,
-          &kCFTypeDictionaryValueCallBacks);
-
-      AutoCFRelease<CTFontDescriptorRef> varDesc =
-          aFontDesc
-              ? ::CTFontDescriptorCreateCopyWithAttributes(aFontDesc, varAttr)
-              : ::CTFontDescriptorCreateWithAttributes(varAttr);
-
-      ctFont = ::CTFontCreateWithGraphicsFont(aCGFont, aSize, nullptr, varDesc);
-    } else {
-      ctFont =
-          ::CTFontCreateWithGraphicsFont(aCGFont, aSize, nullptr, aFontDesc);
-    }
-  } else {
-    ctFont = ::CTFontCreateWithGraphicsFont(aCGFont, aSize, nullptr, aFontDesc);
-  }
-
-  return ctFont;
-}
-
 int32_t gfxMacFont::GetGlyphWidth(uint16_t aGID) {
   if (mVariationFont) {
     // Avoid a potential Core Text crash (bug 1450209) by using
@@ -534,7 +479,23 @@ bool gfxMacFont::GetGlyphBounds(uint16_t aGID, gfxRect* aBounds, bool aTight) {
   gfxRect bounds(bb.origin.x, -(bb.origin.y + bb.size.height), bb.size.width,
                  bb.size.height);
   bounds.Scale(mFUnitsConvFactor);
-  *aBounds = bounds;
+
+  // For bitmap fonts (like Apple Color Emoji), CoreGraphics does not return
+  // accurate bounds, so to try and avoid clipping when the bounds are used
+  // to determine the area to render (e.g. when implementing canvas2d filters),
+  // we inflate the bounds based on global metrics from the font.
+  if (GetFontEntry()->HasColorBitmapTable()) {
+    aBounds->x = std::min(bounds.x, 0.0);
+    aBounds->width = std::max(bounds.width, mMetrics.maxAdvance);
+    // Note that y-coordinates are downwards here, and bounds.y is MINUS the
+    // glyph ascent as it measures from the baseline.
+    aBounds->y = std::min(bounds.y, -mMetrics.maxAscent);
+    aBounds->height =
+        std::max(bounds.YMost(), mMetrics.maxDescent) - aBounds->y;
+  } else {
+    *aBounds = bounds;
+  }
+
   return true;
 }
 
@@ -544,8 +505,8 @@ bool gfxMacFont::GetGlyphBounds(uint16_t aGID, gfxRect* aBounds, bool aTight) {
 // for sfnts, including ALL downloadable fonts, we prefer to use
 // InitMetricsFromSfntTables and avoid platform APIs.
 void gfxMacFont::InitMetricsFromPlatform() {
-  AutoCFRelease<CTFontRef> ctFont =
-      ::CTFontCreateWithGraphicsFont(mCGFont, mAdjustedSize, nullptr, nullptr);
+  AutoCFTypeRef<CTFontRef> ctFont(
+      ::CTFontCreateWithGraphicsFont(mCGFont, mAdjustedSize, nullptr, nullptr));
   if (!ctFont) {
     return;
   }
@@ -588,7 +549,6 @@ already_AddRefed<ScaledFont> gfxMacFont::GetScaledFont(
   bool hasColorGlyphs = fe->HasColorBitmapTable() || fe->TryGetColorGlyphs();
   RefPtr<ScaledFont> newScaledFont = Factory::CreateScaledFontForMacFont(
       GetCGFontRef(), GetUnscaledFont(), GetAdjustedSize(),
-      ToDeviceColor(mFontSmoothingBackgroundColor),
       !mStyle.useGrayscaleAntialiasing, ApplySyntheticBold(), hasColorGlyphs);
   if (!newScaledFont) {
     return nullptr;
@@ -597,7 +557,7 @@ already_AddRefed<ScaledFont> gfxMacFont::GetScaledFont(
   InitializeScaledFont(newScaledFont);
 
   if (mAzureScaledFont.compareExchange(nullptr, newScaledFont.get())) {
-    Unused << newScaledFont.forget();
+    newScaledFont.forget().leak();
   }
   ScaledFont* scaledFont = mAzureScaledFont;
   return do_AddRef(scaledFont);
