@@ -25,7 +25,6 @@
 #include "gfxFT2FontBase.h"
 #include "gfxTextRun.h"
 #include "GLContextProvider.h"
-#include "mozilla/Atomics.h"
 #include "mozilla/Components.h"
 #include "mozilla/dom/ContentChild.h"
 #include "mozilla/FontPropertyTypes.h"
@@ -68,7 +67,7 @@
 #  include "mozilla/widget/nsWaylandDisplay.h"
 #endif
 #ifdef MOZ_WIDGET_GTK
-#  include "mozilla/widget/DMABufLibWrapper.h"
+#  include "mozilla/widget/DMABufDevice.h"
 #  include "mozilla/StaticPrefs_widget.h"
 #endif
 
@@ -83,12 +82,6 @@ using namespace mozilla::unicode;
 using namespace mozilla::widget;
 
 static FT_Library gPlatformFTLibrary = nullptr;
-static int32_t sDPI;
-
-static void screen_resolution_changed(GdkScreen* aScreen, GParamSpec* aPspec,
-                                      gpointer aClosure) {
-  sDPI = 0;
-}
 
 #if defined(MOZ_X11)
 // TODO(aosmond): The envvar is deprecated. We should remove it once EGL is the
@@ -101,7 +94,13 @@ static bool IsX11EGLEnvvarEnabled() {
 
 gfxPlatformGtk::gfxPlatformGtk() {
   if (!gfxPlatform::IsHeadless()) {
-    gtk_init(nullptr, nullptr);
+    if (!gtk_init_check(nullptr, nullptr)) {
+      gfxCriticalNote << "Failed to init Gtk, missing display? DISPLAY="
+                      << getenv("DISPLAY")
+                      << " WAYLAND_DISPLAY=" << getenv("WAYLAND_DISPLAY")
+                      << "\n";
+      abort();
+    }
   }
 
   mIsX11Display = gfxPlatform::IsHeadless() ? false : GdkIsX11Display();
@@ -122,15 +121,11 @@ gfxPlatformGtk::gfxPlatformGtk() {
   MOZ_RELEASE_ASSERT(gPlatformFTLibrary);
   Factory::SetFTLibrary(gPlatformFTLibrary);
 
-  GdkScreen* gdkScreen = gdk_screen_get_default();
-  if (gdkScreen) {
-    g_signal_connect(gdkScreen, "notify::resolution",
-                     G_CALLBACK(screen_resolution_changed), nullptr);
-  }
-
   // Bug 1714483: Force disable FXAA Antialiasing on NV drivers. This is a
   // temporary workaround for a driver bug.
   PR_SetEnv("__GL_ALLOW_FXAA_USAGE=0");
+
+  InitMesaThreading();
 }
 
 gfxPlatformGtk::~gfxPlatformGtk() {
@@ -138,6 +133,14 @@ gfxPlatformGtk::~gfxPlatformGtk() {
   gPlatformFTLibrary = nullptr;
 }
 
+void gfxPlatformGtk::InitAcceleration() {
+  gfxPlatform::InitAcceleration();
+
+  if (XRE_IsContentProcess()) {
+    ImportCachedContentDeviceData();
+  }
+}
+
 void gfxPlatformGtk::InitX11EGLConfig() {
   FeatureState& feature = gfxConfig::GetFeature(Feature::X11_EGL);
 #ifdef MOZ_X11
@@ -172,11 +175,6 @@ void gfxPlatformGtk::InitX11EGLConfig() {
     feature.ForceDisable(FeatureStatus::Broken, "glxtest could not use EGL",
                          "FEATURE_FAILURE_GLXTEST_NO_EGL"_ns);
   }
-
-  if (feature.IsEnabled() && IsX11Display()) {
-    // Enabling glthread crashes on X11/EGL, see bug 1670545
-    PR_SetEnv("mesa_glthread=false");
-  }
 #else
   feature.DisableByDefault(FeatureStatus::Unavailable, "X11 support missing",
                            "FEATURE_FAILURE_NO_X11"_ns);
@@ -187,10 +185,6 @@ void gfxPlatformGtk::InitDmabufConfig() {
   FeatureState& feature = gfxConfig::GetFeature(Feature::DMABUF);
   feature.EnableByDefault();
 
-  if (StaticPrefs::widget_dmabuf_force_enabled_AtStartup()) {
-    feature.UserForceEnable("Force enabled by pref");
-  }
-
   nsCString failureId;
   int32_t status;
   nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
@@ -203,90 +197,86 @@ void gfxPlatformGtk::InitDmabufConfig() {
                     failureId);
   }
 
+  if (StaticPrefs::widget_dmabuf_force_enabled_AtStartup()) {
+    feature.UserForceEnable("Force enabled by pref");
+  } else if (!StaticPrefs::widget_dmabuf_enabled_AtStartup()) {
+    feature.UserDisable("Force disable by pref",
+                        "FEATURE_FAILURE_USER_FORCE_DISABLED"_ns);
+  }
+
   if (!gfxVars::UseEGL()) {
     feature.ForceDisable(FeatureStatus::Unavailable, "Requires EGL",
                          "FEATURE_FAILURE_REQUIRES_EGL"_ns);
   }
 
-  if (feature.IsEnabled()) {
-    nsAutoCString drmRenderDevice;
-    gfxInfo->GetDrmRenderDevice(drmRenderDevice);
-    gfxVars::SetDrmRenderDevice(drmRenderDevice);
+  if (!gfxVars::WebglUseHardware()) {
+    feature.Disable(FeatureStatus::Blocklisted,
+                    "DMABuf disabled with software rendering", failureId);
+  }
 
-    if (!GetDMABufDevice()->IsEnabled(failureId)) {
+  nsAutoCString drmRenderDevice;
+  gfxInfo->GetDrmRenderDevice(drmRenderDevice);
+  gfxVars::SetDrmRenderDevice(drmRenderDevice);
+
+  if (feature.IsEnabled()) {
+    DMABufDeviceLock device;
+    if (!device.GetDMABufDevice()->IsEnabled(failureId)) {
       feature.ForceDisable(FeatureStatus::Failed, "Failed to configure",
                            failureId);
     }
+    // Make sure we have DMABuf formats available.
+    (void)GetGlobalDMABufFormats();
   }
 }
 
-bool gfxPlatformGtk::InitVAAPIConfig(bool aForceEnabledByUser) {
-  FeatureState& feature =
+void gfxPlatformGtk::InitPlatformHardwareVideoConfig() {
+  FeatureState& featureDec =
       gfxConfig::GetFeature(Feature::HARDWARE_VIDEO_DECODING);
-  // We're already configured in parent process
-  if (!XRE_IsParentProcess()) {
-    return feature.IsEnabled();
-  }
-  feature.EnableByDefault();
-
-  int32_t status = nsIGfxInfo::FEATURE_STATUS_UNKNOWN;
-  nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
-  nsCString failureId;
-  if (NS_FAILED(gfxInfo->GetFeatureStatus(
-          nsIGfxInfo::FEATURE_HARDWARE_VIDEO_DECODING, failureId, &status))) {
-    feature.Disable(FeatureStatus::BlockedNoGfxInfo, "gfxInfo is broken",
-                    "FEATURE_FAILURE_NO_GFX_INFO"_ns);
-  } else if (status == nsIGfxInfo::FEATURE_BLOCKED_PLATFORM_TEST) {
-    feature.ForceDisable(FeatureStatus::Unavailable,
-                         "Force disabled by gfxInfo", failureId);
-  } else if (status != nsIGfxInfo::FEATURE_STATUS_OK) {
-    feature.Disable(FeatureStatus::Blocklisted, "Blocklisted by gfxInfo",
-                    failureId);
-  }
-  if (aForceEnabledByUser) {
-    feature.UserForceEnable("Force enabled by pref");
-  }
   if (!gfxVars::UseEGL()) {
-    feature.ForceDisable(FeatureStatus::Unavailable, "Requires EGL",
-                         "FEATURE_FAILURE_REQUIRES_EGL"_ns);
+    featureDec.ForceDisable(FeatureStatus::Unavailable, "Requires EGL",
+                            "FEATURE_FAILURE_REQUIRES_EGL"_ns);
+    gfxConfig::ForceDisable(Feature::HARDWARE_VIDEO_ENCODING,
+                            FeatureStatus::Unavailable, "Requires EGL",
+                            "FEATURE_FAILURE_REQUIRES_EGL"_ns);
+  }
+
+  if (!featureDec.IsEnabled()) {
+    return;
   }
 
   // Configure zero-copy playback feature.
-  if (feature.IsEnabled()) {
-    FeatureState& featureZeroCopy =
-        gfxConfig::GetFeature(Feature::HW_DECODED_VIDEO_ZERO_COPY);
+  FeatureState& featureZeroCopy =
+      gfxConfig::GetFeature(Feature::HW_DECODED_VIDEO_ZERO_COPY);
 
-    featureZeroCopy.EnableByDefault();
-    uint32_t state =
-        StaticPrefs::media_ffmpeg_vaapi_force_surface_zero_copy_AtStartup();
-    if (state == 0) {
-      featureZeroCopy.UserDisable("Force disable by pref",
-                                  "FEATURE_FAILURE_USER_FORCE_DISABLED"_ns);
-    } else if (state == 1) {
-      featureZeroCopy.UserEnable("Force enabled by pref");
-    } else {
-      nsCString failureId;
-      int32_t status = nsIGfxInfo::FEATURE_STATUS_UNKNOWN;
-      nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
-      if (NS_FAILED(gfxInfo->GetFeatureStatus(
-              nsIGfxInfo::FEATURE_HW_DECODED_VIDEO_ZERO_COPY, failureId,
-              &status))) {
-        featureZeroCopy.Disable(FeatureStatus::BlockedNoGfxInfo,
-                                "gfxInfo is broken",
-                                "FEATURE_FAILURE_NO_GFX_INFO"_ns);
-      } else if (status == nsIGfxInfo::FEATURE_BLOCKED_PLATFORM_TEST) {
-        featureZeroCopy.ForceDisable(FeatureStatus::Unavailable,
-                                     "Force disabled by gfxInfo", failureId);
-      } else if (status != nsIGfxInfo::FEATURE_ALLOW_ALWAYS) {
-        featureZeroCopy.Disable(FeatureStatus::Blocklisted,
-                                "Blocklisted by gfxInfo", failureId);
-      }
-    }
-    if (featureZeroCopy.IsEnabled()) {
-      gfxVars::SetHwDecodedVideoZeroCopy(true);
+  featureZeroCopy.EnableByDefault();
+  uint32_t state =
+      StaticPrefs::media_ffmpeg_vaapi_force_surface_zero_copy_AtStartup();
+  if (state == 0) {
+    featureZeroCopy.UserDisable("Force disable by pref",
+                                "FEATURE_FAILURE_USER_FORCE_DISABLED"_ns);
+  } else if (state == 1) {
+    featureZeroCopy.UserEnable("Force enabled by pref");
+  } else {
+    nsCString failureId;
+    int32_t status = nsIGfxInfo::FEATURE_STATUS_UNKNOWN;
+    nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
+    if (NS_FAILED(gfxInfo->GetFeatureStatus(
+            nsIGfxInfo::FEATURE_HW_DECODED_VIDEO_ZERO_COPY, failureId,
+            &status))) {
+      featureZeroCopy.Disable(FeatureStatus::BlockedNoGfxInfo,
+                              "gfxInfo is broken",
+                              "FEATURE_FAILURE_NO_GFX_INFO"_ns);
+    } else if (status == nsIGfxInfo::FEATURE_BLOCKED_PLATFORM_TEST) {
+      featureZeroCopy.ForceDisable(FeatureStatus::Unavailable,
+                                   "Force disabled by gfxInfo", failureId);
+    } else if (status != nsIGfxInfo::FEATURE_ALLOW_ALWAYS) {
+      featureZeroCopy.Disable(FeatureStatus::Blocklisted,
+                              "Blocklisted by gfxInfo", failureId);
     }
   }
-  return feature.IsEnabled();
+  if (featureZeroCopy.IsEnabled()) {
+    gfxVars::SetHwDecodedVideoZeroCopy(true);
+  }
 }
 
 void gfxPlatformGtk::InitWebRenderConfig() {
@@ -297,20 +287,15 @@ void gfxPlatformGtk::InitWebRenderConfig() {
   }
 
   FeatureState& feature = gfxConfig::GetFeature(Feature::WEBRENDER_COMPOSITOR);
-#ifdef RELEASE_OR_BETA
-  feature.ForceDisable(FeatureStatus::Blocked,
-                       "Cannot be enabled in release or beta",
-                       "FEATURE_FAILURE_DISABLE_RELEASE_OR_BETA"_ns);
-#else
+  // HDR requires compositor to work
+#if defined(MOZ_WAYLAND)
   if (feature.IsEnabled()) {
     if (!IsWaylandDisplay()) {
       feature.ForceDisable(FeatureStatus::Unavailable,
                            "Wayland support missing",
                            "FEATURE_FAILURE_NO_WAYLAND"_ns);
-    }
-#  ifdef MOZ_WAYLAND
-    else if (gfxConfig::IsEnabled(Feature::WEBRENDER) &&
-             !gfxConfig::IsEnabled(Feature::DMABUF)) {
+    } else if (gfxConfig::IsEnabled(Feature::WEBRENDER) &&
+               !gfxConfig::IsEnabled(Feature::DMABUF)) {
       // We use zwp_linux_dmabuf_v1 and GBM directly to manage FBOs. In theory
       // this is also possible vie EGLstreams, but we don't bother to implement
       // it as recent NVidia drivers support GBM and DMABuf as well.
@@ -322,9 +307,11 @@ void gfxPlatformGtk::InitWebRenderConfig() {
                            "Requires wp_viewporter protocol support",
                            "FEATURE_FAILURE_REQUIRES_WPVIEWPORTER"_ns);
     }
-#  endif  // MOZ_WAYLAND
   }
-#endif    // RELEASE_OR_BETA
+#else  // MOZ_WAYLAND
+  feature.ForceDisable(FeatureStatus::Unavailable, "Not available on X11",
+                       "FEATURE_FAILURE_NO_WAYLAND"_ns);
+#endif
 
   gfxVars::SetUseWebRenderCompositor(feature.IsEnabled());
 }
@@ -340,6 +327,36 @@ void gfxPlatformGtk::InitPlatformGPUProcessPrefs() {
 #endif
 }
 
+void gfxPlatformGtk::InitMesaThreading() {
+  FeatureState& featureMesaThreading =
+      gfxConfig::GetFeature(Feature::MESA_THREADING);
+  featureMesaThreading.EnableByDefault();
+
+  nsCString failureId;
+  int32_t status;
+  nsCOMPtr<nsIGfxInfo> gfxInfo = components::GfxInfo::Service();
+  if (NS_FAILED(gfxInfo->GetFeatureStatus(nsIGfxInfo::FEATURE_MESA_THREADING,
+                                          failureId, &status))) {
+    featureMesaThreading.Disable(FeatureStatus::BlockedNoGfxInfo,
+                                 "gfxInfo is broken",
+                                 "FEATURE_FAILURE_NO_GFX_INFO"_ns);
+  } else if (status != nsIGfxInfo::FEATURE_STATUS_OK) {
+    featureMesaThreading.Disable(FeatureStatus::Blocklisted,
+                                 "Blocklisted by gfxInfo", failureId);
+  }
+
+  // Enabling glthread crashes on X11/EGL, see bug 1670545
+  if (gfxConfig::IsEnabled(Feature::X11_EGL) && IsX11Display()) {
+    featureMesaThreading.Disable(FeatureStatus::Failed,
+                                 "No glthread with EGL and X11",
+                                 "FEATURE_FAILURE_EGL_X11"_ns);
+  }
+
+  if (!featureMesaThreading.IsEnabled()) {
+    PR_SetEnv("mesa_glthread=false");
+  }
+}
+
 already_AddRefed<gfxASurface> gfxPlatformGtk::CreateOffscreenSurface(
     const IntSize& aSize, gfxImageFormat aFormat) {
   if (!Factory::AllowedSurfaceSize(aSize)) {
@@ -401,7 +418,7 @@ static const char kFontNotoSansSymbols[] = "Noto Sans Symbols";
 static const char kFontNotoSansSymbols2[] = "Noto Sans Symbols2";
 
 void gfxPlatformGtk::GetCommonFallbackFonts(uint32_t aCh, Script aRunScript,
-                                            eFontPresentation aPresentation,
+                                            FontPresentation aPresentation,
                                             nsTArray<const char*>& aFontList) {
   if (PrefersColor(aPresentation)) {
     aFontList.AppendElement(kFontTwemojiMozilla);
@@ -436,50 +453,6 @@ bool gfxPlatformGtk::CreatePlatformFontList() {
   return gfxPlatformFontList::Initialize(new gfxFcPlatformFontList);
 }
 
-int32_t gfxPlatformGtk::GetFontScaleDPI() {
-  MOZ_ASSERT(XRE_IsParentProcess(),
-             "You can access this via LookAndFeel if you need it in child "
-             "processes");
-  if (MOZ_LIKELY(sDPI != 0)) {
-    return sDPI;
-  }
-  GdkScreen* screen = gdk_screen_get_default();
-  // Ensure settings in config files are processed.
-  gtk_settings_get_for_screen(screen);
-  int32_t dpi = int32_t(round(gdk_screen_get_resolution(screen)));
-  if (dpi <= 0) {
-    // Fall back to something reasonable
-    dpi = 96;
-  }
-  sDPI = dpi;
-  return dpi;
-}
-
-double gfxPlatformGtk::GetFontScaleFactor() {
-  // Integer scale factors work well with GTK window scaling, image scaling, and
-  // pixel alignment, but there is a range where 1 is too small and 2 is too
-  // big.
-  //
-  // An additional step of 1.5 is added because this is common scale on WINNT
-  // and at this ratio the advantages of larger rendering outweigh the
-  // disadvantages from scaling and pixel mis-alignment.
-  //
-  // A similar step for 1.25 is added as well, because this is the scale that
-  // "Large text" settings use in gnome, and it seems worth to allow, especially
-  // on already-hidpi environments.
-  int32_t dpi = GetFontScaleDPI();
-  if (dpi < 120) {
-    return 1.0;
-  }
-  if (dpi < 132) {
-    return 1.25;
-  }
-  if (dpi < 168) {
-    return 1.5;
-  }
-  return round(dpi / 96.0);
-}
-
 gfxImageFormat gfxPlatformGtk::GetOffscreenFormat() {
   // Make sure there is a screen
   GdkScreen* screen = gdk_screen_get_default();
@@ -542,23 +515,16 @@ nsTArray<uint8_t> gfxPlatformGtk::GetPlatformCMSOutputProfileData() {
   }
 
   if (XRE_IsContentProcess()) {
-    MOZ_ASSERT(NS_IsMainThread());
-    // This will be passed in during InitChild so we can avoid sending a
-    // sync message back to the parent during init.
-    const mozilla::gfx::ContentDeviceData* contentDeviceData =
-        GetInitContentDeviceData();
-    if (contentDeviceData) {
-      // On Windows, we assert that the profile isn't empty, but on
-      // Linux it can legitimately be empty if the display isn't
-      // calibrated.  Thus, no assertion here.
-      return contentDeviceData->cmsOutputProfileData().Clone();
+    auto& cmsOutputProfileData = GetCMSOutputProfileData();
+    // We should have set our profile data when we received our initial
+    // ContentDeviceData.
+    MOZ_ASSERT(cmsOutputProfileData.isSome(),
+               "Should have created output profile data when we received "
+               "initial content device data.");
+    if (cmsOutputProfileData.isSome()) {
+      return cmsOutputProfileData.ref().Clone();
     }
-
-    // Otherwise we need to ask the parent for the updated color profile
-    mozilla::dom::ContentChild* cc = mozilla::dom::ContentChild::GetSingleton();
-    nsTArray<uint8_t> result;
-    Unused << cc->SendGetOutputColorProfileData(&result);
-    return result;
+    return nsTArray<uint8_t>();
   }
 
   if (!mIsX11Display) {
@@ -720,12 +686,11 @@ class GtkVsyncSource final : public VsyncSource {
     Window root = DefaultRootWindow(mXDisplay);
     int screen = DefaultScreen(mXDisplay);
 
-    ScopedXFree<GLXFBConfig> cfgs;
     GLXFBConfig config;
     int visid;
     bool forWebRender = false;
     if (!gl::GLContextGLX::FindFBConfigForWindow(
-            mXDisplay, screen, root, &cfgs, &config, &visid, forWebRender)) {
+            mXDisplay, screen, root, &config, &visid, forWebRender)) {
       lock.NotifyAll();
       return;
     }
@@ -1015,3 +980,9 @@ void gfxPlatformGtk::BuildContentDeviceData(ContentDeviceData* aOut) {
 
   aOut->cmsOutputProfileData() = GetPlatformCMSOutputProfileData();
 }
+
+// Wrapper for third party code (WebRTC for instance) where
+// gfxVars can't be included.
+namespace mozilla::gfx {
+bool IsDMABufEnabled() { return gfxVars::UseDMABuf(); }
+}  // namespace mozilla::gfx
