@@ -92,7 +92,7 @@
 //! [segment.rs]: ../segment/index.html
 //!
 
-use api::{BorderRadius, ClipMode, ComplexClipRegion, ImageMask, ClipId, ClipChainId};
+use api::{BorderRadius, ClipMode, ImageMask, ClipId, ClipChainId};
 use api::{BoxShadowClipMode, FillRule, ImageKey, ImageRendering};
 use api::units::*;
 use crate::image_tiling::{self, Repetition};
@@ -112,7 +112,7 @@ use crate::render_task_graph::RenderTaskGraphBuilder;
 use crate::resource_cache::{ImageRequest, ResourceCache};
 use crate::scene_builder_thread::Interners;
 use crate::space::SpaceMapper;
-use crate::util::{clamp_to_scale_factor, MaxRect, extract_inner_rect_safe, project_rect, ScaleOffset};
+use crate::util::{clamp_to_scale_factor, extract_inner_rect_safe, project_rect, MatrixHelpers, MaxRect, ScaleOffset};
 use euclid::approxeq::ApproxEq;
 use std::{iter, ops, u32, mem};
 
@@ -187,6 +187,20 @@ impl ClipTree {
         }
     }
 
+    pub fn reset(&mut self) {
+        self.nodes.clear();
+        self.nodes.push(ClipTreeNode {
+            handle: ClipDataHandle::INVALID,
+            children: Vec::new(),
+            parent: ClipNodeId::NONE,
+        });
+
+        self.leaves.clear();
+
+        self.clip_root_stack.clear();
+        self.clip_root_stack.push(ClipNodeId::NONE);
+    }
+
     /// Add a set of clips to the provided tree node id, reusing existing
     /// nodes in the tree where possible
     fn add_impl(
@@ -274,6 +288,18 @@ impl ClipTree {
         &self.nodes[id.0 as usize]
     }
 
+    pub fn get_parent(&self, id: ClipNodeId) -> Option<ClipNodeId> {
+        // Invalid ids point to the first item in the nodes vector which
+        // has an invalid id for the parent so we don't need to handle
+        // `id` being invalid separately.
+        let parent = self.nodes[id.0 as usize].parent;
+        if parent == ClipNodeId::NONE {
+            return None;
+        }
+
+        return Some(parent)
+    }
+
     /// Retrieve a clip tree leaf by id
     pub fn get_leaf(&self, id: ClipLeafId) -> &ClipTreeLeaf {
         &self.leaves[id.0 as usize]
@@ -430,6 +456,24 @@ impl ClipTreeBuilder {
         }
     }
 
+    pub fn begin(&mut self) {
+        self.clip_map.clear();
+        self.clip_chain_map.clear();
+        self.clip_chains.clear();
+        self.clip_stack.clear();
+        self.clip_stack.push(ClipStackEntry {
+            clip_node_id: ClipNodeId::NONE,
+            last_clip_chain_cache: None,
+            seen_clips: FastHashSet::default(),
+        });
+        self.tree.reset();
+        self.clip_handles_buffer.clear();
+    }
+
+    pub fn recycle_tree(&mut self, tree: ClipTree) {
+        self.tree = tree;
+    }
+
     /// Define a new rect clip
     pub fn define_rect_clip(
         &mut self,
@@ -481,10 +525,16 @@ impl ClipTreeBuilder {
         &mut self,
         clip_chain_id: Option<ClipChainId>,
         reset_seen: bool,
+        ignore_ancestor_clips: bool,
     ) {
         let (mut clip_node_id, mut seen_clips) = {
             let prev = self.clip_stack.last().unwrap();
-            (prev.clip_node_id, prev.seen_clips.clone())
+            let clip_node_id = if ignore_ancestor_clips {
+                ClipNodeId::NONE
+            } else {
+                prev.clip_node_id
+            };
+            (clip_node_id, prev.seen_clips.clone())
         };
 
         if let Some(clip_chain_id) = clip_chain_id {
@@ -684,9 +734,20 @@ impl ClipTreeBuilder {
         false
     }
 
+    pub fn get_parent(&self, id: ClipNodeId) -> Option<ClipNodeId> {
+        self.tree.get_parent(id)
+    }
+
     /// Finalize building and return the clip-tree
-    pub fn finalize(self) -> ClipTree {
-        self.tree
+    pub fn finalize(&mut self) -> ClipTree {
+        // Note: After this, the builder's clip tree does not hold allocations and
+        // is not in valid state. `ClipTreeBuilder::begin()` must be called before
+        // building can happen again.
+        std::mem::replace(&mut self.tree, ClipTree {
+            nodes: Vec::new(),
+            leaves: Vec::new(),
+            clip_root_stack: Vec::new(),
+        })
     }
 
     /// Get a clip node by id
@@ -892,24 +953,36 @@ impl From<ClipItemKey> for ClipNode {
 }
 
 // Flags that are attached to instances of clip nodes.
+#[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
+#[derive(Copy, PartialEq, Eq, Clone, PartialOrd, Ord, Hash, MallocSizeOf)]
+pub struct ClipNodeFlags(u8);
+
 bitflags! {
-    #[cfg_attr(feature = "capture", derive(Serialize))]
-    #[cfg_attr(feature = "replay", derive(Deserialize))]
-    #[derive(MallocSizeOf)]
-    pub struct ClipNodeFlags: u8 {
+    impl ClipNodeFlags : u8 {
         const SAME_SPATIAL_NODE = 0x1;
         const SAME_COORD_SYSTEM = 0x2;
         const USE_FAST_PATH = 0x4;
     }
 }
 
+impl core::fmt::Debug for ClipNodeFlags {
+    fn fmt(&self, f: &mut core::fmt::Formatter) -> core::fmt::Result {
+        if self.is_empty() {
+            write!(f, "{:#x}", Self::empty().bits())
+        } else {
+            bitflags::parser::to_writer(self, f)
+        }
+    }
+}
+
 // When a clip node is found to be valid for a
 // clip chain instance, it's stored in an index
 // buffer style structure. This struct contains
 // an index to the node data itself, as well as
 // some flags describing how this clip node instance
 // is positioned.
-#[derive(Debug, MallocSizeOf)]
+#[derive(Debug, Clone, MallocSizeOf)]
 #[cfg_attr(feature = "capture", derive(Serialize))]
 #[cfg_attr(feature = "replay", derive(Deserialize))]
 pub struct ClipNodeInstance {
@@ -954,17 +1027,18 @@ impl ClipNodeRange {
 //TODO: merge with `CoordinateSpaceMapping`?
 #[derive(Debug, MallocSizeOf)]
 #[cfg_attr(feature = "capture", derive(Serialize))]
-enum ClipSpaceConversion {
+pub enum ClipSpaceConversion {
     Local,
     ScaleOffset(ScaleOffset),
-    Transform(LayoutToWorldTransform),
+    Transform(LayoutToVisTransform),
 }
 
 impl ClipSpaceConversion {
     /// Construct a new clip space converter between two spatial nodes.
-    fn new(
+    pub fn new(
         prim_spatial_node_index: SpatialNodeIndex,
         clip_spatial_node_index: SpatialNodeIndex,
+        visibility_spatial_node_index: SpatialNodeIndex,
         spatial_tree: &SpatialTree,
     ) -> Self {
         //Note: this code is different from `get_relative_transform` in a way that we only try
@@ -976,15 +1050,15 @@ impl ClipSpaceConversion {
         if prim_spatial_node_index == clip_spatial_node_index {
             ClipSpaceConversion::Local
         } else if prim_spatial_node.coordinate_system_id == clip_spatial_node.coordinate_system_id {
-            let scale_offset = prim_spatial_node.content_transform
-                .inverse()
-                .accumulate(&clip_spatial_node.content_transform);
+            let scale_offset = clip_spatial_node.content_transform
+                .then(&prim_spatial_node.content_transform.inverse());
             ClipSpaceConversion::ScaleOffset(scale_offset)
         } else {
             ClipSpaceConversion::Transform(
-                spatial_tree
-                    .get_world_transform(clip_spatial_node_index)
-                    .into_transform()
+                spatial_tree.get_relative_transform(
+                    clip_spatial_node_index,
+                    visibility_spatial_node_index,
+                ).into_transform().cast_unit()
             )
         }
     }
@@ -1089,7 +1163,7 @@ impl ClipNodeInfo {
                             }
 
                             let task_id = rg_builder.add().init(
-                                RenderTask::new_image(props.descriptor.size, req)
+                                RenderTask::new_image(props.descriptor.size, req, false)
                             );
 
                             mask_tiles.push(VisibleMaskImageTile {
@@ -1108,7 +1182,7 @@ impl ClipNodeInfo {
                     let tile_range_start = mask_tiles.len();
 
                     let task_id = rg_builder.add().init(
-                        RenderTask::new_image(props.descriptor.size, request)
+                        RenderTask::new_image(props.descriptor.size, request, false)
                     );
 
                     mask_tiles.push(VisibleMaskImageTile {
@@ -1240,6 +1314,14 @@ impl ClipStore {
         }
     }
 
+    pub fn reset(&mut self) {
+        self.clip_node_instances.clear();
+        self.mask_tiles.clear();
+        self.active_clip_node_info.clear();
+        self.active_local_clip_rect = None;
+        self.active_pic_coverage_rect = PictureRect::max_rect();
+    }
+
     pub fn get_instance_from_range(
         &self,
         node_range: &ClipNodeRange,
@@ -1253,6 +1335,7 @@ impl ClipStore {
         &mut self,
         prim_spatial_node_index: SpatialNodeIndex,
         pic_spatial_node_index: SpatialNodeIndex,
+        visibility_spatial_node_index: SpatialNodeIndex,
         clip_leaf_id: ClipLeafId,
         spatial_tree: &SpatialTree,
         clip_data_store: &ClipDataStore,
@@ -1268,13 +1351,14 @@ impl ClipStore {
         let mut local_clip_rect = clip_leaf.local_clip_rect;
         let mut current = clip_leaf.node_id;
 
-        while current != clip_root {
+        while current != clip_root && current != ClipNodeId::NONE {
             let node = clip_tree.get_node(current);
 
             if !add_clip_node_to_current_chain(
                 node.handle,
                 prim_spatial_node_index,
                 pic_spatial_node_index,
+                visibility_spatial_node_index,
                 &mut local_clip_rect,
                 &mut self.active_clip_node_info,
                 &mut self.active_pic_coverage_rect,
@@ -1295,6 +1379,7 @@ impl ClipStore {
         &mut self,
         prim_clip_chain: &ClipChainInstance,
         prim_spatial_node_index: SpatialNodeIndex,
+        visibility_spatial_node_index: SpatialNodeIndex,
         spatial_tree: &SpatialTree,
         clip_data_store: &ClipDataStore,
     ) {
@@ -1313,6 +1398,7 @@ impl ClipStore {
             let conversion = ClipSpaceConversion::new(
                 prim_spatial_node_index,
                 clip.item.spatial_node_index,
+                visibility_spatial_node_index,
                 spatial_tree,
             );
             self.active_clip_node_info.push(ClipNodeInfo {
@@ -1382,18 +1468,41 @@ impl ClipStore {
         Some(inner_rect)
     }
 
+    // Directly construct a clip node range, ready for rendering, from an interned clip handle.
+    // Typically useful for drawing specific clips on custom pattern / child render tasks that
+    // aren't primitives.
+    // TODO(gw): For now, we assume they are local clips only - in future we might want to support
+    //           non-local clips.
+    pub fn push_clip_instance(
+        &mut self,
+        handle: ClipDataHandle,
+    ) -> ClipNodeRange {
+        let first = self.clip_node_instances.len() as u32;
+
+        self.clip_node_instances.push(ClipNodeInstance {
+            handle,
+            flags: ClipNodeFlags::SAME_COORD_SYSTEM | ClipNodeFlags::SAME_SPATIAL_NODE,
+            visible_tiles: None,
+        });
+
+        ClipNodeRange {
+            first,
+            count: 1,
+        }
+    }
+
     /// The main interface external code uses. Given a local primitive, positioning
     /// information, and a clip chain id, build an optimized clip chain instance.
     pub fn build_clip_chain_instance(
         &mut self,
         local_prim_rect: LayoutRect,
         prim_to_pic_mapper: &SpaceMapper<LayoutPixel, PicturePixel>,
-        pic_to_world_mapper: &SpaceMapper<PicturePixel, WorldPixel>,
+        pic_to_vis_mapper: &SpaceMapper<PicturePixel, VisPixel>,
         spatial_tree: &SpatialTree,
         gpu_cache: &mut GpuCache,
         resource_cache: &mut ResourceCache,
         device_pixel_scale: DevicePixelScale,
-        world_rect: &WorldRect,
+        culling_rect: &VisRect,
         clip_data_store: &mut ClipDataStore,
         rg_builder: &mut RenderTaskGraphBuilder,
         request_resources: bool,
@@ -1406,7 +1515,7 @@ impl ClipStore {
 
         let local_bounding_rect = local_prim_rect.intersection(&local_clip_rect)?;
         let mut pic_coverage_rect = prim_to_pic_mapper.map(&local_bounding_rect)?;
-        let world_clip_rect = pic_to_world_mapper.map(&pic_coverage_rect)?;
+        let vis_clip_rect = pic_to_vis_mapper.map(&pic_coverage_rect)?;
 
         // Now, we've collected all the clip nodes that *potentially* affect this
         // primitive region, and reduced the size of the prim region as much as possible.
@@ -1434,8 +1543,8 @@ impl ClipStore {
                     has_non_local_clips = true;
                     node.item.kind.get_clip_result_complex(
                         transform,
-                        &world_clip_rect,
-                        world_rect,
+                        &vis_clip_rect,
+                        culling_rect,
                     )
                 }
             };
@@ -1539,20 +1648,9 @@ impl ClipStore {
     }
 }
 
-pub struct ComplexTranslateIter<I> {
-    source: I,
-    offset: LayoutVector2D,
-}
-
-impl<I: Iterator<Item = ComplexClipRegion>> Iterator for ComplexTranslateIter<I> {
-    type Item = ComplexClipRegion;
-    fn next(&mut self) -> Option<Self::Item> {
-        self.source
-            .next()
-            .map(|mut complex| {
-                complex.rect = complex.rect.translate(self.offset);
-                complex
-            })
+impl Default for ClipStore {
+    fn default() -> Self {
+        ClipStore::new()
     }
 }
 
@@ -1858,10 +1956,8 @@ impl ClipItemKind {
             ClipItemKind::BoxShadow { .. } => {
                 false
             }
-            ClipItemKind::RoundedRectangle { ref radius, .. } => {
-                // The rounded clip rect fast path shader can only work
-                // if the radii are uniform.
-                radius.is_uniform().is_some()
+            ClipItemKind::RoundedRectangle { ref rect, ref radius, .. } => {
+                radius.can_use_fast_path_in(rect)
             }
         }
     }
@@ -1885,11 +1981,11 @@ impl ClipItemKind {
 
     fn get_clip_result_complex(
         &self,
-        transform: &LayoutToWorldTransform,
-        prim_world_rect: &WorldRect,
-        world_rect: &WorldRect,
+        transform: &LayoutToVisTransform,
+        prim_rect: &VisRect,
+        culling_rect: &VisRect,
     ) -> ClipResult {
-        let visible_rect = match prim_world_rect.intersection(world_rect) {
+        let visible_rect = match prim_rect.intersection(culling_rect) {
             Some(rect) => rect,
             None => return ClipResult::Reject,
         };
@@ -1924,13 +2020,13 @@ impl ClipItemKind {
                 let outer_clip_rect = match project_rect(
                     transform,
                     &clip_rect,
-                    &world_rect,
+                    &culling_rect,
                 ) {
                     Some(outer_clip_rect) => outer_clip_rect,
                     None => return ClipResult::Partial,
                 };
 
-                match outer_clip_rect.intersection(prim_world_rect) {
+                match outer_clip_rect.intersection(prim_rect) {
                     Some(..) => {
                         ClipResult::Partial
                     }
@@ -2026,23 +2122,6 @@ impl ClipItemKind {
     }
 }
 
-/// Represents a local rect and a device space
-/// rectangles that are either outside or inside bounds.
-#[derive(Clone, Debug, PartialEq)]
-pub struct Geometry {
-    pub local_rect: LayoutRect,
-    pub device_rect: DeviceIntRect,
-}
-
-impl From<LayoutRect> for Geometry {
-    fn from(local_rect: LayoutRect) -> Self {
-        Geometry {
-            local_rect,
-            device_rect: DeviceIntRect::zero(),
-        }
-    }
-}
-
 pub fn rounded_rectangle_contains_point(
     point: &LayoutPoint,
     rect: &LayoutRect,
@@ -2180,8 +2259,8 @@ pub fn polygon_contains_point(
 
 pub fn projected_rect_contains(
     source_rect: &LayoutRect,
-    transform: &LayoutToWorldTransform,
-    target_rect: &WorldRect,
+    transform: &LayoutToVisTransform,
+    target_rect: &VisRect,
 ) -> Option<()> {
     let points = [
         transform.transform_point2d(source_rect.top_left())?,
@@ -2221,6 +2300,7 @@ fn add_clip_node_to_current_chain(
     handle: ClipDataHandle,
     prim_spatial_node_index: SpatialNodeIndex,
     pic_spatial_node_index: SpatialNodeIndex,
+    visibility_spatial_node_index: SpatialNodeIndex,
     local_clip_rect: &mut LayoutRect,
     clip_node_info: &mut Vec<ClipNodeInfo>,
     pic_coverage_rect: &mut PictureRect,
@@ -2234,6 +2314,7 @@ fn add_clip_node_to_current_chain(
     let conversion = ClipSpaceConversion::new(
         prim_spatial_node_index,
         clip_node.item.spatial_node_index,
+        visibility_spatial_node_index,
         spatial_tree,
     );
 
