@@ -14,9 +14,163 @@ use crate::{
     api::units::*, api::ColorDepth, api::ColorF, api::ExternalImageId, api::ImageRendering, api::YuvRangedColorSpace,
     Compositor, CompositorCapabilities, CompositorSurfaceTransform, NativeSurfaceId, NativeSurfaceInfo, NativeTileId,
     profiler, MappableCompositor, SWGLCompositeSurfaceInfo, WindowVisibility,
-    device::Device,
+    device::Device, ClipRadius
 };
 
+// Size (in pixels) of the indirection buffer used for applying rounded rect alpha masks as required
+const INDIRECT_BUFFER_WIDTH: i32 = 64;
+const INDIRECT_BUFFER_HEIGHT: i32 = 64;
+
+// A rounded rect clip in device-space
+#[derive(Debug, Copy, Clone)]
+struct RoundedClip {
+    rect: DeviceIntRect,
+    radii: ClipRadius,
+}
+
+impl RoundedClip {
+    // Construct an empty clip
+    fn zero() -> Self {
+        RoundedClip {
+            rect: DeviceIntRect::zero(),
+            radii: ClipRadius::EMPTY,
+        }
+    }
+
+    // Returns true if this clip has any non-zero corners
+    fn is_valid(&self) -> bool {
+        self.radii != ClipRadius::EMPTY
+    }
+
+    // Returns true if a given rect in device space is affected by this clip
+    fn affects_rect(&self, rect: &DeviceIntRect) -> bool {
+        // If there are no non-zero rounded corners, no clip needed
+        if !self.is_valid() {
+            return false;
+        }
+
+        // Check if any corners where the mask exists are affected by the clip
+        let rect_tl = DeviceIntRect::from_origin_and_size(
+            self.rect.min,
+            DeviceIntSize::new(self.radii.top_left, self.radii.top_left),
+        );
+        if rect_tl.intersects(rect) {
+            return true;
+        }
+
+        let rect_tr = DeviceIntRect::from_origin_and_size(
+            DeviceIntPoint::new(
+                self.rect.max.x - self.radii.top_right,
+                self.rect.min.y,
+            ),
+            DeviceIntSize::new(self.radii.top_right, self.radii.top_right),
+        );
+        if rect_tr.intersects(rect) {
+            return true;
+        }
+
+        let rect_br = DeviceIntRect::from_origin_and_size(
+            DeviceIntPoint::new(
+                self.rect.max.x - self.radii.bottom_right,
+                self.rect.max.y - self.radii.bottom_right,
+            ),
+            DeviceIntSize::new(self.radii.bottom_right, self.radii.bottom_right),
+        );
+        if rect_br.intersects(rect) {
+            return true;
+        }
+
+        let rect_bl = DeviceIntRect::from_origin_and_size(
+            DeviceIntPoint::new(
+                self.rect.min.x,
+                self.rect.max.y - self.radii.bottom_left,
+            ),
+            DeviceIntSize::new(self.radii.bottom_left, self.radii.bottom_left),
+        );
+        if rect_bl.intersects(rect) {
+            return true;
+        }
+
+        // TODO(gw): If a clip is inside the bounds of the surface, this will fail to
+        //           detect that case. It doesn't happen in existing scenarios.
+
+        false
+    }
+}
+
+// Persistent context that is stored per-thread, and available for use by composite
+// jobs as required.
+pub struct SwCompositeJobContext {
+    // Fixed size R8 texture that can be used to write an alpha mask to
+    mask: swgl::LockedResource,
+    // Fixed size RGBA8 texture that can be used as a temporary indirection buffer
+    indirect: swgl::LockedResource,
+}
+
+// Persistent context for mask and indirection buffers, one per thread
+pub struct SwCompositeContext {
+    // Context for the jobs run on the main thread
+    main: SwCompositeJobContext,
+    // Context for the jobs run on the composite thread
+    thread: SwCompositeJobContext,
+}
+
+impl SwCompositeContext {
+    fn new(gl: &swgl::Context) -> Self {
+        SwCompositeContext {
+            main: SwCompositeJobContext::new(gl),
+            thread: SwCompositeJobContext::new(gl),
+        }
+    }
+
+    fn get_job_context(&self, is_composite_thread: bool) -> &SwCompositeJobContext {
+        if is_composite_thread {
+            &self.thread
+        } else {
+            &self.main
+        }
+    }
+}
+
+impl SwCompositeJobContext {
+    // Construct a new per-thread context for sw composite jobs
+    fn new(gl: &swgl::Context) -> Self {
+        let texture_ids = gl.gen_textures(2);
+        let indirect_id = texture_ids[0];
+        let mask_id = texture_ids[1];
+
+        gl.set_texture_buffer(
+            indirect_id,
+            gl::RGBA8,
+            INDIRECT_BUFFER_WIDTH,
+            INDIRECT_BUFFER_HEIGHT,
+            0,
+            ptr::null_mut(),
+            INDIRECT_BUFFER_WIDTH,
+            INDIRECT_BUFFER_HEIGHT,
+        );
+
+        gl.set_texture_buffer(
+            mask_id,
+            gl::R8,
+            INDIRECT_BUFFER_WIDTH,
+            INDIRECT_BUFFER_HEIGHT,
+            0,
+            ptr::null_mut(),
+            INDIRECT_BUFFER_WIDTH,
+            INDIRECT_BUFFER_HEIGHT,
+        );
+
+        let indirect = gl.lock_texture(indirect_id).expect("bug: unable to lock indirect");
+        let mask = gl.lock_texture(mask_id).expect("bug: unable to lock mask");
+
+        SwCompositeJobContext {
+            indirect,
+            mask,
+        }
+    }
+}
+
 pub struct SwTile {
     x: i32,
     y: i32,
@@ -133,6 +287,8 @@ pub struct SwSurface {
     tiles: Vec<SwTile>,
     /// An attached external image for this surface.
     external_image: Option<ExternalImageId>,
+    // The rounded clip that applies to this surface. All corners are zero if not used.
+    rounded_clip: RoundedClip,
 }
 
 impl SwSurface {
@@ -142,6 +298,7 @@ impl SwSurface {
             is_opaque,
             tiles: Vec::new(),
             external_image: None,
+            rounded_clip: RoundedClip::zero(),
         }
     }
 
@@ -166,6 +323,28 @@ impl SwSurface {
         let device_rect = transform.map_rect(&bounds.to_f32()).round_out();
         Some(device_rect.intersection(&clip_rect.to_f32())?.to_i32())
     }
+
+    /// Check that there are no missing tiles in the interior, or rather, that
+    /// the grid of tiles is solidly rectangular.
+    fn has_all_tiles(&self) -> bool {
+        if self.tiles.is_empty() {
+            return false;
+        }
+        // Find the min and max tile ids to identify the tile id bounds.
+        let mut min_x = i32::MAX;
+        let mut min_y = i32::MAX;
+        let mut max_x = i32::MIN;
+        let mut max_y = i32::MIN;
+        for tile in &self.tiles {
+            min_x = min_x.min(tile.x);
+            min_y = min_y.min(tile.y);
+            max_x = max_x.max(tile.x);
+            max_y = max_y.max(tile.y);
+        }
+        // If all tiles are present within the bounds, then the number of tiles
+        // should equal the area of the bounds.
+        (max_x + 1 - min_x) as usize * (max_y + 1 - min_y) as usize == self.tiles.len()
+    }
 }
 
 fn image_rendering_to_gl_filter(filter: ImageRendering) -> gl::GLenum {
@@ -209,11 +388,271 @@ struct SwCompositeJob {
     filter: ImageRendering,
     /// The total number of bands for this job
     num_bands: u8,
+    // The rounded clip that applies to this surface. All corners are zero if not used.
+    rounded_clip: RoundedClip,
+    context: Arc<SwCompositeContext>,
 }
 
 impl SwCompositeJob {
+    // Construct a mask for this job's rounded clip, that is stored in the
+    // shared mask texture of the supplied composite context.
+    fn create_mask(
+        &self,
+        band_clip: &DeviceIntRect,
+        ctx: &SwCompositeJobContext,
+    ) {
+        assert!(band_clip.width() <= INDIRECT_BUFFER_WIDTH);
+        assert!(band_clip.height() <= INDIRECT_BUFFER_HEIGHT);
+
+        // Write mask
+        let (mask_pixels, mask_width, mask_height, _) = ctx.mask.get_buffer();
+        let mask_pixels = unsafe {
+            std::slice::from_raw_parts_mut(
+                mask_pixels as *mut u8,
+                mask_width as usize * mask_height as usize,
+            )
+        };
+
+        // Rounded rect SDF function taken from the existing WR mask shaders.
+        // No doubt this could be done more efficiently, however it typically
+        // is run on only a very small number of pixels, so it's unlikely to
+        // show up in profiles.
+
+        fn sd_round_box(
+            pos: DevicePoint,
+            half_box_size: DeviceSize,
+            radii: &ClipRadius,
+        ) -> f32 {
+            let radius = if pos.x < 0.0 {
+                if pos.y < 0.0 { radii.bottom_right } else { radii.top_right }
+            } else {
+                if pos.y < 0.0 { radii.bottom_left } else { radii.top_left }
+            } as f32;
+
+            let qx = pos.x.abs() - half_box_size.width + radius;
+            let qy = pos.y.abs() - half_box_size.height + radius;
+
+            let qxp = qx.max(0.0);
+            let qyp = qy.max(0.0);
+
+            let d1 = qx.max(qy).min(0.0);
+            let d2 = ((qxp*qxp) + (qyp*qyp)).sqrt();
+
+            d1 + d2 - radius
+        }
+
+        let half_clip_box_size = self.rounded_clip.rect.size().to_f32() * 0.5;
+
+        for y in 0 .. mask_height {
+            let py = band_clip.min.y + y;
+
+            for x in 0 .. mask_width {
+                let px = band_clip.min.x + x;
+
+                let pos = DevicePoint::new(
+                    -0.5 + self.rounded_clip.rect.min.x as f32 + half_clip_box_size.width - px as f32,
+                    -0.5 + self.rounded_clip.rect.min.y as f32 + half_clip_box_size.height - py as f32,
+                );
+
+                let i = (y * mask_width + x) as usize;
+                let d = sd_round_box(
+                    pos,
+                    half_clip_box_size,
+                    &self.rounded_clip.radii,
+                );
+
+                let d = (0.5 - d).min(1.0).max(0.0);
+                mask_pixels[i] = (d * 255.0) as u8;
+            }
+        }
+    }
+
+    // Composite `band_clip` region for the given source (RGBA or YUV), optionally
+    // using an indirection buffer and applying the current alpha mask.
+    fn composite_rect(
+        &self,
+        band_clip: &DeviceIntRect,
+        use_indirect: bool,
+        ctx: &SwCompositeJobContext,
+    ) {
+        match self.locked_src {
+            SwCompositeSource::BGRA(ref resource) => {
+                if use_indirect {
+                    // Copy tile into temporary buffer
+                    ctx.indirect.composite(
+                        resource,
+
+                        self.src_rect.min.x,
+                        self.src_rect.min.y,
+                        self.src_rect.width(),
+                        self.src_rect.height(),
+
+                        -band_clip.min.x + self.dst_rect.min.x,
+                        -band_clip.min.y + self.dst_rect.min.y,
+                        self.dst_rect.width(),
+                        self.dst_rect.height(),
+
+                        true,
+                        self.flip_x,
+                        self.flip_y,
+                        image_rendering_to_gl_filter(self.filter),
+
+                        0,
+                        0,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+
+                    // Apply the mask
+                    ctx.indirect.apply_mask(&ctx.mask);
+
+                    // Composite indirect buffer to frame buffer
+                    self.locked_dst.composite(
+                        &ctx.indirect,
+
+                        0,
+                        0,
+                        band_clip.width(),
+                        band_clip.height(),
+
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+
+                        false,
+                        false,
+                        false,
+                        gl::NEAREST,
+
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+                } else {
+                    self.locked_dst.composite(
+                        resource,
+                        self.src_rect.min.x,
+                        self.src_rect.min.y,
+                        self.src_rect.width(),
+                        self.src_rect.height(),
+                        self.dst_rect.min.x,
+                        self.dst_rect.min.y,
+                        self.dst_rect.width(),
+                        self.dst_rect.height(),
+                        self.opaque,
+                        self.flip_x,
+                        self.flip_y,
+                        image_rendering_to_gl_filter(self.filter),
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+                }
+            }
+            SwCompositeSource::YUV(ref y, ref u, ref v, color_space, color_depth) => {
+                let swgl_color_space = match color_space {
+                    YuvRangedColorSpace::Rec601Narrow => swgl::YuvRangedColorSpace::Rec601Narrow,
+                    YuvRangedColorSpace::Rec601Full => swgl::YuvRangedColorSpace::Rec601Full,
+                    YuvRangedColorSpace::Rec709Narrow => swgl::YuvRangedColorSpace::Rec709Narrow,
+                    YuvRangedColorSpace::Rec709Full => swgl::YuvRangedColorSpace::Rec709Full,
+                    YuvRangedColorSpace::Rec2020Narrow => swgl::YuvRangedColorSpace::Rec2020Narrow,
+                    YuvRangedColorSpace::Rec2020Full => swgl::YuvRangedColorSpace::Rec2020Full,
+                    YuvRangedColorSpace::GbrIdentity => swgl::YuvRangedColorSpace::GbrIdentity,
+                };
+                if use_indirect {
+                    // Copy tile into temporary buffer
+                    ctx.indirect.composite_yuv(
+                        y,
+                        u,
+                        v,
+                        swgl_color_space,
+                        color_depth.bit_depth(),
+
+                        self.src_rect.min.x,
+                        self.src_rect.min.y,
+                        self.src_rect.width(),
+                        self.src_rect.height(),
+
+                        -band_clip.min.x + self.dst_rect.min.x,
+                        -band_clip.min.y + self.dst_rect.min.y,
+                        self.dst_rect.width(),
+                        self.dst_rect.height(),
+
+                        self.flip_x,
+                        self.flip_y,
+
+                        0,
+                        0,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+
+                    // Apply the mask
+                    ctx.indirect.apply_mask(&ctx.mask);
+
+                    // Composite indirect buffer to frame buffer
+                    self.locked_dst.composite(
+                        &ctx.indirect,
+
+                        0,
+                        0,
+                        band_clip.width(),
+                        band_clip.height(),
+
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+
+                        false,
+                        false,
+                        false,
+                        gl::NEAREST,
+
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+                } else {
+                    self.locked_dst.composite_yuv(
+                        y,
+                        u,
+                        v,
+                        swgl_color_space,
+                        color_depth.bit_depth(),
+                        self.src_rect.min.x,
+                        self.src_rect.min.y,
+                        self.src_rect.width(),
+                        self.src_rect.height(),
+                        self.dst_rect.min.x,
+                        self.dst_rect.min.y,
+                        self.dst_rect.width(),
+                        self.dst_rect.height(),
+                        self.flip_x,
+                        self.flip_y,
+                        band_clip.min.x,
+                        band_clip.min.y,
+                        band_clip.width(),
+                        band_clip.height(),
+                    );
+                }
+            }
+        }
+    }
+
     /// Process a composite job
-    fn process(&self, band_index: i32) {
+    fn process(
+        &self,
+        band_index: i32,
+        is_composite_thread: bool,
+    ) {
+        // Retrive the correct context buffers depending on which thread we're on
+        let ctx = self.context.get_job_context(is_composite_thread);
+
         // Bands are allocated in reverse order, but we want to process them in increasing order.
         let num_bands = self.num_bands as i32;
         let band_index = num_bands - 1 - band_index;
@@ -226,60 +665,52 @@ impl SwCompositeJob {
             DeviceIntPoint::new(self.clipped_dst.min.x, self.clipped_dst.min.y + band_offset),
             DeviceIntSize::new(self.clipped_dst.width(), band_height),
         );
-        match self.locked_src {
-            SwCompositeSource::BGRA(ref resource) => {
-                self.locked_dst.composite(
-                    resource,
-                    self.src_rect.min.x,
-                    self.src_rect.min.y,
-                    self.src_rect.width(),
-                    self.src_rect.height(),
-                    self.dst_rect.min.x,
-                    self.dst_rect.min.y,
-                    self.dst_rect.width(),
-                    self.dst_rect.height(),
-                    self.opaque,
-                    self.flip_x,
-                    self.flip_y,
-                    image_rendering_to_gl_filter(self.filter),
-                    band_clip.min.x,
-                    band_clip.min.y,
-                    band_clip.width(),
-                    band_clip.height(),
+
+        // If this band region is affected by a rounded rect clip, apply an alpha mask during compositing
+
+        if self.rounded_clip.affects_rect(&band_clip) {
+            // The job context allocates a small fixed size buffer for indirections, so split this band
+            // in to a number of tiles that can be individually processed.
+
+            let num_x_tiles = (self.clipped_dst.width() + INDIRECT_BUFFER_WIDTH-1) / INDIRECT_BUFFER_WIDTH;
+
+            for x in 0 .. num_x_tiles {
+                let x_offset = (self.clipped_dst.width() * x) / num_x_tiles;
+                let tile_width = (self.clipped_dst.width() * (x + 1)) / num_x_tiles - x_offset;
+
+                let tile_rect = DeviceIntRect::from_origin_and_size(
+                    DeviceIntPoint::new(
+                        self.clipped_dst.min.x + x_offset,
+                        self.clipped_dst.min.y + band_offset,
+                    ),
+                    DeviceIntSize::new(
+                        tile_width,
+                        band_height,
+                    ),
                 );
-            }
-            SwCompositeSource::YUV(ref y, ref u, ref v, color_space, color_depth) => {
-                let swgl_color_space = match color_space {
-                    YuvRangedColorSpace::Rec601Narrow => swgl::YuvRangedColorSpace::Rec601Narrow,
-                    YuvRangedColorSpace::Rec601Full => swgl::YuvRangedColorSpace::Rec601Full,
-                    YuvRangedColorSpace::Rec709Narrow => swgl::YuvRangedColorSpace::Rec709Narrow,
-                    YuvRangedColorSpace::Rec709Full => swgl::YuvRangedColorSpace::Rec709Full,
-                    YuvRangedColorSpace::Rec2020Narrow => swgl::YuvRangedColorSpace::Rec2020Narrow,
-                    YuvRangedColorSpace::Rec2020Full => swgl::YuvRangedColorSpace::Rec2020Full,
-                    YuvRangedColorSpace::GbrIdentity => swgl::YuvRangedColorSpace::GbrIdentity,
-                };
-                self.locked_dst.composite_yuv(
-                    y,
-                    u,
-                    v,
-                    swgl_color_space,
-                    color_depth.bit_depth(),
-                    self.src_rect.min.x,
-                    self.src_rect.min.y,
-                    self.src_rect.width(),
-                    self.src_rect.height(),
-                    self.dst_rect.min.x,
-                    self.dst_rect.min.y,
-                    self.dst_rect.width(),
-                    self.dst_rect.height(),
-                    self.flip_x,
-                    self.flip_y,
-                    band_clip.min.x,
-                    band_clip.min.y,
-                    band_clip.width(),
-                    band_clip.height(),
+
+                // Check if each individual tile within the band is affected by the clip, and
+                // skip indirect buffer (and mask creation) where possible.
+
+                let use_indirect = self.rounded_clip.affects_rect(&tile_rect);
+
+                if use_indirect {
+                    self.create_mask(&tile_rect, ctx);
+                }
+
+                self.composite_rect(
+                    &tile_rect,
+                    use_indirect,
+                    ctx,
                 );
             }
+        } else {
+            // Simple (direct) composite path if no rounded clip
+            self.composite_rect(
+                &band_clip,
+                false,
+                ctx,
+            );
         }
     }
 }
@@ -334,6 +765,8 @@ impl DerefMut for SwCompositeGraphNodeRef {
 struct SwCompositeGraphNode {
     /// Job to be queued for this graph node once ready.
     job: Option<SwCompositeJob>,
+    /// Whether there is a job that requires processing.
+    has_job: AtomicBool,
     /// The number of remaining bands associated with this job. When this is
     /// non-zero and the node has no more parents left, then the node is being
     /// actively used by the composite thread to process jobs. Once it hits
@@ -356,6 +789,7 @@ impl SwCompositeGraphNode {
     fn new() -> SwCompositeGraphNodeRef {
         SwCompositeGraphNodeRef::new(SwCompositeGraphNode {
             job: None,
+            has_job: AtomicBool::new(false),
             remaining_bands: AtomicU8::new(0),
             available_bands: AtomicI8::new(0),
             parents: AtomicU32::new(0),
@@ -366,6 +800,7 @@ impl SwCompositeGraphNode {
     /// Reset the node's state for a new frame
     fn reset(&mut self) {
         self.job = None;
+        self.has_job.store(false, Ordering::SeqCst);
         self.remaining_bands.store(0, Ordering::SeqCst);
         self.available_bands.store(0, Ordering::SeqCst);
         // Initialize parents to 1 as sentinel dependency for uninitialized job
@@ -384,6 +819,7 @@ impl SwCompositeGraphNode {
     /// that would block immediate composition.
     fn set_job(&mut self, job: SwCompositeJob, num_bands: u8) -> bool {
         self.job = Some(job);
+        self.has_job.store(true, Ordering::SeqCst);
         self.remaining_bands.store(num_bands, Ordering::SeqCst);
         self.available_bands.store(num_bands as _, Ordering::SeqCst);
         // Subtract off the sentinel parent dependency now that job is initialized and check
@@ -403,9 +839,13 @@ impl SwCompositeGraphNode {
     }
 
     /// Try to take the job from this node for processing and then process it within the current band.
-    fn process_job(&self, band_index: i32) {
+    fn process_job(
+        &self,
+        band_index: i32,
+        is_composite_thread: bool,
+    ) {
         if let Some(ref job) = self.job {
-            job.process(band_index);
+            job.process(band_index, is_composite_thread);
         }
     }
 
@@ -417,6 +857,8 @@ impl SwCompositeGraphNode {
         }
         // Clear the job to release any locked resources.
         self.job = None;
+        // Signal that resources have been released.
+        self.has_job.store(false, Ordering::SeqCst);
         let mut lock = None;
         for child in self.children.drain(..) {
             // Remove the child's parent dependency on this node. If there are no more
@@ -494,7 +936,7 @@ impl SwCompositeThread {
                 // result when the job queue is dropped, causing the thread
                 // to eventually exit.
                 while let Some((job, band)) = info.take_job(true) {
-                    info.process_job(job, band);
+                    info.process_job(job, band, true);
                 }
                 profiler::unregister_thread();
             })
@@ -512,9 +954,14 @@ impl SwCompositeThread {
     /// Process a job contained in a dependency graph node received from the job queue.
     /// Any child dependencies will be unblocked as appropriate after processing. The
     /// job count will be updated to reflect this.
-    fn process_job(&self, graph_node: &mut SwCompositeGraphNode, band: i32) {
+    fn process_job(
+        &self,
+        graph_node: &mut SwCompositeGraphNode,
+        band: i32,
+        is_composite_thread: bool,
+    ) {
         // Do the actual processing of the job contained in this node.
-        graph_node.process_job(band);
+        graph_node.process_job(band, is_composite_thread);
         // Unblock any child dependencies now that this job has been processed.
         graph_node.unblock_children(self);
     }
@@ -526,26 +973,17 @@ impl SwCompositeThread {
         locked_dst: swgl::LockedResource,
         src_rect: DeviceIntRect,
         dst_rect: DeviceIntRect,
-        clip_rect: DeviceIntRect,
+        clipped_dst: DeviceIntRect,
+        rounded_clip: RoundedClip,
         opaque: bool,
         flip_x: bool,
         flip_y: bool,
         filter: ImageRendering,
+        num_bands: u8,
         mut graph_node: SwCompositeGraphNodeRef,
         job_queue: &mut SwCompositeJobQueue,
+        context: Arc<SwCompositeContext>,
     ) {
-        // For jobs that would span a sufficiently large destination rectangle, split
-        // it into multiple horizontal bands so that multiple threads can process them.
-        let clipped_dst = match dst_rect.intersection(&clip_rect) {
-            Some(clipped_dst) => clipped_dst,
-            None => return,
-        };
-
-        let num_bands = if clipped_dst.width() >= 64 && clipped_dst.height() >= 64 {
-            (clipped_dst.height() / 64).min(4) as u8
-        } else {
-            1
-        };
         let job = SwCompositeJob {
             locked_src,
             locked_dst,
@@ -557,6 +995,8 @@ impl SwCompositeThread {
             flip_y,
             filter,
             num_bands,
+            rounded_clip,
+            context,
         };
         if graph_node.set_job(job, num_bands) {
             self.send_job(job_queue, graph_node);
@@ -681,7 +1121,7 @@ impl SwCompositeThread {
         // thread if it is busy.
         if !sync {
             while let Some((job, band)) = self.take_job(false) {
-                self.process_job(job, band);
+                self.process_job(job, band, false);
             }
             // Once there are no more jobs, just fall through to waiting
             // synchronously for the composite thread to finish processing.
@@ -742,6 +1182,8 @@ pub struct SwCompositor {
     composite_thread: Option<Arc<SwCompositeThread>>,
     /// SWGL locked resource for sharing framebuffer with SwComposite thread
     locked_framebuffer: Option<swgl::LockedResource>,
+    /// Per-thread buffers used for rendering masks and indirection buffers
+    composite_context: Option<Arc<SwCompositeContext>>,
     /// Whether we are currently in the middle of compositing
     is_compositing: bool,
 }
@@ -755,10 +1197,16 @@ impl SwCompositor {
         // Only create the SwComposite thread if we're not using a native render
         // compositor. Thus, we are compositing into the main software framebuffer,
         // which benefits from compositing asynchronously while updating tiles.
-        let composite_thread = if !use_native_compositor {
-            Some(SwCompositeThread::new())
+        let (composite_thread, composite_context) = if !use_native_compositor {
+            (
+                Some(SwCompositeThread::new()),
+                Some(Arc::new(SwCompositeContext::new(&gl)))
+            )
         } else {
-            None
+            (
+                None,
+                None,
+            )
         };
         SwCompositor {
             gl,
@@ -777,6 +1225,7 @@ impl SwCompositor {
             depth_id: None,
             composite_thread,
             locked_framebuffer: None,
+            composite_context,
             is_compositing: false,
         }
     }
@@ -830,6 +1279,16 @@ impl SwCompositor {
             base.start.min(extra.start)..base.end.max(extra.end)
         }
 
+        // Ensure an occluder surface is both opaque and has all interior tiles.
+        fn valid_occluder(surface: &SwSurface) -> bool {
+            surface.is_opaque &&
+            surface.has_all_tiles() &&
+            // TODO(gw): Skipping an entire surface as an occluder when it has
+            //           a rounded rect is probably too costly. May need to
+            //           just skip tiles or bands from being added as occluders.
+            !surface.rounded_clip.is_valid()
+        }
+
         // Before we can try to occlude any surfaces, we need to fix their clip rects to tightly
         // bound the valid region. The clip rect might otherwise enclose an invalid area that
         // can't fully occlude anything even if the surface is opaque.
@@ -845,7 +1304,7 @@ impl SwCompositor {
         for occlude_index in 0..self.frame_surfaces.len() {
             let (ref occlude_id, _, ref occlude_rect, _) = self.frame_surfaces[occlude_index];
             match self.surfaces.get(occlude_id) {
-                Some(occluder) if occluder.is_opaque && !occlude_rect.is_empty() => {}
+                Some(occluder) if valid_occluder(occluder) && !occlude_rect.is_empty() => {}
                 _ => continue,
             }
 
@@ -866,14 +1325,14 @@ impl SwCompositor {
                     if includes(&occlude_x, &clip_x) {
                         if let Some(visible) = overlaps(&occlude_y, &clip_y) {
                             set_y_range(clip_rect, &visible);
-                            if surface.is_opaque && occlude_x == clip_x {
+                            if occlude_x == clip_x && valid_occluder(surface) {
                                 occlude_y = union(occlude_y, visible);
                             }
                         }
                     } else if includes(&occlude_y, &clip_y) {
                         if let Some(visible) = overlaps(&occlude_x, &clip_x) {
                             set_x_range(clip_rect, &visible);
-                            if surface.is_opaque && occlude_y == clip_y {
+                            if occlude_y == clip_y && valid_occluder(surface) {
                                 occlude_x = union(occlude_x, visible);
                             }
                         }
@@ -1003,19 +1462,35 @@ impl SwCompositor {
                     return;
                 };
                 if let Some(ref framebuffer) = self.locked_framebuffer {
-                    composite_thread.queue_composite(
-                        source,
-                        framebuffer.clone(),
-                        src_rect,
-                        dst_rect,
-                        *clip_rect,
-                        surface.is_opaque,
-                        flip_x,
-                        flip_y,
-                        filter,
-                        tile.graph_node.clone(),
-                        job_queue,
-                    );
+                    if let Some(clipped_dst) = dst_rect.intersection(clip_rect) {
+                        let num_bands = if surface.rounded_clip.affects_rect(&clipped_dst) {
+                            // Create enough bands that we won't exceed the height of the indirection buffer.
+                            ((clipped_dst.height() + INDIRECT_BUFFER_HEIGHT-1) / INDIRECT_BUFFER_HEIGHT) as u8
+                        } else if clipped_dst.width() >= 64 && clipped_dst.height() >= 64 {
+                            // For jobs that would span a sufficiently large destination rectangle, split
+                            // it into multiple horizontal bands so that multiple threads can process them.
+                            (clipped_dst.height() / 64).min(4) as u8
+                        } else {
+                            1
+                        };
+
+                        composite_thread.queue_composite(
+                            source,
+                            framebuffer.clone(),
+                            src_rect,
+                            dst_rect,
+                            clipped_dst,
+                            surface.rounded_clip,
+                            surface.is_opaque,
+                            flip_x,
+                            flip_y,
+                            filter,
+                            num_bands,
+                            tile.graph_node.clone(),
+                            job_queue,
+                            self.composite_context.as_ref().expect("bug").clone(),
+                        );
+                    }
                 }
             }
         }
@@ -1198,6 +1673,11 @@ impl Compositor for SwCompositor {
             composite_thread.deinit();
         }
 
+        // Ensure we drop the last remaining composite context so that the
+        // locked textures are dropped before we try to drop the SWGL context
+        // in the parent caller
+        self.composite_context = None;
+
         for surface in self.surfaces.values() {
             self.deinit_surface(surface);
         }
@@ -1305,6 +1785,12 @@ impl Compositor for SwCompositor {
                         stride = tile_info.stride;
                         buf = tile_info.data;
                     }
+                } else if let Some(ref composite_thread) = self.composite_thread {
+                    // Check if the tile is currently in use before proceeding to modify it.
+                    if tile.graph_node.get().has_job.load(Ordering::SeqCst) {
+                        // Need to wait for the SwComposite thread to finish any queued jobs.
+                        composite_thread.wait_for_composites(false);
+                    }
                 }
                 self.gl.set_texture_buffer(
                     tile.color_id,
@@ -1384,9 +1870,26 @@ impl Compositor for SwCompositor {
         transform: CompositorSurfaceTransform,
         clip_rect: DeviceIntRect,
         filter: ImageRendering,
+        rounded_clip_rect: DeviceIntRect,
+        rounded_clip_radii: ClipRadius,
     ) {
+        // Update the rounded clip on the surface
+        let surface = self.surfaces.get_mut(&id).expect("bug: unknown surface");
+        surface.rounded_clip = RoundedClip {
+            rect: rounded_clip_rect,
+            radii: rounded_clip_radii,
+        };
+
         if self.use_native_compositor {
-            self.compositor.add_surface(device, id, transform, clip_rect, filter);
+            self.compositor.add_surface(
+                device,
+                id,
+                transform,
+                clip_rect,
+                filter,
+                rounded_clip_rect,
+                rounded_clip_radii,
+            );
         }
 
         if self.composite_thread.is_some() {
