@@ -4,10 +4,6 @@
 
 "use strict";
 
-const { Preferences } = ChromeUtils.importESModule(
-  "resource://gre/modules/Preferences.sys.mjs"
-);
-
 const { AppInfo } = ChromeUtils.importESModule(
   "chrome://remote/content/shared/AppInfo.sys.mjs"
 );
@@ -19,9 +15,8 @@ const {
   mergeCapabilities,
   PageLoadStrategy,
   processCapabilities,
-  Proxy,
+  ProxyConfiguration,
   Timeouts,
-  UnhandledPromptBehavior,
   validateCapabilities,
 } = ChromeUtils.importESModule(
   "chrome://remote/content/shared/webdriver/Capabilities.sys.mjs"
@@ -55,7 +50,7 @@ add_task(function test_Timeouts_fromJSON() {
   equal(ts.script, json.script);
 });
 
-add_task(function test_Timeouts_fromJSON_unrecognised_field() {
+add_task(function test_Timeouts_fromJSON_unrecognized_field() {
   let json = {
     sessionId: "foobar",
   };
@@ -63,7 +58,7 @@ add_task(function test_Timeouts_fromJSON_unrecognised_field() {
     Timeouts.fromJSON(json);
   } catch (e) {
     equal(e.name, error.InvalidArgumentError.name);
-    equal(e.message, "Unrecognised timeout: sessionId");
+    equal(e.message, "Unrecognized timeout: sessionId");
   }
 });
 
@@ -92,7 +87,7 @@ add_task(function test_PageLoadStrategy() {
 });
 
 add_task(function test_Proxy_ctor() {
-  let p = new Proxy();
+  let p = new ProxyConfiguration();
   let props = [
     "proxyType",
     "httpProxy",
@@ -108,44 +103,44 @@ add_task(function test_Proxy_ctor() {
 });
 
 add_task(function test_Proxy_init() {
-  let p = new Proxy();
+  let p = new ProxyConfiguration();
 
   // no changed made, and 5 (system) is default
   equal(p.init(), false);
-  equal(Preferences.get("network.proxy.type"), 5);
+  equal(Services.prefs.getIntPref("network.proxy.type"), 5);
 
   // pac
   p.proxyType = "pac";
   p.proxyAutoconfigUrl = "http://localhost:1234";
   ok(p.init());
 
-  equal(Preferences.get("network.proxy.type"), 2);
+  equal(Services.prefs.getIntPref("network.proxy.type"), 2);
   equal(
-    Preferences.get("network.proxy.autoconfig_url"),
+    Services.prefs.getStringPref("network.proxy.autoconfig_url"),
     "http://localhost:1234"
   );
 
   // direct
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "direct";
   ok(p.init());
-  equal(Preferences.get("network.proxy.type"), 0);
+  equal(Services.prefs.getIntPref("network.proxy.type"), 0);
 
   // autodetect
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "autodetect";
   ok(p.init());
-  equal(Preferences.get("network.proxy.type"), 4);
+  equal(Services.prefs.getIntPref("network.proxy.type"), 4);
 
   // system
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "system";
   ok(p.init());
-  equal(Preferences.get("network.proxy.type"), 5);
+  equal(Services.prefs.getIntPref("network.proxy.type"), 5);
 
   // manual
   for (let proxy of ["http", "ssl", "socks"]) {
-    p = new Proxy();
+    p = new ProxyConfiguration();
     p.proxyType = "manual";
     p.noProxy = ["foo", "bar"];
     p[`${proxy}Proxy`] = "foo";
@@ -155,46 +150,49 @@ add_task(function test_Proxy_init() {
     }
 
     ok(p.init());
-    equal(Preferences.get("network.proxy.type"), 1);
-    equal(Preferences.get("network.proxy.no_proxies_on"), "foo, bar");
-    equal(Preferences.get(`network.proxy.${proxy}`), "foo");
-    equal(Preferences.get(`network.proxy.${proxy}_port`), 42);
+    equal(Services.prefs.getIntPref("network.proxy.type"), 1);
+    equal(
+      Services.prefs.getStringPref("network.proxy.no_proxies_on"),
+      "foo, bar"
+    );
+    equal(Services.prefs.getStringPref(`network.proxy.${proxy}`), "foo");
+    equal(Services.prefs.getIntPref(`network.proxy.${proxy}_port`), 42);
     if (proxy === "socks") {
-      equal(Preferences.get(`network.proxy.${proxy}_version`), 4);
+      equal(Services.prefs.getIntPref(`network.proxy.${proxy}_version`), 4);
     }
   }
 
   // empty no proxy should reset default exclustions
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
   p.noProxy = [];
   ok(p.init());
-  equal(Preferences.get("network.proxy.no_proxies_on"), "");
+  equal(Services.prefs.getStringPref("network.proxy.no_proxies_on"), "");
 });
 
 add_task(function test_Proxy_toString() {
-  equal(new Proxy().toString(), "[object Proxy]");
+  equal(new ProxyConfiguration().toString(), "[object Proxy]");
 });
 
 add_task(function test_Proxy_toJSON() {
-  let p = new Proxy();
+  let p = new ProxyConfiguration();
   deepEqual(p.toJSON(), {});
 
   // autoconfig url
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "pac";
   p.proxyAutoconfigUrl = "foo";
   deepEqual(p.toJSON(), { proxyType: "pac", proxyAutoconfigUrl: "foo" });
 
   // manual proxy
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
   deepEqual(p.toJSON(), { proxyType: "manual" });
 
   for (let proxy of ["httpProxy", "sslProxy", "socksProxy"]) {
     let expected = { proxyType: "manual" };
 
-    p = new Proxy();
+    p = new ProxyConfiguration();
     p.proxyType = "manual";
 
     if (proxy == "socksProxy") {
@@ -226,7 +224,7 @@ add_task(function test_Proxy_toJSON() {
   }
 
   // noProxy: add brackets for IPv6 address
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
   p.noProxy = ["2001:db8::1"];
   let expected = { proxyType: "manual", noProxy: "[2001:db8::1]" };
@@ -234,38 +232,48 @@ add_task(function test_Proxy_toJSON() {
 });
 
 add_task(function test_Proxy_fromJSON() {
-  let p = new Proxy();
-  deepEqual(p, Proxy.fromJSON(undefined));
-  deepEqual(p, Proxy.fromJSON(null));
+  let p = new ProxyConfiguration();
+  deepEqual(p, ProxyConfiguration.fromJSON(undefined));
+  deepEqual(p, ProxyConfiguration.fromJSON(null));
 
   for (let typ of [true, 42, "foo", []]) {
-    Assert.throws(() => Proxy.fromJSON(typ), /InvalidArgumentError/);
+    Assert.throws(
+      () => ProxyConfiguration.fromJSON(typ),
+      /InvalidArgumentError/
+    );
   }
 
   // must contain a valid proxyType
-  Assert.throws(() => Proxy.fromJSON({}), /InvalidArgumentError/);
+  Assert.throws(() => ProxyConfiguration.fromJSON({}), /InvalidArgumentError/);
   Assert.throws(
-    () => Proxy.fromJSON({ proxyType: "foo" }),
+    () => ProxyConfiguration.fromJSON({ proxyType: "foo" }),
     /InvalidArgumentError/
   );
 
   // autoconfig url
   for (let url of [true, 42, [], {}]) {
     Assert.throws(
-      () => Proxy.fromJSON({ proxyType: "pac", proxyAutoconfigUrl: url }),
+      () =>
+        ProxyConfiguration.fromJSON({
+          proxyType: "pac",
+          proxyAutoconfigUrl: url,
+        }),
       /InvalidArgumentError/
     );
   }
 
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "pac";
   p.proxyAutoconfigUrl = "foo";
-  deepEqual(p, Proxy.fromJSON({ proxyType: "pac", proxyAutoconfigUrl: "foo" }));
+  deepEqual(
+    p,
+    ProxyConfiguration.fromJSON({ proxyType: "pac", proxyAutoconfigUrl: "foo" })
+  );
 
   // manual proxy
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
-  deepEqual(p, Proxy.fromJSON({ proxyType: "manual" }));
+  deepEqual(p, ProxyConfiguration.fromJSON({ proxyType: "manual" }));
 
   for (let proxy of ["httpProxy", "sslProxy", "socksProxy"]) {
     let manual = { proxyType: "manual" };
@@ -286,10 +294,13 @@ add_task(function test_Proxy_fromJSON() {
       "2001:db8::1",
     ]) {
       manual[proxy] = host;
-      Assert.throws(() => Proxy.fromJSON(manual), /InvalidArgumentError/);
+      Assert.throws(
+        () => ProxyConfiguration.fromJSON(manual),
+        /InvalidArgumentError/
+      );
     }
 
-    p = new Proxy();
+    p = new ProxyConfiguration();
     p.proxyType = "manual";
     if (proxy == "socksProxy") {
       manual.socksVersion = 5;
@@ -313,7 +324,7 @@ add_task(function test_Proxy_fromJSON() {
       p[`${proxy}`] = host_map[host].hostname;
       p[`${proxy}Port`] = host_map[host].port;
 
-      deepEqual(p, Proxy.fromJSON(manual));
+      deepEqual(p, ProxyConfiguration.fromJSON(manual));
     }
 
     // Without a port the default port of the scheme is used
@@ -330,57 +341,60 @@ add_task(function test_Proxy_fromJSON() {
         p[`${proxy}Port`] = default_ports[proxy];
       }
 
-      deepEqual(p, Proxy.fromJSON(manual));
+      deepEqual(p, ProxyConfiguration.fromJSON(manual));
     }
   }
 
   // missing required socks version
   Assert.throws(
-    () => Proxy.fromJSON({ proxyType: "manual", socksProxy: "foo:1234" }),
+    () =>
+      ProxyConfiguration.fromJSON({
+        proxyType: "manual",
+        socksProxy: "foo:1234",
+      }),
     /InvalidArgumentError/
   );
 
-  // Bug 1703805: Since Firefox 90 ftpProxy is no longer supported
+  // missing required socks proxy
   Assert.throws(
-    () => Proxy.fromJSON({ proxyType: "manual", ftpProxy: "foo:21" }),
+    () => ProxyConfiguration.fromJSON({ proxyType: "manual", socksVersion: 4 }),
     /InvalidArgumentError/
   );
 
   // noProxy: invalid settings
   for (let noProxy of [true, 42, {}, null, "foo", [true], [42], [{}], [null]]) {
     Assert.throws(
-      () => Proxy.fromJSON({ proxyType: "manual", noProxy }),
+      () => ProxyConfiguration.fromJSON({ proxyType: "manual", noProxy }),
       /InvalidArgumentError/
     );
   }
 
   // noProxy: valid settings
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
   for (let noProxy of [[], ["foo"], ["foo", "bar"], ["127.0.0.1"]]) {
     let manual = { proxyType: "manual", noProxy };
     p.noProxy = noProxy;
-    deepEqual(p, Proxy.fromJSON(manual));
+    deepEqual(p, ProxyConfiguration.fromJSON(manual));
   }
 
   // noProxy: IPv6 needs brackets removed
-  p = new Proxy();
+  p = new ProxyConfiguration();
   p.proxyType = "manual";
   p.noProxy = ["2001:db8::1"];
   let manual = { proxyType: "manual", noProxy: ["[2001:db8::1]"] };
-  deepEqual(p, Proxy.fromJSON(manual));
+  deepEqual(p, ProxyConfiguration.fromJSON(manual));
 });
 
-add_task(function test_UnhandledPromptBehavior() {
-  equal(UnhandledPromptBehavior.Accept, "accept");
-  equal(UnhandledPromptBehavior.AcceptAndNotify, "accept and notify");
-  equal(UnhandledPromptBehavior.Dismiss, "dismiss");
-  equal(UnhandledPromptBehavior.DismissAndNotify, "dismiss and notify");
-  equal(UnhandledPromptBehavior.Ignore, "ignore");
+add_task(function test_Capabilities_ctor_http_default() {
+  const caps = new Capabilities();
+
+  equal(true, caps.get("moz:webdriverClick"));
 });
 
-add_task(function test_Capabilities_ctor() {
-  let caps = new Capabilities();
+add_task(function test_Capabilities_ctor_http() {
+  const caps = new Capabilities(false);
+
   ok(caps.has("browserName"));
   ok(caps.has("browserVersion"));
   ok(caps.has("platformName"));
@@ -388,21 +402,40 @@ add_task(function test_Capabilities_ctor() {
   equal(PageLoadStrategy.Normal, caps.get("pageLoadStrategy"));
   equal(false, caps.get("acceptInsecureCerts"));
   ok(caps.get("timeouts") instanceof Timeouts);
-  ok(caps.get("proxy") instanceof Proxy);
+  ok(caps.get("proxy") instanceof ProxyConfiguration);
   equal(caps.get("setWindowRect"), !AppInfo.isAndroid);
   equal(caps.get("strictFileInteractability"), false);
   equal(caps.get("webSocketUrl"), null);
 
   equal(false, caps.get("moz:accessibilityChecks"));
   ok(caps.has("moz:buildID"));
-  ok(caps.has("moz:debuggerAddress"));
   ok(caps.has("moz:platformVersion"));
   ok(caps.has("moz:processID"));
   ok(caps.has("moz:profile"));
   equal(true, caps.get("moz:webdriverClick"));
+});
 
-  // No longer supported capabilities
-  ok(!caps.has("moz:useNonSpecCompliantPointerOrigin"));
+add_task(function test_Capabilities_ctor_bidi() {
+  const caps = new Capabilities(true);
+
+  ok(caps.has("browserName"));
+  ok(caps.has("browserVersion"));
+  ok(caps.has("platformName"));
+  ok(["linux", "mac", "windows", "android"].includes(caps.get("platformName")));
+  equal(undefined, caps.get("pageLoadStrategy"));
+  equal(false, caps.get("acceptInsecureCerts"));
+  ok(!caps.has("timeouts"));
+  ok(caps.get("proxy") instanceof ProxyConfiguration);
+  ok(!caps.has("setWindowRect"));
+  ok(!caps.has("strictFileInteractability"));
+  ok(!caps.has("webSocketUrl"));
+
+  ok(!caps.has("moz:accessibilityChecks"));
+  ok(caps.has("moz:buildID"));
+  ok(caps.has("moz:platformVersion"));
+  ok(caps.has("moz:processID"));
+  ok(caps.has("moz:profile"));
+  ok(!caps.has("moz:webdriverClick"));
 });
 
 add_task(function test_Capabilities_toString() {
@@ -426,122 +459,146 @@ add_task(function test_Capabilities_toJSON() {
 
   equal(caps.get("moz:accessibilityChecks"), json["moz:accessibilityChecks"]);
   equal(caps.get("moz:buildID"), json["moz:buildID"]);
-  equal(caps.get("moz:debuggerAddress"), json["moz:debuggerAddress"]);
   equal(caps.get("moz:platformVersion"), json["moz:platformVersion"]);
   equal(caps.get("moz:processID"), json["moz:processID"]);
   equal(caps.get("moz:profile"), json["moz:profile"]);
   equal(caps.get("moz:webdriverClick"), json["moz:webdriverClick"]);
 });
 
-add_task(function test_Capabilities_fromJSON() {
+add_task(function test_Capabilities_fromJSON_http() {
   const { fromJSON } = Capabilities;
 
   // plain
-  for (let typ of [{}, null, undefined]) {
-    ok(fromJSON(typ).has("browserName"));
+  for (const type of [{}, null, undefined]) {
+    ok(fromJSON(type, false).has("browserName"));
   }
 
-  // matching
-  let caps = new Capabilities();
+  let caps;
 
-  caps = fromJSON({ acceptInsecureCerts: true });
+  // Capabilities supported by HTTP and BiDi
+  caps = fromJSON({ acceptInsecureCerts: true }, false);
   equal(true, caps.get("acceptInsecureCerts"));
-  caps = fromJSON({ acceptInsecureCerts: false });
-  equal(false, caps.get("acceptInsecureCerts"));
 
+  let proxyConfig = { proxyType: "manual" };
+  caps = fromJSON({ proxy: proxyConfig }, false);
+  equal("manual", caps.get("proxy").proxyType);
+
+  // WebDriver HTTP-only capabilities
   for (let strategy of Object.values(PageLoadStrategy)) {
-    caps = fromJSON({ pageLoadStrategy: strategy });
+    caps = fromJSON({ pageLoadStrategy: strategy }, false);
     equal(strategy, caps.get("pageLoadStrategy"));
   }
 
-  let proxyConfig = { proxyType: "manual" };
-  caps = fromJSON({ proxy: proxyConfig });
-  equal("manual", caps.get("proxy").proxyType);
-
   let timeoutsConfig = { implicit: 123 };
-  caps = fromJSON({ timeouts: timeoutsConfig });
+  caps = fromJSON({ timeouts: timeoutsConfig }, false);
   equal(123, caps.get("timeouts").implicit);
 
-  caps = fromJSON({ strictFileInteractability: false });
-  equal(false, caps.get("strictFileInteractability"));
-  caps = fromJSON({ strictFileInteractability: true });
+  caps = fromJSON({ strictFileInteractability: true }, false);
   equal(true, caps.get("strictFileInteractability"));
 
-  caps = fromJSON({ webSocketUrl: true });
+  caps = fromJSON({ webSocketUrl: true }, false);
   equal(true, caps.get("webSocketUrl"));
 
-  caps = fromJSON({ "webauthn:virtualAuthenticators": true });
-  equal(true, caps.get("webauthn:virtualAuthenticators"));
-  caps = fromJSON({ "webauthn:virtualAuthenticators": false });
-  equal(false, caps.get("webauthn:virtualAuthenticators"));
-  Assert.throws(
-    () => fromJSON({ "webauthn:virtualAuthenticators": "foo" }),
-    /InvalidArgumentError/
-  );
-
-  caps = fromJSON({ "webauthn:extension:uvm": true });
-  equal(true, caps.get("webauthn:extension:uvm"));
-  caps = fromJSON({ "webauthn:extension:uvm": false });
-  equal(false, caps.get("webauthn:extension:uvm"));
-  Assert.throws(
-    () => fromJSON({ "webauthn:extension:uvm": "foo" }),
-    /InvalidArgumentError/
-  );
-
-  caps = fromJSON({ "webauthn:extension:prf": true });
-  equal(true, caps.get("webauthn:extension:prf"));
-  caps = fromJSON({ "webauthn:extension:prf": false });
-  equal(false, caps.get("webauthn:extension:prf"));
-  Assert.throws(
-    () => fromJSON({ "webauthn:extension:prf": "foo" }),
-    /InvalidArgumentError/
-  );
-
-  caps = fromJSON({ "webauthn:extension:largeBlob": true });
-  equal(true, caps.get("webauthn:extension:largeBlob"));
-  caps = fromJSON({ "webauthn:extension:largeBlob": false });
-  equal(false, caps.get("webauthn:extension:largeBlob"));
-  Assert.throws(
-    () => fromJSON({ "webauthn:extension:largeBlob": "foo" }),
-    /InvalidArgumentError/
-  );
-
-  caps = fromJSON({ "webauthn:extension:credBlob": true });
-  equal(true, caps.get("webauthn:extension:credBlob"));
-  caps = fromJSON({ "webauthn:extension:credBlob": false });
-  equal(false, caps.get("webauthn:extension:credBlob"));
-  Assert.throws(
-    () => fromJSON({ "webauthn:extension:credBlob": "foo" }),
-    /InvalidArgumentError/
-  );
-
-  caps = fromJSON({ "moz:accessibilityChecks": true });
+  // Mozilla specific capabilities
+  caps = fromJSON({ "moz:accessibilityChecks": true }, false);
   equal(true, caps.get("moz:accessibilityChecks"));
-  caps = fromJSON({ "moz:accessibilityChecks": false });
-  equal(false, caps.get("moz:accessibilityChecks"));
 
-  // capability is always populated with null if remote agent is not listening
-  caps = fromJSON({});
-  equal(null, caps.get("moz:debuggerAddress"));
-  caps = fromJSON({ "moz:debuggerAddress": "foo" });
-  equal(null, caps.get("moz:debuggerAddress"));
-  caps = fromJSON({ "moz:debuggerAddress": true });
-  equal(null, caps.get("moz:debuggerAddress"));
-
-  caps = fromJSON({ "moz:webdriverClick": true });
+  caps = fromJSON({ "moz:webdriverClick": true }, false);
   equal(true, caps.get("moz:webdriverClick"));
-  caps = fromJSON({ "moz:webdriverClick": false });
-  equal(false, caps.get("moz:webdriverClick"));
 
-  // No longer supported capabilities
+  // Extension capabilities
+  caps = fromJSON({ "webauthn:virtualAuthenticators": true }, false);
+  equal(true, caps.get("webauthn:virtualAuthenticators"));
   Assert.throws(
-    () => fromJSON({ "moz:useNonSpecCompliantPointerOrigin": false }),
+    () => fromJSON({ "webauthn:virtualAuthenticators": "foo" }, false),
     /InvalidArgumentError/
   );
+
+  caps = fromJSON({ "webauthn:extension:uvm": true }, false);
+  equal(true, caps.get("webauthn:extension:uvm"));
   Assert.throws(
-    () => fromJSON({ "moz:useNonSpecCompliantPointerOrigin": true }),
+    () => fromJSON({ "webauthn:extension:uvm": "foo" }, false),
     /InvalidArgumentError/
   );
+
+  caps = fromJSON({ "webauthn:extension:prf": true }, false);
+  equal(true, caps.get("webauthn:extension:prf"));
+  Assert.throws(
+    () => fromJSON({ "webauthn:extension:prf": "foo" }, false),
+    /InvalidArgumentError/
+  );
+
+  caps = fromJSON({ "webauthn:extension:largeBlob": true }, false);
+  equal(true, caps.get("webauthn:extension:largeBlob"));
+  Assert.throws(
+    () => fromJSON({ "webauthn:extension:largeBlob": "foo" }, false),
+    /InvalidArgumentError/
+  );
+
+  caps = fromJSON({ "webauthn:extension:credBlob": true }, false);
+  equal(true, caps.get("webauthn:extension:credBlob"));
+  Assert.throws(
+    () => fromJSON({ "webauthn:extension:credBlob": "foo" }, false),
+    /InvalidArgumentError/
+  );
+});
+
+add_task(function test_Capabilities_fromJSON_bidi() {
+  const { fromJSON } = Capabilities;
+
+  // plain
+  for (const type of [{}, null, undefined]) {
+    ok(fromJSON(type, true).has("browserName"));
+  }
+
+  let caps;
+
+  // Capabilities supported by HTTP and BiDi
+  caps = fromJSON({ acceptInsecureCerts: true }, true);
+  equal(true, caps.get("acceptInsecureCerts"));
+
+  let proxyConfig = { proxyType: "manual" };
+  caps = fromJSON({ proxy: proxyConfig }, true);
+  equal("manual", caps.get("proxy").proxyType);
+
+  // HTTP capabilities are ignored for BiDi-only sessions
+  for (let strategy of Object.values(PageLoadStrategy)) {
+    caps = fromJSON({ pageLoadStrategy: strategy }, true);
+    ok(!caps.has("pageLoadStrategy"));
+  }
+
+  let timeoutsConfig = { implicit: 123 };
+  caps = fromJSON({ timeouts: timeoutsConfig }, true);
+  ok(!caps.has("timeouts"));
+
+  caps = fromJSON({ strictFileInteractability: true }, true);
+  ok(!caps.has("strictFileInteractability"));
+
+  caps = fromJSON({ webSocketUrl: true }, true);
+  ok(!caps.has("webSocketUrl"));
+
+  // Mozilla specific capabilities
+  caps = fromJSON({ "moz:accessibilityChecks": true }, true);
+  ok(!caps.has("moz:accessibilityChecks"));
+
+  caps = fromJSON({ "moz:webdriverClick": true }, true);
+  equal(undefined, caps.get("moz:webdriverClick"));
+
+  // Extension capabilities
+  caps = fromJSON({ "webauthn:virtualAuthenticators": true }, true);
+  ok(!caps.has("webauthn:virtualAuthenticators"));
+
+  caps = fromJSON({ "webauthn:extension:uvm": true }, true);
+  ok(!caps.has("webauthn:extension:uvm"));
+
+  caps = fromJSON({ "webauthn:extension:prf": true }, true);
+  ok(!caps.has("webauthn:extension:prf"));
+
+  caps = fromJSON({ "webauthn:extension:largeBlob": true }, true);
+  ok(!caps.has("webauthn:extension:largeBlob"));
+
+  caps = fromJSON({ "webauthn:extension:credBlob": true }, true);
+  ok(!caps.has("webauthn:extension:credBlob"));
 });
 
 add_task(function test_mergeCapabilities() {
@@ -586,8 +643,6 @@ add_task(function test_validateCapabilities_invalid() {
     { "moz:accessibilityChecks": "foo" },
     { "moz:webdriverClick": "foo" },
     { "moz:webdriverClick": 1 },
-    { "moz:useNonSpecCompliantPointerOrigin": false },
-    { "moz:debuggerAddress": "foo" },
     { "moz:someRandomString": {} },
   ];
   for (const capabilities of invalidCapabilities) {
@@ -616,7 +671,6 @@ add_task(function test_validateCapabilities_valid() {
     { "moz:firefoxOptions": {} },
     { "moz:accessibilityChecks": true },
     { "moz:webdriverClick": true },
-    { "moz:debuggerAddress": true },
     { "test:extension": "foo" },
   ];
   for (const validCapability of validCapabilities) {
@@ -672,7 +726,7 @@ add_task(function test_processCapabilities() {
 
 // use Proxy.toJSON to test marshal
 add_task(function test_marshal() {
-  let proxy = new Proxy();
+  let proxy = new ProxyConfiguration();
 
   // drop empty fields
   deepEqual({}, proxy.toJSON());
@@ -688,7 +742,7 @@ add_task(function test_marshal() {
   deepEqual({ proxyType: { foo: "bar" } }, proxy.toJSON());
 
   // iterate over complex object that implement toJSON
-  proxy.proxyType = new Proxy();
+  proxy.proxyType = new ProxyConfiguration();
   deepEqual({}, proxy.toJSON());
   proxy.proxyType.proxyType = "manual";
   deepEqual({ proxyType: { proxyType: "manual" } }, proxy.toJSON());
@@ -696,6 +750,6 @@ add_task(function test_marshal() {
   // drop objects with no entries
   proxy.proxyType = { foo: {} };
   deepEqual({}, proxy.toJSON());
-  proxy.proxyType = { foo: new Proxy() };
+  proxy.proxyType = { foo: new ProxyConfiguration() };
   deepEqual({}, proxy.toJSON());
 });
