@@ -1,23 +1,13 @@
 /**
- * Copyright 2018 Google Inc. All rights reserved.
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
- *
- *     http://www.apache.org/licenses/LICENSE-2.0
- *
- * Unless required by applicable law or agreed to in writing, software
- * distributed under the License is distributed on an "AS IS" BASIS,
- * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
- * See the License for the specific language governing permissions and
- * limitations under the License.
+ * @license
+ * Copyright 2018 Google Inc.
+ * SPDX-License-Identifier: Apache-2.0
  */
-import assert from 'assert';
+import assert from 'node:assert';
 
 import expect from 'expect';
 import {Puppeteer} from 'puppeteer-core';
-import {ElementHandle} from 'puppeteer-core/internal/api/ElementHandle.js';
+import type {ElementHandle} from 'puppeteer-core/internal/api/ElementHandle.js';
 
 import {getTestState, setupTestBrowserHooks} from './mocha-utils.js';
 
@@ -40,13 +30,13 @@ describe('Query handler tests', function () {
          shadowRoot.appendChild(div1);
          shadowRoot.appendChild(div2);
          document.documentElement.appendChild(div);
-         </script>`
+         </script>`,
       );
       return state;
     }
     it('should find first element in shadow', async () => {
       const {page} = await setUpPage();
-      const div = (await page.$('pierce/.foo')) as ElementHandle<HTMLElement>;
+      using div = (await page.$('pierce/.foo')) as ElementHandle<HTMLElement>;
       const text = await div.evaluate(element => {
         return element.textContent;
       });
@@ -62,15 +52,15 @@ describe('Query handler tests', function () {
           return div.evaluate(element => {
             return element.textContent;
           });
-        })
+        }),
       );
       expect(text.join(' ')).toBe('Hello World');
     });
     it('should find first child element', async () => {
       const {page} = await setUpPage();
-      const parentElement = (await page.$('html > div'))!;
-      const childElement = (await parentElement.$(
-        'pierce/div'
+      using parentElement = (await page.$('html > div'))!;
+      using childElement = (await parentElement.$(
+        'pierce/div',
       )) as ElementHandle<HTMLElement>;
       const text = await childElement.evaluate(element => {
         return element.textContent;
@@ -79,7 +69,7 @@ describe('Query handler tests', function () {
     });
     it('should find all child elements', async () => {
       const {page} = await setUpPage();
-      const parentElement = (await page.$('html > div'))!;
+      using parentElement = (await page.$('html > div'))!;
       const childElements = (await parentElement.$$('pierce/div')) as Array<
         ElementHandle<HTMLElement>
       >;
@@ -88,7 +78,7 @@ describe('Query handler tests', function () {
           return div.evaluate(element => {
             return element.textContent;
           });
-        })
+        }),
       );
       expect(text.join(' ')).toBe('Hello World');
     });
@@ -115,11 +105,11 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div id="1">a</div><div>a</div>');
 
-        const element = await page.$('text/a');
+        using element = await page.$('text/a');
         expect(
           await element?.evaluate(e => {
             return e.id;
-          })
+          }),
         ).toBe('1');
       });
       it('should return multiple elements', async () => {
@@ -145,11 +135,11 @@ describe('Query handler tests', function () {
           document.body.append(div);
         });
 
-        const element = await page.$('text/a');
+        using element = await page.$('text/a');
         expect(
           await element?.evaluate(e => {
             return e.textContent;
-          })
+          }),
         ).toBe('a');
       });
       it('should query deeply nested text', async () => {
@@ -157,11 +147,11 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div><div>a</div><div>b</div></div>');
 
-        const element = await page.$('text/a');
+        using element = await page.$('text/a');
         expect(
           await element?.evaluate(e => {
             return e.textContent;
-          })
+          }),
         ).toBe('a');
       });
       it('should query inputs', async () => {
@@ -169,13 +159,13 @@ describe('Query handler tests', function () {
 
         await page.setContent('<input value="a">');
 
-        const element = (await page.$(
-          'text/a'
+        using element = (await page.$(
+          'text/a',
         )) as ElementHandle<HTMLInputElement>;
         expect(
           await element?.evaluate(e => {
             return e.value;
-          })
+          }),
         ).toBe('a');
       });
       it('should not query radio', async () => {
@@ -190,22 +180,22 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div><span>a</span> <span>b</span><div>');
 
-        const element = await page.$('text/a b');
+        using element = await page.$('text/a b');
         expect(
           await element?.evaluate(e => {
             return e.textContent;
-          })
+          }),
         ).toBe('a b');
       });
       it('should clear caches', async () => {
         const {page} = await getTestState();
 
         await page.setContent(
-          '<div id=target1>text</div><input id=target2 value=text><div id=target3>text</div>'
+          '<div id=target1>text</div><input id=target2 value=text><div id=target3>text</div>',
         );
-        const div = (await page.$('#target1')) as ElementHandle<HTMLDivElement>;
-        const input = (await page.$(
-          '#target2'
+        using div = (await page.$('#target1')) as ElementHandle<HTMLDivElement>;
+        using input = (await page.$(
+          '#target2',
         )) as ElementHandle<HTMLInputElement>;
 
         await div.evaluate(div => {
@@ -214,7 +204,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$eval(`text/text`, e => {
             return e.id;
-          })
+          }),
         ).toBe('target1');
         await div.evaluate(div => {
           div.textContent = 'foo';
@@ -222,7 +212,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$eval(`text/text`, e => {
             return e.id;
-          })
+          }),
         ).toBe('target2');
         await input.evaluate(input => {
           input.value = '';
@@ -231,7 +221,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$eval(`text/text`, e => {
             return e.id;
-          })
+          }),
         ).toBe('target3');
 
         await div.evaluate(div => {
@@ -244,7 +234,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$$eval(`text/text`, es => {
             return es.length;
-          })
+          }),
         ).toBe(3);
         await div.evaluate(div => {
           div.textContent = 'foo';
@@ -252,7 +242,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$$eval(`text/text`, es => {
             return es.length;
-          })
+          }),
         ).toBe(2);
         await input.evaluate(input => {
           input.value = '';
@@ -261,7 +251,7 @@ describe('Query handler tests', function () {
         expect(
           await page.$$eval(`text/text`, es => {
             return es.length;
-          })
+          }),
         ).toBe(1);
       });
     });
@@ -271,7 +261,7 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div class="a"><span>a</span></div>');
 
-        const elementHandle = (await page.$('div'))!;
+        using elementHandle = (await page.$('div'))!;
         expect(await elementHandle.$(`text/a`)).toBeTruthy();
         expect(await elementHandle.$$(`text/a`)).toHaveLength(1);
       });
@@ -281,7 +271,7 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div class="a"></div>');
 
-        const elementHandle = (await page.$('div'))!;
+        using elementHandle = (await page.$('div'))!;
         expect(await elementHandle.$(`text/a`)).toBeFalsy();
         expect(await elementHandle.$$(`text/a`)).toHaveLength(0);
       });
@@ -302,10 +292,10 @@ describe('Query handler tests', function () {
         const {page} = await getTestState();
 
         expect(
-          await page.$('xpath/html/body/non-existing-element')
+          await page.$('xpath/html/body/non-existing-element'),
         ).toBeFalsy();
         expect(
-          await page.$$('xpath/html/body/non-existing-element')
+          await page.$$('xpath/html/body/non-existing-element'),
         ).toHaveLength(0);
       });
       it('should return first element', async () => {
@@ -313,11 +303,11 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div>a</div><div></div>');
 
-        const element = await page.$('xpath/html/body/div');
+        using element = await page.$('xpath/html/body/div');
         expect(
           await element?.evaluate(e => {
             return e.textContent === 'a';
-          })
+          }),
         ).toBeTruthy();
       });
       it('should return multiple elements', async () => {
@@ -335,7 +325,7 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div class="a">a<span></span></div>');
 
-        const elementHandle = (await page.$('div'))!;
+        using elementHandle = (await page.$('div'))!;
         expect(await elementHandle.$(`xpath/span`)).toBeTruthy();
         expect(await elementHandle.$$(`xpath/span`)).toHaveLength(1);
       });
@@ -345,7 +335,7 @@ describe('Query handler tests', function () {
 
         await page.setContent('<div class="a">a</div>');
 
-        const elementHandle = (await page.$('div'))!;
+        using elementHandle = (await page.$('div'))!;
         expect(await elementHandle.$(`xpath/span`)).toBeFalsy();
         expect(await elementHandle.$$(`xpath/span`)).toHaveLength(0);
       });
@@ -360,12 +350,21 @@ describe('Query handler tests', function () {
     it('should work with CSS selectors', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('div > button');
+      using element = await page.$('div > button');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
+      ).toBeTruthy();
+
+      using root = await page.$('div');
+      using button = await root!.$('& > button');
+      assert(button, 'Could not find element');
+      expect(
+        await button.evaluate(element => {
+          return element.id === 'b';
+        }),
       ).toBeTruthy();
 
       // Should parse more complex CSS selectors. Listing a few problematic
@@ -374,21 +373,34 @@ describe('Query handler tests', function () {
         '.user_row[data-user-id="\\38 "]:not(.deactivated_user)',
         `input[value='Search']:not([class='hidden'])`,
         `[data-test-id^="test-"]:not([data-test-id^="test-foo"])`,
+        `& > table`,
       ]) {
         await page.$$(selector);
       }
     });
 
+    it('should work with puppeteer pseudo classes', async () => {
+      const {server, page} = await getTestState();
+      await page.goto(`${server.PREFIX}/p-selectors.html`);
+      using element = await page.$('button::-p-text(world)');
+      assert(element, 'Could not find element');
+      expect(
+        await element.evaluate(element => {
+          return element.id === 'b';
+        }),
+      ).toBeTruthy();
+    });
+
     it('should work with deep combinators', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
       {
-        const element = await page.$('div >>>> div');
+        using element = await page.$('div >>>> div');
         assert(element, 'Could not find element');
         expect(
           await element.evaluate(element => {
             return element.id === 'c';
-          })
+          }),
         ).toBeTruthy();
       }
       {
@@ -397,7 +409,7 @@ describe('Query handler tests', function () {
         expect(
           await elements[1]?.evaluate(element => {
             return element.id === 'd';
-          })
+          }),
         ).toBeTruthy();
       }
       {
@@ -406,7 +418,7 @@ describe('Query handler tests', function () {
         expect(
           await elements[0]?.evaluate(element => {
             return element.id === 'd';
-          })
+          }),
         ).toBeTruthy();
       }
       {
@@ -415,7 +427,7 @@ describe('Query handler tests', function () {
         expect(
           await elements[0]?.evaluate(element => {
             return element.id === 'd';
-          })
+          }),
         ).toBeTruthy();
       }
     });
@@ -423,81 +435,81 @@ describe('Query handler tests', function () {
     it('should work with text selectors', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('div ::-p-text(world)');
+      using element = await page.$('div ::-p-text(world)');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
     it('should work ARIA selectors', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('div ::-p-aria(world)');
+      using element = await page.$('div ::-p-aria(world)');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
     it('should work for ARIA selectors in multiple isolated worlds', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      let element = await page.waitForSelector('::-p-aria(world)');
+      using element = await page.waitForSelector('::-p-aria(world)');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
       // $ would add ARIA query handler to the main world.
       await element.$('::-p-aria(world)');
-      element = await page.waitForSelector('::-p-aria(world)');
-      assert(element, 'Could not find element');
+      using element2 = await page.waitForSelector('::-p-aria(world)');
+      assert(element2, 'Could not find element');
       expect(
-        await element.evaluate(element => {
+        await element2.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
     it('should work ARIA selectors with role', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('::-p-aria(world[role="button"])');
+      using element = await page.$('::-p-aria(world[role="button"])');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
     it('should work ARIA selectors with name and role', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('::-p-aria([name="world"][role="button"])');
+      using element = await page.$('::-p-aria([name="world"][role="button"])');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
     it('should work XPath selectors', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('div ::-p-xpath(//button)');
+      using element = await page.$('div ::-p-xpath(//button)');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'b';
-        })
+        }),
       ).toBeTruthy();
     });
 
@@ -510,12 +522,12 @@ describe('Query handler tests', function () {
 
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      const element = await page.$('::-p-div');
+      using element = await page.$('::-p-div');
       assert(element, 'Could not find element');
       expect(
         await element.evaluate(element => {
           return element.id === 'a';
-        })
+        }),
       ).toBeTruthy();
     });
 
@@ -533,39 +545,39 @@ describe('Query handler tests', function () {
       });
 
       {
-        const element = await page.$('::-p-div(true)');
+        using element = await page.$('::-p-div(true)');
         assert(element, 'Could not find element');
         expect(
           await element.evaluate(element => {
             return element.id === 'a';
-          })
+          }),
         ).toBeTruthy();
       }
       {
-        const element = await page.$('::-p-div("true")');
+        using element = await page.$('::-p-div("true")');
         assert(element, 'Could not find element');
         expect(
           await element.evaluate(element => {
             return element.id === 'a';
-          })
+          }),
         ).toBeTruthy();
       }
       {
-        const element = await page.$("::-p-div('true')");
+        using element = await page.$("::-p-div('true')");
         assert(element, 'Could not find element');
         expect(
           await element.evaluate(element => {
             return element.id === 'a';
-          })
+          }),
         ).toBeTruthy();
       }
       {
-        const element = await page.$('::-p-div');
+        using element = await page.$('::-p-div');
         assert(element, 'Could not find element');
         expect(
           await element.evaluate(element => {
             return element.id === 'b';
-          })
+          }),
         ).toBeTruthy();
       }
     });
@@ -573,14 +585,13 @@ describe('Query handler tests', function () {
     it('should work with :hover', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      let button = await page.$('div ::-p-text(world)');
+      using button = await page.$('div ::-p-text(world)');
       assert(button, 'Could not find element');
       await button.hover();
-      await button.dispose();
 
-      button = await page.$('div ::-p-text(world):hover');
-      assert(button, 'Could not find element');
-      const value = await button.evaluate(span => {
+      using button2 = await page.$('div ::-p-text(world):hover');
+      assert(button2, 'Could not find element');
+      const value = await button2.evaluate(span => {
         return {textContent: span.textContent, tagName: span.tagName};
       });
       expect(value).toMatchObject({textContent: 'world', tagName: 'BUTTON'});
@@ -597,7 +608,7 @@ describe('Query handler tests', function () {
       const results: T[][] = [];
       for (let i = 0; i < inputs.length; ++i) {
         const permutation = permute(
-          inputs.slice(0, i).concat(inputs.slice(i + 1))
+          inputs.slice(0, i).concat(inputs.slice(i + 1)),
         );
         const value = inputs[i] as T;
         if (permutation.length === 0) {
@@ -620,14 +631,14 @@ describe('Query handler tests', function () {
             .map(selector => {
               return selector === 'button' ? '::-p-text(world)' : selector;
             })
-            .join(',')
+            .join(','),
         );
         const actual = await Promise.all(
           elements.map(element => {
             return element.evaluate(element => {
               return element.id;
             });
-          })
+          }),
         );
         expect(actual.join()).toStrictEqual('a,b,f,c');
       }
@@ -643,22 +654,22 @@ describe('Query handler tests', function () {
     it('should handle escapes', async () => {
       const {server, page} = await getTestState();
       await page.goto(`${server.PREFIX}/p-selectors.html`);
-      let element = await page.$(
-        ':scope >>> ::-p-text(My name is Jun \\(pronounced like "June"\\))'
+      using element = await page.$(
+        ':scope >>> ::-p-text(My name is Jun \\(pronounced like "June"\\))',
       );
       expect(element).toBeTruthy();
-      element = await page.$(
-        ':scope >>> ::-p-text("My name is Jun (pronounced like \\"June\\")")'
+      using element2 = await page.$(
+        ':scope >>> ::-p-text("My name is Jun (pronounced like \\"June\\")")',
       );
-      expect(element).toBeTruthy();
-      element = await page.$(
-        ':scope >>> ::-p-text(My name is Jun \\(pronounced like "June"\\)")'
+      expect(element2).toBeTruthy();
+      using element3 = await page.$(
+        ':scope >>> ::-p-text(My name is Jun \\(pronounced like "June"\\)")',
       );
-      expect(element).toBeFalsy();
-      element = await page.$(
-        ':scope >>> ::-p-text("My name is Jun \\(pronounced like "June"\\))'
+      expect(element3).toBeFalsy();
+      using element4 = await page.$(
+        ':scope >>> ::-p-text("My name is Jun \\(pronounced like "June"\\))',
       );
-      expect(element).toBeFalsy();
+      expect(element4).toBeFalsy();
     });
   });
 });
